/*
 * O.Filangi : 1ere version d'une implementation linaeraisation de la vraisemblance pour un modèle homoscedastique.
 *
 * nvcc --cuda analyse_lin_cuda.cu
 */


#define __PRINT_DEBUG_ANALYSE 1

/* uncomment to print information about analysis */
/* #define __PRINT_DEBUG_ANALYSE 1 */ 

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <cuda.h>
#include <cublas.h>

extern "C" {
#include "struct_qtlmap.h"
}
#define N 8192                 // Matrix s i z e
#define BLOCK_SIZE 16          // Block s i z e ( must d i v i d e N and be <= 22)
#define DISPLAY 0              // 1 t o e n a b l e m a t r i x d i s p l a y (0 t o d i s a b l e )

#define CUDA_MAX_BLOCK_SIZE 22
#if (N % BLOCK_SIZE != 0 )
#error "BLOCK_SIZE must divide matrix size N"
#endif
#if (BLOCK_SIZE > CUDA_MAX_BLOCK_SIZE)
#error "BLOCK_SIZE must be inferior or equal to 22"
#endif
#define elem ( i , j ) m[ j ∗n+i ]
#define MIN( x , y ) ( y < x ? y : x )

/*********************************************************************************************************************************************************************************************
*
*   KERNEL - GPU
*
**********************************************************************************************************************************************************************************************/

/*********************************************************************
* ETAPE 1 : Etape de descente dans la resolution a partir de la matrice triangulaire decompose par la methode Cholesky
*
* On appelle NTNIV fois la fonction kernel
* chaqua thread execute pour la solution RHS
* RHS(iniv_thread) = RHS(iniv_thread) - (RHS(iniv) / XX(iniv,iniv)) * XX(iniv,iniv_thread) 
*
*
**********************************************************************/

__global__ void downstep_cholesky(DT *xx,DT *rhs,bool * vecsol, int ntniv) {
    
    int ithread = threadIdx.x; // calcul de t(iniv) = rhs(iniv) - t(1) * XX(1,iniv) - ... - t(iniv-1) * XX(iniv-1,iniv)
    int bsize   = blockDim.x;
   
    for (int iniv = 0;iniv<bsize;iniv++) {
     if ( ithread > iniv && vecsol[ithread] && vecsol[iniv] ) {
       rhs[ithread] = rhs[ithread] - (rhs[iniv] / xx[iniv*ntniv+iniv]) * xx[ithread*ntniv+iniv] ;
     }
     __syncthreads();
    }
}

/*********************************************************************
* ETAPE 2 et 4: Rescale de RHS. Chaque niveau est divise par la diagonale superieure de la matrice triangulaire CHOLESKY
*
* On appelle NTNIV fois la fonction kernel
* chaqua thread execute pour la solution RHS
* RHS(iniv_thread) =  (RHS(iniv_thread) / XX(iniv_thread,iniv_thread))
*
*
**********************************************************************/

__global__ void scale_solution(DT *xx, DT *rhs,bool * vecsol,int ntniv) {
    
    int ithread = threadIdx.x;
    
    if ( vecsol[ithread] ) {
       rhs[ithread] = rhs[ithread] / xx[ithread*ntniv+ithread] ;
    }
}

/*********************************************************************
* ETAPE 3 : Etape de remonte dans la resolution a partir de la matrice triangulaire decompose par la methode Cholesky
*
* On appelle NTNIV fois la fonction kernel
* chaqua thread execute pour la solution RHS
* RHS(iniv_thread) = RHS(iniv_thread) - (RHS(iniv) / XX(iniv,iniv)) * XX(iniv,iniv_thread) 
*
*
**********************************************************************/
__global__ void upstep_cholesky(DT *xx,DT *rhs,bool * vecsol, int ntniv) {
    
    int ithread = threadIdx.x; // calcul de t(iniv) = rhs(iniv) - t(1) * XX(1,iniv) - ... - t(iniv-1) * XX(iniv-1,iniv)
    int bsize   = blockDim.x;
   
   for (int iniv = bsize-1;iniv>=0;iniv--) {
     if ( ithread < iniv && vecsol[ithread] && vecsol[iniv] ) {
        rhs[ithread] = rhs[ithread] - (rhs[iniv] / xx[iniv*ntniv+iniv]) * xx[iniv*ntniv+ithread] ;
     }
     __syncthreads();
    }
}



/*********************************************************************
* Pour ne pas prenre en compte les parametres non estimable, on les fixe a zero
*
*
*
**********************************************************************/


__global__ void fix_estimability_rhs(DT *rhs,bool * vecsol) {
    
    int ithread = threadIdx.x; 
   
    if ( ! vecsol[ithread] ) {
       rhs[ithread] = 0.0 ;
    }
}


__global__ void copy_upper_triang_to_lower(DT *xx,int ntniv) {
    
    int x = threadIdx.x;
    int y = threadIdx.y;
    
    if ( x > y ) {
      xx[y*ntniv+x] = xx[x*ntniv+y] ;
    }
}

__global__ void set_zero_lower_triang(DT *xx,int ntniv) {
    
    int x = threadIdx.x;
    int y = threadIdx.y;
    
    if ( x > y ) {
      xx[y*ntniv+x] = 0.0 ;
    }
}

/******************************/
/* Colesky CUDA iterative     */
/******************************/


/*
 * : ptrDv
*/
__global__ void cuda_chol_iter(DT *m, bool *dVecsol, int n,float seuil_cho) {

   int i;
   int x = threadIdx.x ;
   int y = threadIdx.y ;

  __shared__ DT b[CUDA_MAX_BLOCK_SIZE][CUDA_MAX_BLOCK_SIZE];
   
   b[x][y] = m[y*n + x];
   __syncthreads();
   
   for (i=0;i<n;i++) {
      if ( x == i ) {
         DT fac = sqrtf(b[x][x]);
         if ( y >= x) {
           b[x][y] /= fac;
           if ( x == y ) dVecsol[x] = (b[x][x] > seuil_cho) ;
         }
       }  
     __syncthreads(); 
     
     
     if ( x > i && y >= x && dVecsol[i] ) {
      b[x][y] -= b[i][y] * b[i][x];
     }
   __syncthreads();
   }
   m[y*n + x] = b[x][y];
}

/*********************************************************************************************************************************************************************************************
*
*   CPU
*
**********************************************************************************************************************************************************************************************/


void cholesky_cuda(DT *dm, bool *dVecsol, int ntniv, float seuil_cho) {
   dim3 threads(ntniv,ntniv);
   cuda_chol_iter<<<1,threads>>>(dm,dVecsol,ntniv,seuil_cho);
   cudaThreadSynchronize();
}

/*
void cholesky_cuda_block(float *dm, int n, int bsize) {
  
   int iter = (n + bsize -1) / bsize ;
   int i;
   
   dim3 threads(bsize,bsize);
   
   for (i=0;i<iter;i++) {
     //top left block
     cuda_chol_iter(dm,n,);
   
   }
  
}
*/




void cholesky_cpu(float *XX,int ntniv,bool *vecsol,float seuil_cho,float *triang, int*nbnivest) {

    *nbnivest = 0 ;
    //float triang[ntniv][ntniv;
    
    for (int i=0;i<ntniv;i++) {
      for (int j=0;j<ntniv;j++) {
        triang[j*ntniv+i] = 0.0;
      }
    }
    
    for (int j=0;j<ntniv;j++) {
       triang[j*ntniv+j] = XX[j*ntniv+j] ;
       //printf("%i ============>%f\n",j,XX[j*ntniv+j]);
       for (int k=0;k<=j-1;k++) {
          triang[j*ntniv+j] = triang[j*ntniv+j] - triang[k*ntniv+j]*triang[k*ntniv+j];
         // printf("============>%f %f\n", triang[j*ntniv+j],triang[k*ntniv+j]);
       }
      // printf("%i %f %f\n",j,triang[j*ntniv+j],seuil_cho);
       //si estimable
       if ( triang[j*ntniv+j] > seuil_cho ) {
          vecsol[j] = true;
        //  printf("vecsol(%i) : TRUE\n",j);
          *nbnivest = *nbnivest + 1 ;
          triang[j*ntniv+j] =  sqrt(triang[j*ntniv+j]);
          for (int i=j+1;i<ntniv;i++ ) {
            triang[j*ntniv+i] = XX[i*ntniv+j];
            for ( int k=0;k<=j-1;k++) {
               triang[j*ntniv+i] = triang[j*ntniv+i] -  (triang[k*ntniv+i] * triang[k*ntniv+j]);
               //printf("============>%f %f %f\n",triang[i][j],triang[i][k],triang[j][k]);
            }
            //XX[j*ntniv+i]=XX[j*ntniv+i]/XX[j*ntniv+j];
            triang[j*ntniv+i] = triang[j*ntniv+i] / triang[j*ntniv+j] ;
            triang[i*ntniv+j] = triang[j*ntniv+i] ;
          }
       } else {
         // printf("vecsol(%i) : FALSE\n",j);
         vecsol[j] = false;
       }
    }     
         
}

/*********************************************************************************************************************************************************************************************
*
*   INTERFACE WITH QTLMAP ( FORTRAN )
    * cuda_set_qtlmap_user_struct_(float *y, int *ndP, int *ntnivmaxP,ADD_FORTRAN * idUrlStructUserData) 
*   * cuda_init_qtlmap_process_struct_(int *nkdP, int *ntnivP,ADD_FORTRAN * idUrlStruct)                                : initialize the contexte
*   * cuda_homoscedatic_resolv_(float *xinc,int *nkdPtr,int*ntnivPtr,double * seuil_choPtr,ADD_FORTRAN * idUrlStruct )    : apply analysis
*   * cuda_free_qtlmap_struct_process_(ADD_FORTRAN * idUrlStruct)                                                                 ! release memory
*   * cuda_free_qtlmap_struct_user_(ADD_FORTRAN * idUrlStruct)
*
**********************************************************************************************************************************************************************************************/

/*
  xinc  = matrice d'incidence
  nkd   = nombre de ligne
  ntniv = nombre de colonne
  seuil_choPtr = seuil d'estimabilite pour la solution
  
  column major order ===> offset = row + column*NUMROWS
  
*/
extern "C" void cuda_homoscedatic_resolv_(ADD_FORTRAN * idUrlStructUser,        /* IN  : Structure contenant les adresses memoires des donnees de l'utilisateur (performance) sur la carte GPU*/
                                          ADD_FORTRAN * idUrlStructProcess,     /* IN  : Structure contenant les adresses memoires des structures temporaires pour chaque process sur la carte GPU*/
                                          double *xinc_d,                       /* IN  : matrice de contingence */
                                          int * nd,                             /* IN  : nombre d'individu total  : nombre de ligne reelle de XINC */
                                          int * ntnivmax,                       /* IN  : nombre maximum de niveau : nombre de colonne reelle de XINC */
                                          int *ntnivPtr,                        /* IN  : nombre de niveau a estimer : M Colonnes de la matrice d'incidence */
                                          double * seuil_choPtr,                /* IN  : seuil pour considere grace a la decomposition de cholesky si un niveau est estimable */
                                          int    * vecsol,                      /* OUT : tableau de booleen de taille NTNIV  pour l'estimabilite de chaque niveau */
                                          double * bestim,                      /* OUT : tableau solution de l'estimation de chaque niveau (taille NTNIV) */
                                          double * osigsquare,                  /* OUT : variance residuelle pour chaque famille de pere */
                                          int    * npP        ,                 /* IN  : nombre de famille de pere */        
                                          bool * performPrecision,              /* IN  : indicateur si le tableau tempForConfusion doit etre calcule */
                                          double * tempForConfusion             /* OUT : tableau a remplir si  performPrecision est a VRAI */
                                           ) {     
      
        QTLMapDevicePtrStructProcess * idStructProcess ;
        QTLMapDevicePtrStructUser    * idStructUser ;
        
        float seuil_cho = *seuil_choPtr;
        int incx=1,incy=1;
        float alpha,beta;
        cublasStatus status = 0;
        int ntniv = *ntnivPtr;
        
        
        /* Cast de l'adresse stocker dans un entier long en veritable adressage C */
        idStructProcess  = (QTLMapDevicePtrStructProcess *) (*idUrlStructProcess) ;
        idStructUser     = (QTLMapDevicePtrStructUser *) (*idUrlStructUser) ;
        
#if(__PRINT_DEBUG_ANALYSE)
        printf("USER Adressage passe en argument       :%p\n",idStructUser);
        printf("USER Valeur du cast en entier idStruct :%lld\n",(ADD_FORTRAN)idStructUser);
        printf("PROCESS Adressage passe en argument       :%p\n",idStructProcess);
        printf("PROCESS Valeur du cast en entier idStruct :%lld\n",(ADD_FORTRAN)idStructProcess);
        printf("NKD     :%i\n",idStructUser->nkd);
        printf("NTNIVMAX:%i\n",idStructProcess->ntnivmax);
        printf("NTNIV   :%i\n",ntniv);
#endif
        /* CAST DU TABLEAU XINC */
        
        DT * xinc = (DT *)malloc(sizeof(DT)*(idStructUser->nkd)*ntniv);
        for (int i=0;i<idStructUser->nkd;i++) {
         for (int j=0;j<ntniv;j++ ){
           xinc[j*(idStructUser->nkd)+i] = (DT) ( xinc_d[j*(*nd) +i] );
           //printf(" %f ",( xinc_d[j*(*nd) +i] ));
          }
          //printf("\n");
        }
        
        /*
          column major order : Passage par fortran : pour obtenir XINC(i,j) => xinc[j*nkd + i]
        */
       
        /* Transfert de la matrice de contingence sur le device */
        status = cublasSetMatrix(idStructUser->nkd,ntniv,sizeof(DT),xinc,idStructUser->nkd,idStructProcess->devicePtrXinc,idStructUser->nkd);
        EXIT_ERROR_USE_DEVICE(status,"cublasSetMatrix","devicePtrXinc");
        free(xinc);
        
        /*--------------------------------------------------------------------------------------------------------------*/
        //calcul of XX via ** CUBLAS ** 
#if(__PRINT_DEBUG_ANALYSE)
        printf("\n=============================== CUBLAS:cublasSsyrk  X . X ==============================================\n") ;
#endif
        alpha=1.0;
        beta=0.0;
        cublasSsyrk ('U','T', ntniv, idStructUser->nkd,alpha, (DT *)idStructProcess->devicePtrXinc,idStructUser->nkd,beta,(DT *)idStructProcess->devicePtrXX,ntniv);
#if(__PRINT_DEBUG_ANALYSE)        
       debug_print_gpu_memory_XX(idStructProcess,ntniv);
#endif              
        // ===================== CHOLESKY CPU ===================================
        //DT triang[ntniv*ntniv];
        
       // cholesky_cpu(XX,ntniv,vecsol,seuil_cho,triang,&nbnivest) ;
        // out => nbnivest nombre de niveau estimable
        // ON MET A JOUR SUR LE DEVICE VECSOL ET XX .........A EVITER
       /*  
        status = cublasSetMatrix(ntniv,ntniv,sizeof(float),triang,ntniv,idStructProcess->devicePtrXX,ntniv);
        if (status != CUBLAS_STATUS_SUCCESS) {
            fprintf (stderr, "!!!! cublasSetMatrix error (devicePtrXX)\n");
            exit(1);
        }
        status = cublasSetVector(ntniv,sizeof(bool),vecsol,incx,idStructProcess->devicePtrVecsol,incy);
        if (status != CUBLAS_STATUS_SUCCESS) {
            fprintf (stderr, "!!!! cublasSetMatrix error (devicePtrVecsol)\n");
            exit(1);
        }
        */
        
        // ===================== CHOLESKY GPU ===================================
#if(__PRINT_DEBUG_ANALYSE)
        printf("\n================================ Decomposition de Cholesky A = L . L' ==========================================\n") ;
#endif
         
       cholesky_cuda((DT *)idStructProcess->devicePtrXX,(bool *)idStructProcess->devicePtrVecsol,ntniv,seuil_cho);
       dim3 threads(ntniv,ntniv);
       //copy_upper_triang_to_lower<<<1,threads>>>((DT *)idStructProcess->devicePtrXX,ntniv);
       set_zero_lower_triang<<<1,threads>>>((DT *)idStructProcess->devicePtrXX,ntniv);
#if(__PRINT_DEBUG_ANALYSE)        
       debug_print_gpu_memory_XX(idStructProcess,ntniv);
       debug_print_gpu_memory_Vecsol(idStructProcess,ntniv);
#endif
        /* construction de la solution RHS = Xr'.Y */
        //on prend pas en compte nbnivest.....
#if(__PRINT_DEBUG_ANALYSE)
        printf("\n================================ Calcul de RHS Cublas cublasSgemv : RHS = XINC . Y ==========================================\n") ;
#endif
        //calcul of RHS via ** CUBLAS **
        alpha = 1.0 ;
        beta  = 0.0 ;     
        cublasSgemv('T',idStructUser->nkd,ntniv,alpha,(DT *)idStructProcess->devicePtrXinc,idStructUser->nkd,(DT *)(idStructUser->devicePtrY),incx,beta,(DT *)(idStructProcess->devicePtrRHS),incy);
        
#if(__PRINT_DEBUG_ANALYSE)        
       debug_print_gpu_memory_RHS(idStructProcess,ntniv);
#endif
        
#if(__PRINT_DEBUG_ANALYSE)
        printf("\n================================ Resolution LU a partir de RHS,Vecsol => BESTIM est contenue dans RHS ==========================================\n") ;
#endif        
        /* Etape de descente dans la resolution LU a partir de la matrice triangulaire L (decomposition Cholesky) */
        /*
           la solution RHS(i) depend des resultats calcule iterativement RHS(i-1),...,RHS(1).
           Parallelisme.
           Pour NTNIV etape, on soutrait la solution RHS(i-1) trouve a l iteration precedente, a tous les RHS(i') avec i'>i 
        */
        //for (int iniv=0;iniv<ntniv;iniv++) {
           downstep_cholesky<<<1,ntniv>>>((DT *)idStructProcess->devicePtrXX,(DT *)idStructProcess->devicePtrRHS,(bool *)idStructProcess->devicePtrVecsol,ntniv);
	  // cudaThreadSynchronize();
        //}
        
        scale_solution<<<1,ntniv>>>((DT *)idStructProcess->devicePtrXX,(DT *)idStructProcess->devicePtrRHS,(bool *)idStructProcess->devicePtrVecsol,ntniv);
        cudaThreadSynchronize();
#if(__PRINT_DEBUG_ANALYSE)  
        printf("\n              ================================ Etape Descente Resolution LU\n"); 
        debug_print_gpu_memory_RHS(idStructProcess,ntniv);
#endif
        /* Etape de remonte dans la resolution LU a partir de la matrice triangulaire L (decomposition Cholesky) */
        //for (int iniv=ntniv-1;iniv>=0;iniv--) {
           upstep_cholesky<<<1,ntniv>>>((DT *)idStructProcess->devicePtrXX,(DT *)idStructProcess->devicePtrRHS,(bool *)idStructProcess->devicePtrVecsol,ntniv);
	  
        //}
        
        scale_solution<<<1,ntniv>>>((DT *)idStructProcess->devicePtrXX,(DT *)idStructProcess->devicePtrRHS,(bool *)idStructProcess->devicePtrVecsol,ntniv);
        cudaThreadSynchronize();

#if(__PRINT_DEBUG_ANALYSE)    
        printf("\n              ================================ Etape Monte Resolution LU\n"); 
        debug_print_gpu_memory_RHS(idStructProcess,ntniv);
#endif
        /* On met a zero les RHS(I) tel que VECSOL(I) == FAUX */
        fix_estimability_rhs<<<1,ntniv>>>((DT *)idStructProcess->devicePtrRHS,(bool *)idStructProcess->devicePtrVecsol); 
#if(__PRINT_DEBUG_ANALYSE)        
       debug_print_gpu_memory_Vecsol(idStructProcess,ntniv);
       debug_print_gpu_memory_RHS(idStructProcess,ntniv);
#endif

#if(__PRINT_DEBUG_ANALYSE)
        printf("\n================================ Cublas cublasSgemv : XB = XINC * BEstim ==========================================\n") ;
#endif    
        /* Calcul de la valeur XB = XINC * Solution  . Note RHS est devenue Bestim  */
        /* Pour prendre en compte le tableau vecsol pour les valeur estimable */
        /* XB = XINC . RHS - XINC(:,INIV).RHS pour chaque iniv tel que vecsol(iniv) = false */
        
        
        // XB = XINC . RHS
        alpha = 1.0 ;
        beta  = 0.0 ;
        cublasSgemv('N',
                    idStructUser->nkd,
                    ntniv,
                    alpha,
                    (DT *)idStructProcess->devicePtrXinc,
                    idStructUser->nkd,
                    (DT *)idStructProcess->devicePtrRHS,
                    incx,
                    beta,
                    (DT *)idStructProcess->devicePtrXB,
                    incy);
#if(__PRINT_DEBUG_ANALYSE)        
       debug_print_gpu_memory_XB(idStructProcess,idStructUser->nkd,ntniv);
#endif        
        
        /* on soustrait les valeurs*/
        //for (in)
#if(__PRINT_DEBUG_ANALYSE)
        printf("\n================================ Cublas cublasSscal : XB = -XB ==========================================\n") ;
#endif   
        // XB = -XB
        alpha = -1.0 ;
        cublasSscal(idStructUser->nkd,alpha,(DT *)idStructProcess->devicePtrXB,incx);
#if(__PRINT_DEBUG_ANALYSE)        
       debug_print_gpu_memory_XB(idStructProcess,idStructUser->nkd,ntniv);
#endif       

#if(__PRINT_DEBUG_ANALYSE)
        printf("\n================================ Cublas cublasSaxpy : XB = Y + (-XB) ==========================================\n") ;
#endif   
        // XB = Y + (-XB)
        alpha=1.0 ;
        
        cublasSaxpy(idStructUser->nkd,alpha,(DT *)idStructUser->devicePtrY,incx,(DT *)idStructProcess->devicePtrXB,incy);
#if(__PRINT_DEBUG_ANALYSE)        
       debug_print_gpu_memory_XB(idStructProcess,idStructUser->nkd,ntniv);
#endif       

#if(__PRINT_DEBUG_ANALYSE)
        printf("\n================================ Cublas cublasSdot : SISQ = XB . XB ==========================================\n") ;
#endif                   
        osigsquare[0] = (double) cublasSdot(idStructUser->nkd,(DT *)idStructProcess->devicePtrXB,incx,(DT *)idStructProcess->devicePtrXB,incx) ;
        osigsquare[0] = osigsquare[0] / idStructUser->nkd ;
        for ( int i=1;i<(*npP);i++) {
          osigsquare[i] = osigsquare[0];  
        }

        /* Recuperation des tableaux resultats */ 
        
        /* vecsol */
        bool vecsolT[ntniv];
        status = cublasGetVector(ntniv,sizeof(bool),idStructProcess->devicePtrVecsol,incx,vecsolT,incy);
        EXIT_ERROR_USE_DEVICE(status,"cublasGetVector","devicePtrVecsol");
        for ( int i=0;i<ntniv;i++) {
          vecsol[i] = (int)vecsolT[i];
        }
        
        /* bestim */
        
        DT bestim_temp[ntniv];
        status = cublasGetVector(ntniv,sizeof(DT),idStructProcess->devicePtrRHS,incx,bestim_temp,incy);
        EXIT_ERROR_USE_DEVICE(status,"cublasGetVector","devicePtrVecsol");
        int ii=-1;
        for (int i=0;i<ntniv;i++) { 
           if ( vecsol[i] ) { 
            ii++;
            bestim[ii] = (double)bestim_temp[i];
           }
        } ;

#if(__PRINT_DEBUG_ANALYSE)     
       printf("\n========================== RESULT ============================================");
       printf("\n\nSIGSQ    =%f\n\n",(*(osigsquare+1)));
       debug_print_gpu_memory_Vecsol(idStructProcess,ntniv);
       debug_print_gpu_memory_RHS(idStructProcess,ntniv);
#endif       
       
       
        return;
}




