/** \file expression.h Implements a qtlmap parser class group. */

#ifndef EXPRESSION_H
#define EXPRESSION_H

//To disable the asserts
//#define NDEBUG

#include <map>
#include <vector>
#include <stdexcept>
#include <ostream>
#include <sstream>
#include <assert.h>

using namespace std;

/**
 * QtlMapNode is the abstract base class for QTLMap parser nodes.
 */
class QtlMapNode
{
public:
	/// required for virtual functions.
	virtual ~QtlMapNode()
	{
	}

	/// output the calculation tree to the given stream. tries to format the
	/// output to make tree levels visible.
	virtual void	print(std::ostream &os, unsigned int depth=0) const = 0;
    /// check the coherency of the object
	virtual bool	check(string & message) const = 0;

	/// helper function for print() which makes the indention string
	static inline std::string indent(unsigned int d)
	{
		return std::string(d * 2, ' ');
	}
};

/**
 * Genealogy generic node.
 */
class  GenealogyGenNode : public QtlMapNode
{
public:

public:

	explicit GenealogyGenNode()
	: QtlMapNode()
	{
	}

	virtual ~GenealogyGenNode()
	{
	}
};

/**
 * Model description of a trait : correction with fixed effect or covariate and qtl effect
 */
class  ModelNode : public QtlMapNode
{
public:

	vector<string> lfeOrCov; /// list of fixed effect or covariate which interact in the model
	vector<string> lfeQtl;   /// list of fixed effect with a qtl effect


public:
    /// constructor of a model
	explicit ModelNode()
	: QtlMapNode()
	{
	}

	virtual ~ModelNode()
	{
	}
    /**
     *  print the model
     */
	virtual void print(std::ostream &os, unsigned int depth) const
    {
		string model="";

		for (unsigned int i=0;i<lfeOrCov.size();i++) {
			model=model+lfeOrCov[i]+"+";
		}
		for (unsigned int i=0;i<lfeQtl.size();i++) {
			model=model+lfeQtl[i]+"xQTL+";
		}
		if ( model.size()>0) {
			os << indent(depth) << "model [" << model.substr(0,model.size()-1) << "]" <<  std::endl;
		}
	}

	/**
	 * Check internal model :
	 * nothing to check, the model can be empty
	 */
	virtual bool check(string & message) const {
		return true;
	}
};

/**
 * Generaic object phenotype, abstract class
 */
class  ObjectPhenoNode : public QtlMapNode
{
public:
	string name ; /// name of the phenotype
	ModelNode * model;             /// model of the trait

public:

	explicit ObjectPhenoNode()
	: QtlMapNode()
	{
		name="";
		model=NULL;
	}

	virtual ~ObjectPhenoNode()
	{
		delete model;
	}

	virtual void getPositions(vector<int> & positions) const = 0 ;
	virtual bool checkPosition(ObjectPhenoNode * obj,string & message) const = 0;
};

/**
 * Trait Node : description of of real trait referenced in a phenotype file
 */
class  TraitNode : public ObjectPhenoNode
{
public:
	int position ;                 /// index position the datafile (colomn or line depends the context)
	string type  ;                 /// type of the trait : integer, real, categorial, average
	double heritability ;          /// heritability of the trait
	bool isUndefinedValue ;        /// true if the undefinedValue is set
	double undefinedValue ;        /// value of the undefinedValue
	bool isPresentPosition ;       /// true if presentPosition is set
	int presentPosition ;          /// index of the present value : 1 the trait is available, 0 is not.
	bool isCensoredPosition ;      /// true if the censoredPosition is set
	int censoredPosition ;         /// index position of the censoredValue : available only for average trait

public:
    /**
     * constructor
     */
	explicit TraitNode()
	: ObjectPhenoNode()
	{
		position=0;
		name="";
		isUndefinedValue=false;
		isPresentPosition=false;
		isCensoredPosition=false;
	}
    /**
     * Destructor
     */
	virtual ~TraitNode()
	{
	}

	/**
	 * print the information about the object
	 */
	virtual void print(std::ostream &os, unsigned int depth) const
	{
		os << indent(depth) << "trait [" << std::endl;
		os << indent(depth) << "name :" << name << std::endl;
		os << indent(depth) << "position :" << position << std::endl;
		os << indent(depth) << "type :" << type << std::endl;
		os << indent(depth) << "heritability :" << heritability << std::endl;
		if ( isUndefinedValue ) {
			os << indent(depth) << "undifinedValue :" << undefinedValue << std::endl;
		}
		if ( isPresentPosition ) {
			os << indent(depth) << "presentPosition :" << presentPosition << std::endl;
		}
		if ( isCensoredPosition ) {
			os << indent(depth) << "censoredPosition :" << censoredPosition << std::endl;
		}
		if ( model ) {
			model->print(os,depth+1);
		}
		os << indent(depth) << "]" << std::endl;
	}

	/**
	 * Check the trait
	 * 1 : name is not empty
	 * 2 : position > 0
	 * 3 : undefinedValue or presentPosition have to be set
	 * 4 : censor position is set if average data
	 * 5 : heritabilit >= 0 and heritability <= 1.0
	 * 6 : position, presentPosition,censorPosition can have the same value
	 */
	virtual bool check(string & message) const {
       if ( name == "" ) {
          message="name trait is empty";
          return false;
       }

       if (position < 1) {
          message="["+name+"]"+"bad value of the position";
          return false;
       }

       if (!isUndefinedValue && !isPresentPosition) {
    	   message="["+name+"] none undefined values or a present positon in the datafile.";
    	   return false;
       }

       if ( type=="average" && !isCensoredPosition ) {
    	   message="["+name+"]"+"censor position is not defined (average data)";
    	   return false;
       }

       if ( heritability < 0.0 || heritability > 1.0 ) {
    	   message = "["+name+"]"+"heritability is a real < 0.0 and > 1.0";
    	   return false;
       }

       if ( isPresentPosition && (position == presentPosition) ) {
    	   message="["+name+"]"+"position and presentPosition value have the same value";
           return false;
       }
       if ( isCensoredPosition && (position == censoredPosition) ) {
           message="["+name+"]"+"position and censorPosition value have the same value";
           return false;
       }

       if ( isCensoredPosition && isPresentPosition && (presentPosition == censoredPosition) ) {
           message="["+name+"]"+"presentPosition and censorPosition value have the same value";
           return false;
       }

       return true;
	}

	virtual void getPositions(vector<int> & positions) const {
		positions.push_back(position);
		if ( isPresentPosition ) positions.push_back(presentPosition);
		if ( isCensoredPosition ) positions.push_back(censoredPosition);

	}

	virtual bool checkPosition(ObjectPhenoNode * obj,string & message) const {
		vector<int> l ;
		obj->getPositions(l);
		for (unsigned int i=0;i<l.size();i++) {
			if ( position == l[i] ) {
				message = "Conflict position with ["+name+"] and ["+obj->name+"]";
				return false ;
			}
			if ( isPresentPosition && (presentPosition == l[i])) {
				message = "Conflict presentPosition with ["+name+"] and ["+obj->name+"]";
				return false ;
			}
			if ( isCensoredPosition && (presentPosition == l[i])) {
				message = "Conflict censoredPosition with ["+name+"] and ["+obj->name+"]";
				return false ;
			}
		}
		return true;
	}
};

/**
 * Fixed effect node
 */
class  FixedEffectNode : public ObjectPhenoNode
{
public:
	int position ;  /// position in the datafile of this fixed effect
public:

	explicit FixedEffectNode()
	: ObjectPhenoNode()
	{
		position=0;
	}

	virtual ~FixedEffectNode()
	{
		cout << "~FixedEffectNode" << endl ;
	}
	virtual void print(std::ostream &os, unsigned int depth) const
	{
		os << indent(depth) << "FixedEffect [" << std::endl;
		os << indent(depth) << "name :" << name << std::endl;
		os << indent(depth) << "position :" << position << std::endl;
		os << indent(depth) << "]" << std::endl;
	}

	virtual bool check(string & message) const {
           return true;
	}

	virtual void getPositions(vector<int> & positions) const {
		positions.push_back(position);
	}

	virtual bool checkPosition(ObjectPhenoNode * obj,string & message) const {
		vector<int> l ;
		obj->getPositions(l);
		for (unsigned int i=0;i<l.size();i++) {
	      if ( position == l[i] ) {
	    	  message = "Conflict position with ["+name+"] and ["+obj->name+"]";
	    	  return false;
	      }
		}
	      return true;
    }

};

/**
 * Covariate node : similar to FixedEffectNode
 */
class  CovariateNode : public FixedEffectNode
{
public:

public:

	explicit CovariateNode()
	: FixedEffectNode()
	{
	}

	virtual ~CovariateNode()
	{
	}
	virtual void print(std::ostream &os, unsigned int depth) const
			{
		os << indent(depth) << "Covariate [" << std::endl;
		os << indent(depth) << "name :" << name << std::endl;
		os << indent(depth) << "position :" << position << std::endl;
		os << indent(depth) << "]" << std::endl;
			}
};

/**
 *
 */
class  SimulTraitNode : public ObjectPhenoNode
{
public:
	string type  ;
	double heritability ;
	ModelNode * model;

public:

	explicit SimulTraitNode()
	: ObjectPhenoNode()
	{
		name="";
		heritability=0.0;
		model=NULL;
	}

	virtual ~SimulTraitNode()
	{
	}
	virtual void print(std::ostream &os, unsigned int depth) const
	{
		os << indent(depth) << "simulTrait [" << std::endl;
		os << indent(depth) << "name :" << name << std::endl;
		os << indent(depth) << "type :" << type << std::endl;
		os << indent(depth) << "heritability :" << heritability << std::endl;
		if ( model ) {
			model->print(os,depth+1);
		}
		os << indent(depth) << "]" << std::endl;
	}

	virtual bool check(string & message) const {
			return true;
	}
	virtual void getPositions(vector<int> & positions) const {

	}

	virtual bool checkPosition(ObjectPhenoNode * obj,string & message) const {
		return true;
	}
};

class  GeneProbNode : public ObjectPhenoNode
{
public:
	int positionStart ;
	int positionEnd ;
	double heritability ;
	double undefinedValue ;
	ModelNode * model;

public:

	explicit GeneProbNode()
	: ObjectPhenoNode()
	{
		positionStart=-1;
		positionEnd =-1;
	}

	virtual ~GeneProbNode()
	{
	}
	virtual void print(std::ostream &os, unsigned int depth) const
			{
		os << indent(depth) << "trait [" << std::endl;
		os << indent(depth) << "positionStart :" << positionStart << std::endl;
		os << indent(depth) << "positionEnd :" << positionEnd << std::endl;
		os << indent(depth) << "heritability :" << heritability << std::endl;
		os << indent(depth) << "undifinedValue :" << undefinedValue << std::endl;
		if ( model ) {
			model->print (os,depth+1);
		}
		os << indent(depth) << "]" << std::endl;
	}

	virtual bool check(string & message) const {
            if (positionStart <= 0 ) {
            	message="geneProb node : bad value of positionStart (>0).";
            	return false;
            }
            if (positionStart > positionEnd ) {
                message="geneProb node : bad value of positionStart and positionEnd( positionEnd>=positionStart).";
                return false;
            }

            if (heritability<=0) {
            	message="geneProb node : heritability have to >0.";
            	return false;
            }
            if (heritability>=1) {
                message="geneProb node : heritability have to <1.";
                return false;
            }

            if ( model ) {
            	return model->check(message);
            }
			return true;
	}

	virtual void getPositions(vector<int> & positions) const {
        for (unsigned int i=positionStart;i<positionEnd;i++) {
        	positions.push_back(i);
        }
	}

	virtual bool checkPosition(ObjectPhenoNode * obj,string & message) const {
		vector<int> l ;
		obj->getPositions(l);
		for (unsigned int i=0;i<l.size();i++) {
			if ( (l[i] < positionEnd) && (l[i] > positionStart) ) {
				message="Conflict position find with geneProb and "+obj->name;
				return false;
			}
		}
		return true;
	}
};

class  PhenotypeGenNode : public QtlMapNode
{
public:
	string file ;
	bool isFile ;
	vector<FixedEffectNode *> lfe;
	vector<CovariateNode *> lcov;
	vector<ObjectPhenoNode *> ltraits;

public:

	explicit PhenotypeGenNode()
	: QtlMapNode()
	{
		isFile=false;
	}

	virtual ~PhenotypeGenNode()
	{
	}

	virtual void print(std::ostream &os, unsigned int depth) const
			{
		os << indent(depth) << "phenotype [" << std::endl;
		if ( isFile ) {
			os << indent(depth) << "file :" << file << std::endl;
		}
		for (unsigned int i=0;i<lfe.size();i++) {
			lfe[i]->print(os,depth+1);
		}
		for (unsigned int i=0;i<lcov.size();i++) {
			lcov[i]->print(os,depth+1);
		}
		for (unsigned int i=0;i<ltraits.size();i++) {
			ltraits[i]->print(os,depth+1);
		}

		os << indent(depth) << "]" << std::endl;
	}
    /**
     * 1) Check the model for each trait : the covariate and fixed effects have to be defined
     * 2) Check position (a position can be defined twice) for each Phenotype
     */
	virtual bool check(string & message) const {
			for (unsigned int i=0;i<ltraits.size();i++) {
				///checking the model
				if ( ltraits[i]->model ) {
					vector<string> l1 =  ltraits[i]->model->lfeOrCov;

                    //Each element of the model have to be referenced in lfe or lcov
					for (unsigned int j=0;j<l1.size();j++) {
                        bool find = false ;

                        for (unsigned int k=0;(k<lfe.size())&&(!find);k++) {
                        	find = (lfe[k]->name == l1[j]);
                        }
                        for (unsigned int k=0;(k<lcov.size())&&(!find);k++) {
                            find = (lcov[k]->name == l1[j]);
                        }
                        if (!find) {
                        	message = "name ["+l1[j]+"] is not defined as a fixed effect or a covariate!";
                        	return false;
                        }
					}
					l1 =  ltraits[i]->model->lfeQtl;
					//Each element of the model have to be referenced in lfe
					for (unsigned int j=0;j<l1.size();j++) {
						bool find = false ;
						for (unsigned int k=0;(k<lfe.size())&&(!find);k++) {
							find = (lfe[k]->name.compare(l1[j])==0);
						}

						if (!find) {
							message = "name ["+l1[j]+"] is not defined as a fixed effect!";
							return false;
						}
					}
				}
			}

			for (unsigned int i=0;i<ltraits.size();i++) {
				for (unsigned int j=0;j<lcov.size();j++) {
                    if ( ! ltraits[i]->checkPosition(lcov[j],message) ) {
                    	return false;
                    }
				}
				for (unsigned int j=0;j<lfe.size();j++) {
					if ( ! ltraits[i]->checkPosition(lfe[j],message) ) {
					    return false;
					}
				}
				for (unsigned int j=0;j<ltraits.size();j++) {
			        if ( i == j) continue;
					if ( ! ltraits[i]->checkPosition(ltraits[j],message) ) {
						return false;
					}
				}
			}

			for (unsigned int i=0;i<lcov.size();i++) {
				for (unsigned int j=0;j<lfe.size();j++) {
					if ( ! lcov[i]->checkPosition(lfe[j],message) ) {
					    return false;
					}
				}
				for (unsigned int j=0;j<lcov.size();j++) {
					 if ( i == j) continue;
				    if ( ! lcov[i]->checkPosition(lcov[j],message) ) {
				        return false;
				    }
				}
			}
			for (unsigned int i=0;i<lfe.size();i++) {
				for (unsigned int j=0;j<lfe.size();j++) {
					if ( i == j) continue;
					if ( ! lfe[i]->checkPosition(lfe[j],message) ) {
						return false;
					}
				}
			}

			return true;
	}
};

class  PhenotypeNode : public PhenotypeGenNode
{
public:

public:

	explicit PhenotypeNode()
	: PhenotypeGenNode()
	{
	}

	virtual ~PhenotypeNode()
	{
	}

};

class EffectNode : public QtlMapNode
{
public:
	string nameTrait ;
	double value;
public:
	explicit EffectNode()
	: QtlMapNode()
	{
		nameTrait="";
		value =0.0;
	}

	virtual ~EffectNode()
	{
	}
	virtual void print(std::ostream &os, unsigned int depth) const
	{
		os << indent(depth) << "effect [" << std::endl;
		os << indent(depth) << "nameTrait :" << nameTrait <<  std::endl;
		os << indent(depth) << "value :" << value <<  std::endl;
		os << indent(depth) << "]" << std::endl;
	}

	virtual bool check(string & message) const {
			return true;
	}
};

class QTLNode : public QtlMapNode
{
public :
	string chromosome ;
	double position   ;
	double allFreqOnGrdParent;
	vector<EffectNode*> lEffects;

public:
	explicit QTLNode()
	: QtlMapNode()
	{
		chromosome = "";
		position = 0.0;
		allFreqOnGrdParent = 0.0;
	}

	virtual ~QTLNode()
	{
		for (unsigned int i=0;i<lEffects.size();i++) {
			delete lEffects[i];
		}
		lEffects.clear();
	}
	virtual void print(std::ostream &os, unsigned int depth) const
	{
		os << indent(depth) << "QTL [" << std::endl;
		os << indent(depth) << "chromosome :" << chromosome <<  std::endl;
		os << indent(depth) << "position :" << position <<  std::endl;
		os << indent(depth) << "allFreqOnGrdParent :" << allFreqOnGrdParent <<  std::endl;
		for (unsigned int i=0;i<lEffects.size();i++) {
			lEffects[i]->print(os,depth+1);
		}
		os << indent(depth) << "]" << std::endl;
	}

	virtual bool check(string & message) const {
			return true;
	}
};

class  PhenotypeSimulNode : public PhenotypeGenNode
{
public:
	vector<vector<double>*> * correlationMatrix ;
	vector<QTLNode *> lQtls;
public:

	explicit PhenotypeSimulNode()
	: PhenotypeGenNode()
	{
	}

	virtual ~PhenotypeSimulNode()
	{
		for (unsigned int i=0;i<correlationMatrix->size();i++) {
			delete (*correlationMatrix)[i];
		}
		delete correlationMatrix;
		for (unsigned int i=0;i<lQtls.size();i++) {
			delete lQtls[i];
		}
	}

	virtual void print(std::ostream &os, unsigned int depth) const
	{
		os << indent(depth) << "phenotype [" << std::endl;
		if ( isFile ) {
			os << indent(depth) << "file :" << file << std::endl;
		}
		for (unsigned int i=0;i<lfe.size();i++) {
			lfe[i]->print(os,depth+1);
		}
		for (unsigned int i=0;i<lcov.size();i++) {
			lcov[i]->print(os,depth+1);
		}
		for (unsigned int i=0;i<ltraits.size();i++) {
			ltraits[i]->print(os,depth+1);
		}

		if ( correlationMatrix ) {
			os << indent(depth) << "correlationMatrix [" << std::endl;
			for (unsigned int i=0;i<correlationMatrix->size();i++) {
				os << indent(depth) << " [";
				for (unsigned int j=0;j<(*correlationMatrix)[i]->size();j++) {
					os << indent(depth) << (*(*correlationMatrix)[i])[j];
				}
				os << indent(depth) << "]"<< endl;
			}
			os << indent(depth) << " ]" << std::endl;
		}

		os << indent(depth) << "]" << std::endl;
		for (unsigned int i=0;i<lQtls.size();i++) {
			lQtls[i]->print(os,depth+1);
		}

	}
    /**
     * Check correlation matrix
     * 1) first elt: size:1
     *    second elt:size 2
     *    etc...
     * 2) each value is between -1 and 1
     * 3) size of matrix is nb trait - 1
     * Check Qtl effect
     * 1) qtl effect have to referenced a trait name
     */
	virtual bool check(string & message) const {
		    int count = 1 ;
		    std::ostringstream out,out2;
		    assert(correlationMatrix);
		    for (unsigned int i=0;i<correlationMatrix->size();i++) {
		        if ( (*correlationMatrix)[i]->size() != count ) {
		        	out << count;
		        	message ="Bad definition of Correlation matrix. The "+out.str()+" vector must have "+out.str()+" element(s)]";
		        	return false;
		        }
		        count++;
		    	for (unsigned int j=0;j<(*correlationMatrix)[i]->size();j++) {
		    		double val = (*(*correlationMatrix)[i])[j];
		    		if ( val < -1 ) {
		    			out << (i+1);
		    			out2 << (j+1) ;
		    			message ="Bad definition of Correlation matrix. Bad value of element ["+out.str()+","+out2.str()+"] (val<-1)";
		    			return false;
		    		}
		    		if ( val > 1 ) {
		    			out << (i+1);
		    			out2 << (j+1) ;
		    		    message ="Bad definition of Correlation matrix. Bad value of element ["+out.str()+","+out2.str()+"] (val>1)";
		    		    return false;
		    		}
		    	}
		    }

		    if ( correlationMatrix->size() !=  (ltraits.size()-1) ) {
		    	std::ostringstream out;
		    	out << ltraits.size() ;
		    	out2 << ltraits.size()-1 ;
		    	message ="Number of traits : "+out.str()+". User have to defined a correlation matrix with "+out2.str()+" vector.";
                message+="\n       ";
                for (int i=0;i<(ltraits.size()-1);i++) {
                	std::ostringstream out;
                	out << (i+1);
                	message+="trait"+out.str()+" ";
                }
                message+="\n";
                for (int i=1;i<ltraits.size();i++) {
                	std::ostringstream out;
                    out << (i+1);
                    message+="trait"+out.str()+" ";
                    for (int j=0;j<i;j++) {
                    	message+=" val ";
                    }
                    message+="\n";
                }
                message+="\n";

		    	return false;
		    }

		    //check qtleffet with traits names
		    for (unsigned int i=0;i<lQtls.size();i++) {
		    	vector<EffectNode*> leffts = lQtls[i]->lEffects;
		    	for (unsigned int j=0; j<leffts.size();j++) {
		    		bool find = false;
		    		for(unsigned k=0;(k<ltraits.size()&&(!find));k++) {
		    			find = ltraits[k]->name == leffts[j]->nameTrait ;
		    		}
                    if (!find) {
                    	message="Qtl effect definition. The trait does not exist ["+leffts[j]->nameTrait+"]";
                    	return false;
                    }
		    	}
		    }

			return PhenotypeGenNode::check(message);
	}

};

class  GenealogyNode : public GenealogyGenNode
{
public:
	string file ;

public:

	explicit GenealogyNode()
	: GenealogyGenNode()
	{
		file = "";
	}

	virtual ~GenealogyNode()
	{
	}

	virtual void print(std::ostream &os, unsigned int depth) const
	{
		os << indent(depth) << "genealogy [" << std::endl;
		os << indent(depth) << "file :" << file << std::endl;
		os << indent(depth) << "]" << std::endl;
	}

	virtual bool check(string & message) const {
			return true;
	}
};

class  GenalogySimulNode : public GenealogyGenNode
{
public:
	string type ;
	int nbSires;
	int nbDamsBySire;
	int nbProgeniesByDam;

public:

	explicit GenalogySimulNode()
	: GenealogyGenNode()
	{
		type = "";
		nbSires = 0;
		nbDamsBySire = 0;
		nbProgeniesByDam = 0;
	}

	virtual ~GenalogySimulNode()
	{
	}

	virtual void print(std::ostream &os, unsigned int depth) const
	{
		os << indent(depth) << "genealogySimul [" << std::endl;
		os << indent(depth) << "type :" << type << std::endl;
		os << indent(depth) << "nbSires :" << nbSires << std::endl;
		os << indent(depth) << "nbDamsBySire :" << nbDamsBySire << std::endl;
		os << indent(depth) << "nbProgeniesByDam :" << nbProgeniesByDam << std::endl;
		os << indent(depth) << "]" << std::endl;
	}

	virtual bool check(string & message) const {
			return true;
	}
};


class  MapNode : public QtlMapNode
{
public:
	string file ;
	string unit ;
	vector<string> chromosomes;
public:

	explicit MapNode()
	: QtlMapNode()
	{
		file = "";
		unit = "";
	}

	virtual ~MapNode()
	{

	}
	virtual void print(std::ostream &os, unsigned int depth) const
    {
		os << indent(depth) << "map ["<<  std::endl;
		os << indent(depth) << "file :" << file << std::endl;
		os << indent(depth) << "unit :" << unit << std::endl;
		string v="[";
		for (unsigned int i=0;i<chromosomes.size();i++) {
			v+=chromosomes[i]+" ";
		}
		v+="]";
		os << indent(depth) << "chromsomes :" << v << std::endl;
		os << indent(depth) << "]"<<  std::endl;
	}

	virtual bool check(string & message) const {
			return true;
	}
};

/*
 * Simulation Map Node
 */
class  ChromosomeNode : public MapNode
{
public:
	string name ;
	double density ;
	int nbAlleleByMarker;
	double size ;
	string unit ;
public:

	explicit ChromosomeNode()
	: MapNode()
	{
		name = "";
		unit = "";
		density = 0.0;
		nbAlleleByMarker = 0;
		size = 0.0;
	}

	virtual ~ChromosomeNode()
	{

	}
	virtual void print(std::ostream &os, unsigned int depth) const
	{
		os << indent(depth) << "chromosome ["<<  std::endl;
		os << indent(depth) << "name :" << name << std::endl;
		os << indent(depth) << "density :" << density << std::endl;
		os << indent(depth) << "nbAlleleByMarker :" << nbAlleleByMarker << std::endl;
		os << indent(depth) << "size :" << size << std::endl;
		os << indent(depth) << "unit :" << unit << std::endl;
		os << indent(depth) << "]"<<  std::endl;
	}

	virtual bool check(string & message) const {
			return true;
	}
};

class  GenotypeGenNode : public QtlMapNode
{
public:
	string file ;
	vector<MapNode *> lMap;
public:

	explicit GenotypeGenNode()
	: QtlMapNode()
	{
		file = "";
	}

	virtual ~GenotypeGenNode()
	{
		for (unsigned int i=0 ; i<lMap.size();i++) {
			delete lMap[i];
		}
		lMap.clear();
	}

	virtual bool check(string & message) const {

		if ( lMap.size() < 1 ) {
			message = "none map is defined.";
			return false;
		}
		for (unsigned int i=0;i<lMap.size();i++ ){
			if (! lMap[i]->check(message)) return false;
		}
		return true;
	}
};

class  GenotypeNode : public GenotypeGenNode
{
public:
	string unknownChar;

public:

	explicit GenotypeNode()
	: GenotypeGenNode()
	{
		unknownChar="";
	}

	virtual ~GenotypeNode()
	{

	}

	virtual void print(std::ostream &os, unsigned int depth) const
	{
		os << indent(depth) << "genotype ["<<  std::endl;
		os << indent(depth) << "file :" << file << std::endl;
		os << indent(depth) << "unknownChar :" << unknownChar << std::endl;
		os << indent(depth) << "]"<<  std::endl;
		for (unsigned int i=0;i<lMap.size();i++ ){
			lMap[i]->print(os,depth+1);
		}
	}

	virtual bool check(string & message) const {
			if ( unknownChar == "" ) {
				message = "unknown char is empty.";
				return false;
			}
			return GenotypeGenNode::check(message);
			return true;
	}

};

class  GenotypeSimulNode : public GenotypeGenNode
{
public:
	bool isFile ;


public:

	explicit GenotypeSimulNode()
	: GenotypeGenNode()
	{
		isFile=false;
	}

	virtual ~GenotypeSimulNode()
	{

	}

	virtual void print(std::ostream &os, unsigned int depth) const
			{
		os << indent(depth) << "genotypeSimul ["<<  std::endl;
		if ( isFile ) {
			os << indent(depth) << "file :" << file << std::endl;
		}
		for (unsigned int i=0;i<lMap.size();i++ ){
			lMap[i]->print(os,depth+1);
		}
		os << indent(depth) << "]"<<  std::endl;
			}

	virtual bool check(string & message) const {
		    if ( isFile ) {
		    	if ( file == "" ) {
		    		message = "genotype file is empty!";
		    		return false;
		    	}
		    }
			return GenotypeGenNode::check(message);
		}

};

/**
 * General generic object to describe a simulation or a analyse
 */
class  QtlMapGenericObject : public QtlMapNode {
public:
	string outputFile ;                /// output file result in ascii format
	string outputSummary ;               /// summary file of the analyse or simulation
	GenotypeGenNode * genotypeGen;     /// genotype generic data
	GenealogyGenNode * genealogyGen;   /// genealogy generic data
	PhenotypeGenNode * phenotypeGen;   /// phenotype generic data

public :
	/// construct a constant calculation node from a value
	explicit QtlMapGenericObject()
	: QtlMapNode()
	{
		outputFile="";
		outputSummary="";
		genotypeGen=NULL;
		genealogyGen=NULL;
		phenotypeGen=NULL;
	}

	virtual ~QtlMapGenericObject()
	{
		if ( genotypeGen ) delete genotypeGen;
	}

	/**
	 * Checking
	 */
	virtual bool check(string & message) const {
           if (outputFile == "") {
        	   message = "outputFile definition is empty";
        	   return false;
           }
           if (outputSummary == "") {
               message = "outputSummary definition is empty";
               return false;
           }

           if ( ! genotypeGen ) {
        	   message = "Genotype definition is empty";
        	   return false;
           }

           if ( ! genotypeGen->check(message)) return false;

           if ( ! genealogyGen ) {
               message = "Genealogy definition is empty";
               return false;
           }

           if ( ! genealogyGen->check(message)) return false;

           if ( ! phenotypeGen ) {
               message = "Phenotype definition is empty";
               return false;
           }

           if ( ! phenotypeGen->check(message)) return false;

           return true;
	}
};

class  SimulationNode : public QtlMapGenericObject {

public:
	string outputSimul;
public :
	/// construct a constant calculation node from a value
	explicit SimulationNode()
	: QtlMapGenericObject()
	{
		outputSimul="";
	}

	virtual ~SimulationNode()
	{

	}
	virtual void print(std::ostream &os, unsigned int depth) const
			{
		os << indent(depth) << "Output file :" << outputFile << std::endl;
		os << indent(depth) << "Summary file :" << outputSummary << std::endl;
		os << indent(depth) << "Simulation file :" << outputSimul << std::endl;
		if ( genotypeGen ) {
			genotypeGen->print(os,depth+1);
		}
		if ( genealogyGen ) {
			genealogyGen->print(os,depth+1);
		}
		if ( phenotypeGen ) {
			phenotypeGen->print(os,depth+1);
		}
			}
	/**
	 * Check the chromosome referenced in QTL effect have to be defined in the Genotype definition
	 */
	virtual bool check(string & message) const {

		if ( ! genotypeGen ) {
			message ="Genotype node is not defined.";
			return false;
		}
		if ( ! phenotypeGen ) {
			message ="Phenotype node is not defined.";
			return false;
		}

		PhenotypeSimulNode * psimul = dynamic_cast<PhenotypeSimulNode*>(phenotypeGen);

		if ( psimul) {
			vector<MapNode *> lm = genotypeGen->lMap;
			vector<string> lc ;
			lc.clear();

			for (unsigned int i=0;i<lm.size();i++) {
				ChromosomeNode * c = dynamic_cast<ChromosomeNode*>(lm[i]);
				if ( ! c ) {
					//not a simulation map....
					for (unsigned int j=0;j<lm[i]->chromosomes.size();j++) {
						lc.push_back(lm[i]->chromosomes[j]);
					}
				} else {
					lc.push_back(c->name);
				}
			}

			vector<QTLNode *> lqtls = psimul->lQtls;

			for (unsigned int i=0;i<lqtls.size();i++) {
				bool find = false;
				for (unsigned int j=0;(j<lc.size()&&!find);j++) {
					find = (lc[j] == lqtls[i]->chromosome);
				}
				if (!find) {
					message="Qtl effect definition. Chromosome ["+lqtls[i]->chromosome+"] is not defined.";
					return false;
				}
			}
		}

		return QtlMapGenericObject::check(message);
	}
};

class  AnalyseNode : public QtlMapGenericObject
{
public :
	string xmlFile ;
	string pdedFile ;
	string pdedJoinFile ;
	string haplotypeFile ;
	string lrtFile;
	string qtlEffectFile;


public:
	/// construct a constant calculation node from a value
	explicit AnalyseNode()
	: QtlMapGenericObject()
	{
		xmlFile = "";
		pdedFile = "";
		pdedJoinFile = "";
		haplotypeFile = "";
		lrtFile = "";
		qtlEffectFile = "";

	}

	virtual ~AnalyseNode()
	{
		this->print(cout,2);
		//delete node;
	}

	virtual void print(std::ostream &os, unsigned int depth) const
			{
		os << indent(depth) << "Output file :" << outputFile << std::endl;
		os << indent(depth) << "Summary file :" << outputSummary << std::endl;
		os << indent(depth) << "Xml file :" << xmlFile << std::endl;
		os << indent(depth) << "pdedJoin file :" << pdedJoinFile << std::endl;
		os << indent(depth) << "pded file :" << pdedFile << std::endl;
		os << indent(depth) << "haplotype file :" << haplotypeFile << std::endl;
		os << indent(depth) << "LRT file :" << lrtFile << std::endl;
		os << indent(depth) << "QTLEffect file :" << qtlEffectFile << std::endl;
		if ( genotypeGen ) {
			genotypeGen->print(os,depth+1);
		}
		if ( genealogyGen ) {
			genealogyGen->print(os,depth+1);
		}
		if ( phenotypeGen ) {
			phenotypeGen->print(os,depth+1);
		}
			}

	 virtual bool check(string & message) const {
			return QtlMapGenericObject::check(message);
	 }
};

/**
 * Intermediary Class to instance node grammar
 */

class TraitsAndFileInformation {
public:
	string file ;
	bool isFile ;
	vector<ObjectPhenoNode*>* lfep ;
	vector<ObjectPhenoNode*>* lcovp ;
	vector<ObjectPhenoNode*>* ltraitsp ;
};

class FileAndVectorMapNode {
public:
	string file ;
	bool isFile ;
	vector<MapNode*>* lmaps ;
};


/** Calculator context used to save the parsed expressions. This context is
 * passed along to the example::Driver class and fill during parsing via bison
 * actions. */
class QtlMapContext
{
public:
	/// array of unassigned expressions found by the parser. these are then
	/// outputted to the user.
	std::map<std::string,QtlMapNode*>	nodes;

	/// free the saved expression trees
	~QtlMapContext()
	{
		clear();
	}

	/// free all saved expression trees
	void clear()
	{
		/* for(map<string,QtlMapNode*>::iterator it=nodes.begin(); it != nodes.end(); ++it)
	  {
	    delete (*it).second;
	  }*/
		nodes.clear();
	}

	void add(const string & id,QtlMapNode* obj) {
		map<string,QtlMapNode*>::iterator it=nodes.find(id) ;
		if ( it == nodes.end() ) {
			nodes.insert(pair<string,QtlMapNode*>(id,obj));
		}
	}

	bool remove(const string & id) {
		map<string,QtlMapNode*>::iterator it=nodes.find(id) ;
		if ( it != nodes.end() ) {
			//delete (*it).second;
			nodes.erase(it);
		}
	}

	QtlMapNode* get(const string & id) {
		map<string,QtlMapNode*>::iterator it=nodes.find(id) ;
		if ( it != nodes.end() ) return (*it).second;
		cout << "NULL " << endl ;
		return NULL;
	}
};

#endif // EXPRESSION_H
