!Copyright or © or Copr. INRA (2010)
!contributors :
!
!  Pascale Le Roy     <Pascale.LeRoy@rennes.inra.fr>
!  Jean-Michel Elsen  <Jean-Michel.Elsen@toulouse.inra.fr>
!  Helene Gilbert     <Helene.Gilbert@jouy.inra.fr>
!  Carole Moreno      <Carole.Moreno@toulouse.inra.fr>
!  Andres Legarra     <Andres.Legarra@toulouse.inra.fr>
!  Olivier Filangi    <Olivier.Filangi@rennes.inra.fr>
!
!
!QTLMap is a computer program dedicated to the detection
!of QTL from experimental designs in outbred population.
!The statistical techniques used are linkage analysis (LA) and
!linkage disequilibrium linkage analysis (LDLA) using interval mapping.
!Different versions of the LA are proposed from a  quasi Maximum Likelihood
! approach to a fully linear (regression) model.
!The LDLA is a regression approach (Legarra and Fernando, 2009).
!The population may be sets of half-sib families or mixture of
!full- and half- sib families.
!The computations of Phase and Transmission probabilities are optimized to be rapid and as exact as possible.
!QTLMap is able to deal with large numbers of markers (SNP) and traits (eQTL).

!This software is governed by the CeCILL license under French law and
!abiding by the rules of distribution of free software.  You can  use,
!modify and/ or redistribute the software under the terms of the CeCILL
!license as circulated by CEA, CNRS and INRIA at the following URL
!"http://www.cecill.info".
!
!As a counterpart to the access to the source code and  rights to copy,
!modify and redistribute granted by the license, users are provided only
!with a limited warranty  and the software's author,  the holder of the
!economic rights,  and the successive licensors  have only  limited
!liability.
!
!In this respect, the user's attention is drawn to the risks associated
!with loading,  using,  modifying and/or developing or reproducing the
!software by the user in light of its specific status of free software,
!that may mean  that it is complicated to manipulate,  and  that  also
!therefore means  that it is reserved for developers  and  experienced
!professionals having in-depth computer knowledge. Users are therefore
!encouraged to load and test the software's suitability as regards their
!requirements in conditions enabling the security of their systems and/or
!data to be ensured and,  more generally, to use and operate it in the
!same conditions as regards security.
!
!The fact that you are presently reading this means that you have had
!knowledge of the CeCILL license and that you accept its terms.


!!****J* QTLMap/qtlmap
!!  NAME
!!    QTLMap
!!  DESCRIPTION
!!
!!   QTLMap is a software dedicated to the detection of QTL from experimental designs in outbred population.
!!   QTLMap software is developed by the Animal Genetics Division at INRA (French National Institute for Agronomical Research).
!!   The statistical techniques used are linkage analysis (LA) and linkage disequilibrium linkage analysis (LDLA) using interval mapping.
!!   Different versions of the LA are proposed from a  quasi Maximum Likelihood approach to a fully linear (regression) model.
!!   The LDLA is a regression approach (Legarra and Fernando, 2009). The population may be sets of half-sib families or mixture of full- and half- sib families.
!!   The computations of Phase and Transmission probabilities are optimized to be rapid and as exact as possible.
!!   QTLMap is able to deal with large numbers of markers (SNP) and traits (eQTL).
!!
!!   The aim of QTLMap developers is to propose various genetic models depending on:
!!
!!     * the number of QTL alleles segregating (biallelic in crosses between monomorphic breeds, biallelic without hypothesis on the origin, multiallelic, haplotype identity),
!!     * the number of QTL segregating (one, two linked, several unlinked),
!!     * the number of traits under the QTL influence. The trait determinism may vary depending on :
!!
!!         - the trait distribution (gaussian trait, survival trait or threshold distribution),
!!         - the interactions between the QTL and fixed effects or other loci,
!!         - the residual variance structure (homo- or heteroskedasticity for half-sib families).
!!
!!   Due to differences with the asymptotical conditions from the chi2 theory, the test statistic significance are evaluated either through numerical approximations,
!!   or through empirical calculations obtained from permutations or simulations under the null hypothesis.
!!
!!   QTLmap is written in fortran 95 and use the OpenMP API (Parallel Programming).
!!
!!   Up to now, the following functionnalities have been implemented :
!!
!!    * QTL detection in half-sib families or mixture of full- and half-sib families
!!    * One or several linked QTL segregating in the population
!!    * Single trait or multiple trait
!!    * Nuisance parameters (e.g. sex, batch, weight...) and their interactions with QTL can be included in the analysis
!!    * Gaussian, discrete or survival (Cox model) data
!!    * Familial heterogeneity of variances (heteroscedasticity)
!!    * Can handle eQTL analyses
!!    * Computation of transmission and phase probabilities adapted to high throughput genotyping (SNP)
!!    * Empirical thresholds are estimated using simulations under the null hypothesis or permutations of trait values
!!    * Computation of power and accuracy of your design or any simulated design
!!
!!  BUGS
!!
!!  SEE ALSO
!!
!!    http://www.inra.fr/qtlmap
!!
!!  CREATION DATE
!!  28/08/2009
!!
!!  COPYRIGHT
!!***

program qtlmap
#ifdef MPI_ACTIVE
      use mpi
#endif
#ifdef MANAGE_OMP
       use omp_lib
#endif
      use m_qtlmap_log
      use m_qtlmap_constant
      use m_qtlmap_data
      use m_qtlmap_haplotype_data
      use m_qtlmap_tools
      use m_qtlmap_haplotype
      use m_qtlmap_analyse
      use m_qtlmap_parameter_file
      use m_qtlmap_genealogy
      use m_qtlmap_genotype
      use m_qtlmap_traits
      use m_qtlmap_map
      use m_qtlmap_simulation
      use m_qtlmap_output_handler
      use m_qtlmap_cli
      use m_qtlmap_optimization
      use m_qtlmap_math
      use m_qtlmap_output_xml

      implicit none

! Divers
      character(len=LENGTH_MAX_FILE) :: param
      integer                        :: step
      real (kind=dp), dimension (:),allocatable :: xmutanalyse,sigtanalyse
      integer :: nsim,thread_rang = 0 ! by default / identifiant du processus fils mpi
#ifdef MPI_ACTIVE
        integer :: code          ! identifiant du processus global mpi
        integer :: nb_procs      ! nombre de processus alloue par l utilisateur
        integer :: nb_slaves=1   ! nombre de processus alloue
        logical :: drapeau=.true.
        logical :: logique
        integer :: intra_comm,inter_comm
        integer(kind = MPI_ADDRESS_KIND) :: nb_procs_total

        call MPI_INIT (code)
        call MPI_COMM_SIZE ( MPI_COMM_WORLD , nb_procs , code)
        call MPI_COMM_RANK (MPI_COMM_WORLD, thread_rang , code)
        if ( thread_rang == 0 ) then
          call log_mess(" ",INFO_DEF)
          call log_mess("   ****************************** ",INFO_DEF)
          call log_mess("   ****           MPI         *** ",INFO_DEF)
          call log_mess("   NB_PROC:        :"//trim(strint(nb_procs)),INFO_DEF )
          call log_mess("   ****************************** ",INFO_DEF)
          call log_mess(" ",INFO_DEF)
        end if
#endif
#ifdef MANAGE_OMP
     !parallelisation sur le parcours du groupe de liaison
     call omp_set_nested(.true.)

     call log_mess(" ",INFO_DEF)
     call log_mess("   ****************************** ",INFO_DEF)
     call log_mess("   **** Use OpenMP directives *** ",INFO_DEF)
     call log_mess("   OMP_NUM_THREADS   :"//trim(strint(omp_get_max_threads())),INFO_DEF )
     call log_mess("   TO CHANGE  EXPORT OMP_NUM_THREADS=<NB> or SETENV OMP_NUM_THREADS <NB>",INFO_DEF )
     !call log_mess("   OMP_NESTED        :"//trim(strbool(omp_get_nested())),INFO_DEF )
     call log_mess("   ****************************** ",INFO_DEF)
     call log_mess(" ",INFO_DEF)
#endif
      ! Printing Version
      call cli_print_qtlmap_version

      !----- PRINT HELP IF OPTION --HELP DETECTED
      if (cli_is_help_panalyse()) then
         call help_panalyse
         stop
      end if

       if (cli_is_help_paramsim()) then
         call help_paramsim
         stop
      end if

      if (cli_is_help()) then
          call cli_print_help
          stop
      end if

      ! ----- INIT LOG LEVEL
      call init_log(INFO_DEF)

      if (cli_is_quiet()) call init_log(ERROR_DEF)
      if (cli_is_verbose()) call init_log(VERBOSE_DEF)
      if (cli_is_debug()) call init_log(DEBUG_DEF)

     !Getting the optimization implementation for minimizing likelihood
      call set_optimization(cli_get_optimization())
      ! ------ GETTING PARAMETER FILE
      param = cli_get_parameters_file()
       !----- Give information about library used by the program
      call MATH_QTLMAP_INFO

      ! ------ READING PARAMETER FILE
      call read_analyse_file(param)
      ! --> PARAMETER PART OF QTLMAP_DATA IS INITIALIZED
      call initialize_qtlmap_context(cli_is_simulation() .or. cli_is_nsim(),cli_is_permute())
      ! ------ PRINTING ANALAYSE AND HAPLOTYPE VERSION DETECTED
      call log_mess('opt_calcul:'//trim(str(cli_get_analyse())),INFO_DEF)
      call log_mess('opt_version:'//trim(str(cli_get_haplotype())),INFO_DEF)
      opt_sib = cli_get_family()
      call log_mess('opt_sib:'//trim(str(opt_sib)),INFO_DEF)

      step     = pas

      !------- IF SIMULATION OPTION OR NSIM OPTION IS DETECTED --> SIMULATION
     !print *, 'SIMUL', cli_is_simulation(),cli_is_nsim() 
      if(cli_is_simulation() .or. cli_is_nsim())then
         nsim = cli_get_nsim()
#ifdef MPI_ACTIVE
         nsim = nsim - mod(nsim,nb_procs)
#endif
         ! ----RUNNING SIMULATION
         call qtlmap_simul(.false.,cli_is_transcriptomic_data(),nsim,cli_get_analyse(),&
                 cli_get_nqtl(),cli_is_hdam())
      else
         ! ----RUNNING ANALYZE
         call qtlmap_analyze(cli_is_transcriptomic_data(),cli_get_analyse(),cli_get_nqtl(),cli_is_hdam(),ndmin)
      end if

      call end_analyse_parameterfile
      call release_data
      call release_haplotype_data
#ifdef MPI_ACTIVE
      print *,"PROC ["//trim(str(thread_rang))//"] END ** "
      call MPI_Finalize (code);
#endif
      contains

!***********************************  SUBROUTINE *********************************************************


      subroutine qtlmap_analyze(type_data_transcr,opt_calcul,opt_qtl,hdam,ndmin)
        logical,         intent(in)               :: type_data_transcr,hdam
        integer         , intent(in)              :: opt_calcul,opt_qtl,ndmin

        type(TYPE_LRT_SOLUTION)  , dimension(:,:),allocatable       :: lrtsol
        type(TYPE_INCIDENCE_SOLUTION) ,dimension(:,:), allocatable  :: listincsol

        real (kind=dp), dimension (:),allocatable   :: ih2
        real (kind=dp),dimension (:,:,:),allocatable :: rhoi
        integer, dimension(:),pointer :: filter  => NULL()

        real(kind=dp) :: hdeux
        integer ,dimension(10) :: position_qtl
        real(kind=dp) ,dimension(10) ::effet_qtl

        integer                  :: c,stat,opt_sim,jm,igeno,nsim,i,j,nnnqtl,sizeList,nb_thread
        character(len=LEN_DEF)   :: simula_t,cc
        character(len=LEN_L) ,dimension(NUMBER_ALL_KEYS) :: index_key,values
        character(len=LENGTH_MAX_FILE) :: genin,perfin,perfout
        character(len=300) :: name_funct

        if (type_data_transcr .and. .not. cli_is_print_all() ) then
           opt_sim = TRANSCIPTOME_ANALYSE
        else
           opt_sim = COMMON_ANALYSE
        end if

        name_funct = get_name_analyse(opt_calcul)
#ifdef MANAGE_OMP
        nb_thread = omp_get_max_threads()
#else
        nb_thread = 1
#endif
        call init_output_handler(opt_qtl,opt_calcul,name_funct,nb_thread)

        call get_summary_panalyse(index_key,values,c)
        call print_summary_panalyse(c,index_key,values)

        !******** Lecture de la map et de la genealogie en parallele ****

        !$OMP PARALLEL DEFAULT(SHARED)
        !$OMP SECTIONS
        !$OMP SECTION
        call read_map
        !$OMP SECTION
        call read_genealogy(cli_is_calcul_cd())
        !$OMP END SECTIONS NOWAIT
        !$OMP END PARALLEL

        !******** Lecture des genotypes et de la genealogie en parallele ****

        !$OMP PARALLEL DEFAULT(SHARED)
        !$OMP SECTIONS
        !$OMP SECTION
        call read_genotype(ndmin)
        !$OMP SECTION
        call read_model(in_param_ef,filter)
        call fixe_structure_model(filter)
        call read_traits(type_data_transcr,filter,cli_is_calcul_cd())
        !$OMP END SECTIONS NOWAIT
        !$OMP END PARALLEL

        if (associated(filter)) deallocate (filter)

        call recup
        call check_typage
        call log_descriptif_genealogy(in_genea)
        call log_marker_description
        call print_freqall
        call check_HWE

        if ( cli_is_calcul_cd() ) then
          call calcul_y_cd(datasetUser)
          call calcul_corcd(datasetUser)
        end if

        ! Normalizing data and print information about performance trait
        call manage_data(type_data_transcr,.false.,need_normalize_data(opt_calcul))

        if (.not. type_data_transcr) then
           call log_descriptif_traits()
        end if


        allocate(lrtsol(size(carac),opt_qtl))
        allocate (listincsol(size(carac),opt_qtl+1))
        allocate (rhoi(nchr,size(carac),size(carac)))
        allocate (ih2(size(carac)))

        !computing phase,haplotype
        call haplotype(cli_get_haplotype())
        do c=1,nchr
         do jm=1,nm
           call log_mess('Mother :'//trim(mere(jm))//'-->genotypes found [Chr '//trim(chromo(c))//']:'//&
       trim(str(ngenom(c,jm+1)-ngenom(c,jm))),VERBOSE_DEF)
           do igeno=ngenom(c,jm)+1,ngenom(c,jm+1)
                call log_mess('genotype '//trim(str((igeno)))//' : '//trim(str(probg(c,igeno)*100.d0))//'%',VERBOSE_DEF)
           end do
          end do
        end do


        !$OMP PARALLEL DEFAULT(SHARED)
        !$OMP SECTIONS
        !$OMP SECTION
        do c=1,nchr
          ! Printing probabilities of transmission in associated file
          call print_pded(c)
        end do
        !$OMP SECTION
        ! Printing genotype finded for sire
        call print_sire_phase()
        ! Printing all genotyp available for dam
        call print_dam_phase()
        !$OMP END SECTIONS NOWAIT
        !$OMP END PARALLEL

        call log_mess("-- ANALYSIS --  CALCUL="//get_name_analyse(opt_calcul),INFO_DEF)
        !analysis
        call analyse(opt_calcul,opt_qtl,lrtsol,listincsol,rhoi,hdam,opt_sim)


        if (.not. type_data_transcr) then
           if (.not. is_multitrait_analysis(opt_calcul)) then
             !print to the console max LRT
              call print_information_max_lrt(opt_qtl,INFO_DEF,lrtsol)
          else
             !print to the console max LRT
              call print_information_max_lrt_multi(opt_qtl,INFO_DEF,lrtsol(1,:))
          end if
        end if

        if (cli_is_xmloutput()) then
          call print_pded_xml
          call print_solution_xml(opt_qtl,lrtsol,listincsol,is_multitrait_analysis(opt_calcul))
        end if
        !---------------------------------------------------------------------------
        !  Estimate the rejection thresholds
        !  NOTE OFI : On ne gere pas 2 QTL contre 1 pour l instant pour les raisons suivantes :
        !          1)   -- difficile de recuperer l effet moyen qtl selon l analyse
        !            ** exemple modfil : un effet moyen general est estime (parfois pas d ailleur.....
        !             faut il continuer l estimation de seuil si celui ci n est pas estime ?)
        !            ** exemple analyse classique : seul les effets peres sont accesible : on prend la moyenne general ?
        !            si oui, dans ce cas on va retrouver des resultats differents selon si on lance avec modlin ou le cas classique
        !          2) Ne faut il pas plutot associe un effet qtl par pere et non un effet moyen, la simul est fausse
        !             --> actuellement le parsing de parsim ne permet pas de faire ca ......
        !
        !
        if ( cli_is_estimate_thresholds(nsim) .and. opt_qtl == 2 ) then
           call log_mess('Estimate the rejection threshold with 2 QTL are not available.',WARNING_DEF)
        else
          if ( cli_is_estimate_thresholds(nsim) ) then
             simula='simulation_thresholds'
             in_parsim='param_sim_threholds'
             ih2=0.3d0
             allocate (xmutanalyse(ncar))
             allocate (sigtanalyse(ncar))
             xmutanalyse = xmut
             sigtanalyse = sigt
             ! Check only first chromosome
             call create_simulation_file(in_parsim,rhoi(1,:,:),ih2)
             call qtlmap_simul(.true.,&
                    cli_is_transcriptomic_data(),nsim,cli_get_analyse(),cli_get_nqtl(),cli_is_hdam())

             deallocate (xmutanalyse)
             deallocate (sigtanalyse)

          end if
        end if

       do i=1,size(lrtsol,1)
          do j=1,size(lrtsol,2)
            call release(lrtsol(i,j))
          end do
        end do

         do i=1,size(listincsol,1)
          do j=1,size(listincsol,2)
            call release(listincsol(i,j))
          end do
        end do


        deallocate(lrtsol)
        deallocate(rhoi)
        deallocate (ih2)


        call end_output_handler(opt_qtl)

      end subroutine qtlmap_analyze



       subroutine qtlmap_simul(isThresholdEstim,type_data_transcr,&
                               nsim,opt_calcul,opt_qtl,hdam)
        logical,         intent(in)               :: isThresholdEstim,type_data_transcr,hdam
        integer         , intent(in)              :: nsim,opt_calcul,opt_qtl
        type(TYPE_LRT_SOLUTION)  , dimension(:,:,:),allocatable      :: lrtsol
        type(TYPE_INCIDENCE_SOLUTION) ,dimension(:,:), allocatable   :: listincsol
        real (kind=dp),dimension (:,:,:),allocatable :: rhoi

        integer      ,dimension(:,:,:)  ,allocatable         :: qtl !(np+nfem+nd,nqtl,2)

        integer                        :: iseed1,iseed2,stat,i,ic,isim,opt_sim,opt_version,j,iqtl
        integer, dimension(8)          :: value
        character(len=100)             :: phrase
        character(len=8)               :: date
        character(len=10)              :: time
        character(len=5)               :: zone

        logical                        :: simQtl,simulMap,permut_active
        character(len=LEN_DEF)           :: tempfile
        character(len=LEN_BUFFER_WORD) :: croisement

        integer                        :: nbAff,nAff,o,nqtlsimul,startSim,endSim,iq,s1,s2,sp1,sp2,nb_thread
        real                           :: pasAff
        integer, dimension(:),pointer          :: filter => NULL()
#ifdef MPI_ACTIVE
        integer                                  :: code              ! code erreur pour les routine MPI
        integer                                  :: nbsimulByThread   ! nombre de simulation par thread
        integer                                  :: dim3double        ! identifiant de type MPI pour le vecteur de double passer par message
        integer                                  :: dim3integer       ! identifiant de type MPI pour le vecteur d'entier passer par message
        integer                                  :: deb_indice
        logical           ,allocatable ,dimension(:,:,:,:) :: unpackAll ! utiliser pour decompacter un vecteur passer en message MPI

        double precision  ,allocatable ,dimension(:,:,:,:) :: lrtmaxTab
        double precision  ,allocatable ,dimension(:)       :: lrtmaxTab_lin   ! compactage des valeurs du LRT pour chaque processus
        double precision  ,allocatable ,dimension(:)       :: lrtmaxTab_final ! agregation de l ensemble des resultats et gerer par le processus maitre

        integer           ,allocatable ,dimension(:,:,:,:) :: chrmaxTab
        integer  ,allocatable ,dimension(:)                :: chrmaxTab_lin   ! compactage des valeurs du LRT pour chaque processus
        integer  ,allocatable ,dimension(:)                :: chrmaxTab_final ! agregation de l ensemble des resultats et gerer par le processus maitre

        integer           ,allocatable ,dimension(:,:,:,:) :: nxmaxTab
        integer  ,allocatable ,dimension(:)                :: nxmaxTab_lin   ! compactage des valeurs du LRT pour chaque processus
        integer  ,allocatable ,dimension(:)                :: nxmaxTab_final ! agregation de l ensemble des resultats et gerer par le processus maitre
!        integer,dimension(1)                     :: profil_tab,profil_sous_tab,coord_debut
!        integer,dimension( MPI_STATUS_SIZE )     :: statut


#endif
        if ( cli_is_print_all() ) then
           opt_sim = COMMON_ANALYSE
        else
           opt_sim = SIMULATION
        end if
        opt_version = cli_get_haplotype()

        permut_active = ( (  get_type_data_analyse(opt_calcul) == TYPE_DATA_COX ) .or. cli_is_permute() )

#ifdef MANAGE_OMP
        nb_thread = omp_get_max_threads()
#else
        nb_thread = 1
#endif

        if (.not. isThresholdEstim ) call init_output_handler(opt_qtl,opt_calcul,get_name_analyse(opt_calcul),nb_thread)

        ! Initialize Random number Generator CoMmon (for Randlib)
        call inrgcm()
        ! Init the random generator
        call DATE_AND_TIME(date, time, zone, value)
        phrase=date//time//zone
        call phrtsd(phrase,iseed1,iseed2)
        call setall(iseed1,iseed2)
 
        if ( .not. permut_active ) then
          call log_mess("*******************************",INFO_DEF)
          call log_mess("******* SIMULATION TRAITS *****",INFO_DEF)
          call log_mess("*******************************",INFO_DEF)
          call read_simulation_file(type_data_transcr,simQtl,nqtlsimul,simulMap,ndmin,croisement)
        else
          call log_mess("*******************************",INFO_DEF)
          call log_mess("****** PERMUTATION TRAITS *****",INFO_DEF)
          call log_mess("*******************************",INFO_DEF)
          simQtl=.false.
          simulMap=.false.
!INITIALISATION DE NQTLSIMUL car sinon il etait egal à 14000 (CAROLE)
          nqtlsimul=0
          call read_genealogy
          call read_model(in_param_ef,filter)
          call fixe_structure_model(filter)
          call read_traits(type_data_transcr,filter)

          if (associated(filter)) deallocate (filter)
        end if
        if (.not. isThresholdEstim ) then

          if (simulGenea .and. permut_active ) then
             call stop_application("Permutation only are available for threshold estimation.")
          end if
          if (simulMap) then
            call init_simul_marker()
            call sim_carte(1,dens,nalle)
            call set_absi
            call init_genotype_simul()
            call sim_typ(ndmin)
          else
            call read_map
            call read_genotype(ndmin)
          end if
        end if

        call log_infoqtldefinedbyuser(nqtlsimul)

        allocate (qtl(np+nm+nd,nqtlsimul,2),STAT=stat)
        call check_allocate(stat,'qtl')
 
        allocate(lrtsol(nsim,size(carac),opt_qtl))
        allocate(listincsol(size(carac),opt_qtl+1))

        if ( isThresholdEstim ) then
          xmut = xmutanalyse
          sigt = sigtanalyse
        end if

        allocate (rhoi(nchr,size(carac),size(carac)))

        if (.not. isThresholdEstim .and. .not. simulMap) then
           call recup
           call check_typage
           call haplotype(opt_version)
           !call manage_simulator_traits(is_multitrait_analysis(opt_calcul),permut_active,ndmin,rho,h2)
           !call manage_data(type_data_transcr,.true.)
           !call log_simulation_message(simulMap,h2,ue,dens,nalle,taille)
        end if

        ! ON ENLEVE LES MESSAGES POUR LES SIMUL....
        call init_log(ERROR_DEF)

        !**************************************
        ! MAIN LOPP
        !**************************************
#ifdef MPI_ACTIVE

       ! on divise l ensemble des simulation en bloc
       ! chaque processus a a sa charge NSIM / NB_PROCESSUS simulation
       nbsimulByThread = nsim / (nb_procs)

       ! Ce qui revient pour nos structure de donnees a faire une boucle partielle pour chaque processus
       startSim = 1!nbsimulByThread * (thread_rang)+1
       endSim   = nbsimulByThread !* (thread_rang+1)

       allocate (lrtmaxTab_final(nsim*ncar*opt_qtl*opt_qtl))
       allocate (chrmaxTab_final(nsim*ncar*opt_qtl*opt_qtl))
       allocate (nxmaxTab_final(nsim*ncar*opt_qtl*opt_qtl))
       allocate (unpackAll(nbsimulByThread,ncar,opt_qtl,opt_qtl))
       unpackAll = .true.

       allocate (lrtmaxTab(nbsimulByThread,ncar,opt_qtl,opt_qtl))
       allocate (chrmaxTab(nbsimulByThread,ncar,opt_qtl,opt_qtl))
       allocate (nxmaxTab(nbsimulByThread,ncar,opt_qtl,opt_qtl))

       allocate (lrtmaxTab_lin(nbsimulByThread*ncar*opt_qtl*opt_qtl))
       allocate (chrmaxTab_lin(nbsimulByThread*ncar*opt_qtl*opt_qtl))
       allocate (nxmaxTab_lin(nbsimulByThread*ncar*opt_qtl*opt_qtl))

       lrtmaxTab=0.d0
       chrmaxTab=0
       nxmaxTab=0

#else
       startSim = 1
       endSim   = nsim
        if (thread_rang == 0  ) then
          nbAff=20
          if ( endSim < nbAff ) nbAff = endSim
          pasAff = real(endSim) / real(nbAff)
          nAff = 1
          write (*,fmt='(a3,'//trim(str(nbAff))//'(a4),a5)') '0%[',(' ',i=1,nbAff),']100%'
          write (*,fmt='(a3)',advance='no') ' '
        end if

#endif


        do isim=startSim,endSim
        !print *,"isim:",isim
#ifndef MPI_ACTIVE
         if ( isim > int(real(nAff)*pasAff) .and. (isim < 10000) ) then
            !write (*,fmt='(a1)',advance='no') '='
            write (*,fmt='(i4)',advance='no') isim
            nAff = nAff + 1
         end if
#endif
          if (simulMap) then
           ! call sim_carte(1,dens,nalle)
            if (simulGenea) call sim_genea(inmp,indm,croisement)
            call sim_typ(ndmin)
            call recup
            call check_typage
           ! call manage_data(type_data_transcr,.true.)
            call haplotype(opt_version)
          end if

          call manage_simulator_traits(is_multitrait_analysis(opt_calcul),permut_active,step,&
          simqtl,nqtlsimul,qtl,croisement)
          call manage_data(type_data_transcr,.true.,need_normalize_data(opt_calcul))

          ! Analyse according to the option opt_calcul
          call analyse(opt_calcul,opt_qtl, lrtsol(isim,:,:),listincsol(:,:),rhoi,hdam,opt_sim)

          ! release incsol
          do i=1,size(listincsol,1)
           do j=1,size(listincsol,2)
            call release(listincsol(i,j))
           end do
          end do

          do i=1,size(lrtsol,2)
           do j=1,size(lrtsol,3)
            call clean_lrt_solution(lrtsol(isim,i,j))
           end do
          end do

#ifdef MPI_ACTIVE
          do ic=1,ncar
           do iqtl=1,opt_qtl
            do iq=1,iqtl
             lrtmaxTab(isim,ic,iqtl,iq)=lrtsol(isim,ic,iqtl)%lrtmax(iq-1)
             chrmaxTab(isim,ic,iqtl,iq)=lrtsol(isim,ic,iqtl)%chrmax(iq-1)
             nxmaxTab(isim,ic,iqtl,iq)=lrtsol(isim,ic,iqtl)%nxmax(iq-1)
            end do
           end do
          end do
#endif
          !print *,'thread_rang:',thread_rang,' isim=',isim,lrtsol(isim,1,1)%lrtmax(0),lrtsol(isim,1,1)%nxmax(0)
        end do

#ifdef MPI_ACTIVE
!        print *,"proc:",thread_rang,"car1",lrtmaxTab(:,1,1,1),chrmaxTab(:,1,1,1),nxmaxTab(:,1,1,1)
!        print *,"proc:",thread_rang,"car2",lrtmaxTab(:,2,1,1),chrmaxTab(:,2,1,1),nxmaxTab(:,2,1,1)
!        print *,"proc:",thread_rang,"car3",lrtmaxTab(:,3,1,1),chrmaxTab(:,3,1,1),nxmaxTab(:,3,1,1)

        !compactage des infos a envoyer au maitre
        lrtmaxTab_lin = pack(lrtmaxTab,mask=.true.)
        chrmaxTab_lin = pack(chrmaxTab,mask=.true.)
        nxmaxTab_lin = pack(nxmaxTab,mask=.true.)

        !Declaration des types a passer par message
        call MPI_TYPE_CONTIGUOUS (nbsimulByThread*ncar*opt_qtl*opt_qtl, MPI_DOUBLE_PRECISION, dim3double,code)
        call MPI_TYPE_COMMIT(dim3double,code)
        call MPI_TYPE_CONTIGUOUS (nbsimulByThread*ncar*opt_qtl*opt_qtl, MPI_INTEGER, dim3integer,code)
        call MPI_TYPE_COMMIT(dim3integer,code)


        ! Ici le processus Maitre (0) recupere l'ensemble des resultats de la simulation....
        deb_indice=nbsimulByThread * (thread_rang)+1
        call MPI_GATHER (lrtmaxTab_lin(1),1, dim3double ,lrtmaxTab_final(deb_indice),1,dim3double, 0 ,MPI_COMM_WORLD ,code)
        call MPI_GATHER (chrmaxTab_lin(1),1, dim3integer ,chrmaxTab_final(deb_indice),1,dim3integer, 0 ,MPI_COMM_WORLD ,code)
        call MPI_GATHER (nxmaxTab_lin(1),1, dim3integer ,nxmaxTab_final(deb_indice),1,dim3integer, 0 ,MPI_COMM_WORLD ,code)

        call MPI_TYPE_FREE(dim3double,code)
        call MPI_TYPE_FREE(dim3integer,code)

        !Gestion seulement par le processus maitre
        if (thread_rang == 0  ) then
           !on libere l allocation
           deallocate (lrtmaxTab)
           deallocate (chrmaxTab)
           deallocate (nxmaxTab)

           !on realloue pour l ensemble de la simulation
           allocate (lrtmaxTab(nsim,ncar,opt_qtl,opt_qtl))
           allocate (chrmaxTab(nsim,ncar,opt_qtl,opt_qtl))
           allocate (nxmaxTab(nsim,ncar,opt_qtl,opt_qtl))

           !agregation des infos
           do i=1,nb_procs
             s1 = nbsimulByThread*ncar*opt_qtl*opt_qtl * (i-1)+1
             s2 = nbsimulByThread*ncar*opt_qtl*opt_qtl * (i)

             sp1 = nbsimulByThread * (i-1)+1
             sp2 = nbsimulByThread * (i)

             lrtmaxTab(sp1:sp2,:,:,:) = unpack(lrtmaxTab_final(s1:s2),mask=unpackAll,field=lrtmaxTab(sp1:sp2,:,:,:))
             chrmaxTab(sp1:sp2,:,:,:) = unpack(chrmaxTab_final(s1:s2),mask=unpackAll,field=chrmaxTab(sp1:sp2,:,:,:))
             nxmaxTab(sp1:sp2,:,:,:) = unpack(nxmaxTab_final(s1:s2),mask=unpackAll,field=nxmaxTab(sp1:sp2,:,:,:))
           end do
!             print *,""
!             print *,"Processus maitre...",nsim,ncar,opt_qtl
!             print *,""
!             print *,"car1",lrtmaxTab(:,1,1,1),chrmaxTab(:,1,1,1),nxmaxTab(:,1,1,1)
!             print *,"car2",lrtmaxTab(:,2,1,1),chrmaxTab(:,2,1,1),nxmaxTab(:,2,1,1)
!             print *,"car3",lrtmaxTab(:,3,1,1),chrmaxTab(:,3,1,1),nxmaxTab(:,3,1,1)

             ! on libere les lrtsols effectuer par le maitre
             do isim=startSim,endSim
               do i=1,size(lrtsol,2)
                do j=1,size(lrtsol,3)
                 call release(lrtsol(isim,i,j))
                end do
               end do
             end do

             !remplissage de la structure lrtsol pour calculer les quantile et ecrire par le processus maitre le fichier de simulation
             do isim=1,nsim
                do ic=1,ncar
                 do iqtl=1,opt_qtl
                  call new(opt_qtl,lrtsol(isim,ic,iqtl))
                  do iq=1,iqtl
                   lrtsol(isim,ic,iqtl)%lrtmax(iq-1) = lrtmaxTab(isim,ic,iqtl,iq)
                   lrtsol(isim,ic,iqtl)%chrmax(iq-1) = chrmaxTab(isim,ic,iqtl,iq)
                   lrtsol(isim,ic,iqtl)%nxmax(iq-1) = nxmaxTab(isim,ic,iqtl,iq)
                  end do
                 end do
                end do
             end do

             ! Pour le maitre on doit liberer tous les lrtsol
             startSim = 1
             endSim  = nsim
        end if
#endif
        !retour chariot pour l affichage de progression
        if (thread_rang == 0  ) then
         write(*,*)''

         call write_simulation_file(opt_qtl,is_multitrait_analysis(opt_calcul),nsim,lrtsol)
         call log_mess('computing quantiles...',VERBOSE_DEF)
         call quantile(simula,nsim,opt_calcul,opt_qtl,nqtlsimul,lrtsol)

         if (.not. isThresholdEstim ) call end_output_handler(opt_qtl)
        end if

         do isim=startSim,endSim
          do i=1,size(lrtsol,2)
           do j=1,size(lrtsol,3)
            call release(lrtsol(isim,i,j))
           end do
          end do
         end do

        deallocate(lrtsol)
        deallocate(qtl)
        deallocate (rhoi)

#ifdef MPI_ACTIVE
        deallocate (lrtmaxTab)
        deallocate (chrmaxTab)
        deallocate (nxmaxTab)
        deallocate (lrtmaxTab_lin)
        deallocate(lrtmaxTab_final)
        deallocate (chrmaxTab_lin)
        deallocate(chrmaxTab_final)
        deallocate (nxmaxTab_lin)
        deallocate(nxmaxTab_final)
        deallocate (unpackAll)
#endif
       end subroutine qtlmap_simul


       subroutine print_information_max_lrt(opt_qtl,level,lrtsol)
         integer            , intent(in)                      :: opt_qtl,level
         type(TYPE_LRT_SOLUTION)  , dimension(ncar,opt_qtl), intent(in)      :: lrtsol

         integer :: i,c,iqtl,nqtl,iqtl2
         character(len=LEN_LINE) :: dxlist

         do nqtl=1,opt_qtl
          do iqtl=0,nqtl-1

           call log_mess('    ***************************************',level)
           call log_mess('    *******   Test '//trim(str(iqtl))//' QTL / '//trim(str(nqtl))//' QTL   ********',level)
           call log_mess('    ***************************************',level)

           do i=1,size(carac)
            if ( .not. associated(lrtsol(i,nqtl)%lrtmax) ) cycle
            call log_mess('Trait ['//trim(trim(carac(i)))//']',level)
            call log_mess('Maximum likelihood ratio test :',level)

            dxlist=adjustl("[")
          !  vallist="["

            do iqtl2=0,nqtl-1
              dxlist=trim(dxlist)//trim(str(absi(lrtsol(i,nqtl)%chrmax(iqtl2),lrtsol(i,nqtl)%nxmax(iqtl2))))//&
              ", Ch "//trim(chromo(lrtsol(i,nqtl)%chrmax(iqtl2)))
              if ( iqtl2 < nqtl-1) dxlist=trim(dxlist)//"] ["
            !  vallist=vallist//" "//trim(str(lrtsol(i,nqtl)%lrtmax(iqtl2)))
            end do
            dxlist=trim(dxlist)//"]"

            call log_mess('The maximum is reached at position '//trim(dxlist)&
            //' Morgan, with value '//trim(str(lrtsol(i,nqtl)%lrtmax(iqtl))),level)
          end do
         end do
        end do

       end subroutine


       subroutine print_information_max_lrt_multi(opt_qtl,level,lrtsol)
         integer            , intent(in)                      :: opt_qtl,level
         type(TYPE_LRT_SOLUTION)  , dimension(opt_qtl), intent(in)      :: lrtsol
         character(len=LEN_L)         :: list_traits,sep

         integer :: i,c
         list_traits = ''
         sep=' '
         do i=1,size(carac)
            list_traits=list_traits//trim(carac(i))//sep
         end do
         call log_mess('    ****   Join traits [ '//trim(list_traits)//']  ****',level)
          call log_mess('    ***************************************',level)
          call log_mess('    ***** Test 0 QTL / 1 QTL   ************',level)
          call log_mess('    ***************************************',level)
          call log_mess('Maximum likelihood ratio test :',level)
          call log_mess('The maximum is reached at position ['//trim(str(absi(lrtsol(1)%chrmax(0),lrtsol(1)%nxmax(0))))&
          //", Ch "//trim(chromo(lrtsol(1)%chrmax(0)))//&
          '] M, with value '//trim(str(lrtsol(1)%lrtmax(0))),level)

!          if (opt_qtl == 2 ) then
!           call log_mess('Test 0 QTL / 2 QTL',level)
!           call log_mess('Maximum likelihood ratio test :',level)
!           call log_mess('The maximum is reached at position ['//trim(str(absi(lrtsol(1)%chrmax(0),lrtsol(1)%nxmax(0))))//&
!             ',Ch '//trim(chromo(lrtsol(1)%chrmax(0)))//'] and ['//trim(str(absi(lrtsol(1)%chrmax(1),lrtsol(1)%nxmax(1))))//&
!             ',Ch '//trim(chromo(lrtsol(1)%chrmax(1)))//'] M, with values [0/2] '//trim(str(lrtsol(1)%lrtmax(0)))//&
!             ' and [1/2] '//trim(str(lrtsol(1)%lrtmax(1))),level)
!          end if
       end subroutine


end program qtlmap

