!!****m* TOOLS/m_qtlmap_math
!!  NAME
!!    m_qtlmap_math -- Mathematical subroutines
!!  SYNOPSIS
!!    This modules purposes an interface with NAG specification. In a NAG environment, the nag subroutines is called
!!    otherwise an implementation of the subroutine using SLATEC is called
!!  DESCRIPTION
!!
!!  NOTES
!!
!!  BUGS
!!
!!  SEE ALSO
!!
!!  COPYRIGHT
!!***
!! You can use this space for remarks that should not be included
!! in the documentation.
!!/
module m_qtlmap_math
    use m_qtlmap_constant
    use m_qtlmap_log
    use m_qtlmap_tools

    implicit none
    private

#ifndef HAVE_LIBNAG
    double precision    ,  parameter , private    :: SQRT_2 = 1.414213562373095d0 ! sqrt(2.d0)
#endif

    ! Give some information about library used by this module
    public :: MATH_QTLMAP_INFO
    ! computes the upper and lower tail probabilities and the probability density function of the beta
    ! distribution with parameters a and b.
    public :: MATH_QTLMAP_G01EEF

    ! calculates the mean, standard deviation, coefficients of skewness and kurtosis, and the maximum
    ! and minimum values for a set of ungrouped data. Weighting may be used.
    public :: MATH_QTLMAP_G01AAF

    !G01EAF returns a one or two-tail probability for the standard Normal distribution, via the routine name.
    public :: MATH_QTLMAP_G01EAF

    !G01FAF returns the deviate associated with the given probability of the standard Normal distribution, via
    ! the routine name.
    public :: MATH_QTLMAP_G01FAF

    ! returns the lower tail, upper tail or two-tail probability for the Student’s t-distribution with real
    ! degrees of freedom, via the routine name.
    public :: MATH_QTLMAP_G01EBF

    ! Performs a canonical variate (canonical discrimination) analysis.
    public :: MATH_QTLMAP_G03ACF

    public :: test_G03ACF

    ! performs a pseudo-random permutation of a vector of integers.
    public :: MATH_QTLMAP_G05EHF

    ! rearranges a vector of real numbers into ascending or descending order.
    public :: MATH_QTLMAP_M01CAF

    ! ranks a vector of real numbers in ascending or descending order.
    public :: MATH_QTLMAP_M01DAF

    ! Ranks a vector of integer numbers in ascending or descending order.
    public :: MATH_QTLMAP_M01DBF

    ! Ranks a vector of character data in ASCII or reverse ASCII order of a specified substring.
    public :: MATH_QTLMAP_M01DCF

    ! Inverts a permutation, and hence converts a rank vector to an index vector, or vice versa.
    public :: MATH_QTLMAP_M01ZAF

    ! Ranks the rows of a matrix of integer numbers in ascending or descending order.
    public :: MATH_QTLMAP_M01DFF

    ! returns the lower or upper tail probability for the 12 distribution with real degrees of freedom, via
    ! the routine name
    public :: MATH_QTLMAP_G01ECF

    !calculates the determinant of a real symmetric positive-definite matrix using a Cholesky factorization.
    public :: MATH_QTLMAP_F03ABF
    !calculates the inverse and the determinant of a real symmetric positive-definite.
    public :: MATH_QTLMAP_INVDETMATSYM
    ! calculates the approximate inverse of a real symmetric positive-definite matrix, using a Cholesky
    ! factorization.
    public :: MATH_QTLMAP_F01ADF
    ! Computes the lower of the beta distribution with parameters a and b.
    public :: MATH_QTLMAP_LOWERTAIL_BETA

    !S15AEF returns the value of the error function erf x, via the routine name.
    public :: MATH_QTLMAP_S15AEF
    ! Returns the value of the complementary error function, erfc x, via the routine name.
    public :: MATH_QTLMAP_S15ADF
    !returns generalized inverse of matrix x of size n x n declared
    ! as m x m.
    public :: ginv1

    contains

!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_INFO
!!  NAME
!!    MATH_QTLMAP_INFO
!!  DESCRIPTION
!!   Give some information about library used by this module
!!  NOTES
!!***
      subroutine MATH_QTLMAP_INFO

#ifdef HAVE_LIBNAG
           call log_mess("************* QTLMap use NAG library *****************************",INFO_DEF)
#else
           call log_mess("************* QTLMap use SLATEC/Orderpack-2.0 library:*****************************",INFO_DEF)
#endif

      end subroutine MATH_QTLMAP_INFO
!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_G01EEF
!!  NAME
!!    MATH_QTLMAP_G01EEF
!!  DESCRIPTION
!!   Computes the upper and lower tail probabilities and the probability density function of the beta
!!   distribution with parameters a and b.
!!  NOTES
!!***
      subroutine MATH_QTLMAP_G01EEF(X, A, B, TOL, P, Q, PDF, IFAIL)
           double precision,intent(in)     ::  X, A, B, TOL
           double precision,intent(out)    ::  P, Q, PDF
           integer , intent(inout)         ::  IFAIL

#ifndef HAVE_LIBNAG
           double precision                :: AGAMMA,BGAMMA
#endif

           call log_mess('MATH_QTLMAP_G01EEF (IN) ==== X:'//str(X),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01EEF (IN) ==== A:'//str(A),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01EEF (IN) ==== B:'//str(B),DEBUG_DEF)
           P=0.d0
           Q=0.d0
           PDF=0.d0

           IF ( X < 0 .or. X > 1 ) then
              IFAIL = 1
              CALL log_mess('MATH_QTLMAP_G01EEF:'//str(X)//' have to be between 0 and 1.',ERROR_DEF)
              RETURN
           END IF

           IF ( A < 0 )  then
             IFAIL = 2
             CALL log_mess('MATH_QTLMAP_G01EEF:A is less than 0',ERROR_DEF)
             RETURN
           END IF

           IF ( B < 0 ) then
             IFAIL = 2
             CALL log_mess('MATH_QTLMAP_G01EEF:B is less than 0',ERROR_DEF)
             RETURN
           END IF

#ifdef HAVE_LIBNAG
           call G01EEF(X, A, B, TOL, P, Q, PDF, IFAIL)
#else
           !The lower tail is also knwon as incomplete beta function
           P = DBETAI(X,A,B)

           IF (IFAIL /= 0 ) RETURN

           Q = 1 - P
           IF ( P == 0 .or. P == 1 ) THEN
             PDF = 0.d0
           ELSE
             AGAMMA = DGAMMA(A)
             BGAMMA = DGAMMA(B)
             IF (AGAMMA == 0 ) then
               IFAIL = 3
               CALL log_mess('MATH_QTLMAP_G01EEF:GAMMA(A) =  0',ERROR_DEF)
               RETURN
             END IF
             IF (BGAMMA == 0 ) then
               IFAIL = 3
               CALL log_mess('MATH_QTLMAP_G01EEF:GAMMA(B) =  0',ERROR_DEF)
               RETURN
             END IF
             PDF = (DGAMMA(A+B))/(AGAMMA*BGAMMA)
             PDF = PDF*(X**(A-1))*((1-X)**(B-1))
           ENDIF
#endif
           call log_mess('MATH_QTLMAP_G01EEF (OUT) ==== P:'//str(P),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01EEF (OUT) ==== Q:'//str(Q),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01EEF (OUT) ==== PDF:'//str(PDF),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01EEF (OUT) ==== IFAIL:'//str(IFAIL),DEBUG_DEF)
      end subroutine MATH_QTLMAP_G01EEF


!****************************************************************************************************************

!!****f* m_qtlmap_math/MATH_QTLMAP_LOWERTAIL_BETA
!!  NAME
!!    MATH_QTLMAP_LOWERTAIL_BETA
!!  DESCRIPTION
!!   Computes the lower of the beta distribution with parameters a and b.
!!  NOTES
!!   NAG : G01EEF
!!
!!***
      subroutine MATH_QTLMAP_LOWERTAIL_BETA(X, A, B, TOL, P, IFAIL)
           double precision,intent(in)     ::  X, A, B, TOL
           double precision,intent(out)    ::  P
           integer , intent(inout)         ::  IFAIL

#ifdef HAVE_LIBNAG
           double precision                :: Q, PDF
           Q=0.d0
           PDF=0.d0

#endif
           P=0.d0
#ifdef DEBUG
           call log_mess('MATH_QTLMAP_LOWERTAIL_BETA (IN) ==== X:'//str(X),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_LOWERTAIL_BETA (IN) ==== A:'//str(A),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_LOWERTAIL_BETA (IN) ==== B:'//str(B),DEBUG_DEF)
#endif
           IF ( X < 0 .or. X > 1 ) then
              IFAIL = 1
              CALL log_mess('MATH_QTLMAP_LOWERTAIL_BETA:'//str(X)//' have to be between 0 and 1.',ERROR_DEF)
              RETURN
           END IF

           IF ( A < 0 )  then
             IFAIL = 2
             CALL log_mess('MATH_QTLMAP_LOWERTAIL_BETA :A is less than 0',ERROR_DEF)
             RETURN
           END IF

           IF ( B < 0 ) then
             IFAIL = 2
             CALL log_mess('MATH_QTLMAP_LOWERTAIL_BETA :B is less than 0',ERROR_DEF)
             RETURN
           END IF

#ifdef HAVE_LIBNAG
           call G01EEF(X, A, B, TOL, P, Q, PDF, IFAIL)
#else
           !The lower tail is also knwon as incomplete beta function
           P = DBETAI(X,A,B)
#endif
#ifdef DEBUG
           call log_mess('MATH_QTLMAP_LOWERTAIL_BETA (OUT) ==== P:'//str(P),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_LOWERTAIL_BETA (OUT) ==== IFAIL:'//str(IFAIL),DEBUG_DEF)
#endif
      end subroutine MATH_QTLMAP_LOWERTAIL_BETA

!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_G01EAF
!!  NAME
!!    MATH_QTLMAP_G01EAF
!!  DESCRIPTION
!!   Returns a one or two-tail probability for the standard Normal distribution, via the routine name.
!!
!!   The lower tail probability :
!!          P(X<=x) = 1/2*derfc(-x/sqrt(2))
!!
!!   The upper tail probability :
!!          P(X>=x) = P(X<=-x)
!!
!!   'S':P(X>=x)+P(X<=-|x|)
!!
!!   'C':P(X<=x)-P(X<=-|x|)
!!  NOTES
!!***
      function MATH_QTLMAP_G01EAF (TAIL, X, IFAIL) result (P)
           INTEGER ,intent(inout)      :: IFAIL
           double precision,intent(in) :: X
           CHARACTER(len=1),intent(in) :: TAIL

           double precision            :: P
#ifdef HAVE_LIBNAG
           double precision, external :: G01EAF
           P = G01EAF (TAIL, X, IFAIL)
#else
           double precision            :: X1,X2
           IFAIL = 0
           ! The lower tail probability : P(X<=x) = 1/2*derfc(-x/sqrt(2))
           if (TAIL == 'L' ) then
               X1=-X/SQRT_2
               P = 0.5*derfc(X1)
               return
           end if

           ! The upper tail probability : P(X>=x) = P(X<=-x)
           if ( TAIL == 'U') then
               X1=x/SQRT_2
               P = 0.5*derfc(X1)
               return
           end if

           !P(X>=x)+P(X<=-|x|)
           if ( TAIL == 'S') then
              X1=x/SQRT_2
              X2=-abs(x)/SQRT_2
              P = 0.5*derfc(X1) + 0.5*derfc(X2)
              return
           end if

           !P(X<=x)-P(X<=-|x|)
           if ( TAIL == 'C') then
             X1=-x/SQRT_2
             X2=-abs(x)/SQRT_2
             P = 0.5*derfc(X1) - 0.5*derfc(X2)
             return
           end if

           IFAIL = 1
           call stop_application('MATH_QTLMAP_G01EAF : unknown value for TAIL:'//TAIL)

#endif
           return
      end function MATH_QTLMAP_G01EAF

!****************************************************************************************************************

!!****f* m_qtlmap_math/MATH_QTLMAP_G01FAF
!!  NAME
!!    MATH_QTLMAP_G01FAF
!!  DESCRIPTION
!!   Returns the deviate associated with the given probability of the standard Normal distribution, via
!!   the routine name.
!!  NOTES
!!***
       function MATH_QTLMAP_G01FAF (TAIL, P, IFAIL) result (X)
           INTEGER ,intent(inout)      :: IFAIL
           double precision,intent(in) :: P
           CHARACTER(len=1),intent(in) :: TAIL
           double precision            :: X

#ifdef HAVE_LIBNAG
           double precision, external :: G01FAF
           x = G01FAF (TAIL, P, IFAIL)
#else
           x = dinvnorm(p)
#endif
           return

      end function MATH_QTLMAP_G01FAF


!****************************************************************************************************************

!!****f* m_qtlmap_math/MATH_QTLMAP_G01AAF
!!  NAME
!!    MATH_QTLMAP_G01AAF
!!  DESCRIPTION
!!   Calculates the mean, standard deviation, coefficients of skewness and kurtosis, and the maximum
!!   and minimum values for a set of ungrouped data. Weighting may be used.
!!  NOTES
!!***
      subroutine MATH_QTLMAP_G01AAF(N, X, IWT, WT, XMEAN, S2, S3, S4, XMIN, XMAX, WTSUM,IFAIL)
           integer , intent(in)          :: N
           integer , intent(inout)       :: IWT
           integer, intent(out)          :: IFAIL
           double precision,intent(in)   :: X(N)
           double precision,intent(inout):: WT(N)
           double precision,intent(out)  :: XMEAN, S2, S3, S4, XMIN, XMAX, WTSUM

#ifndef HAVE_LIBNAG
           integer                       :: I
           double precision              :: SOMMWTCARR,BUF,BUF2,D
#endif
#ifdef DEBUG
           call log_mess('MATH_QTLMAP_G01AAF (IN) ==== N:'//str(N),DEBUG_DEF)
          ! call log_mess('MATH_QTLMAP_G01AAF (IN) ==== X:'//str(X),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01AAF (IN) ==== IWT:'//str(IWT),DEBUG_DEF)
#endif
#ifdef HAVE_LIBNAG
           call G01AAF(N, X, IWT, WT, XMEAN, S2, S3, S4, XMIN, XMAX, WTSUM,IFAIL)

#else

           IF (N < 1) THEN
              IFAIL = 1
              CALL log_mess('MATH_QTLMAP_G01AAF: N < 1 N:'//str(N),ERROR_DEF)
              RETURN
           END IF

           !no weight is associated with data
           IF (IWT == 0) THEN
              WT=1.d0
              IWT=1
           ELSE
              DO I=1,N
               IF ( WT(I) < 0 ) THEN
                IFAIL = 3
                CALL log_mess('MATH_QTLMAP_G01AAF: Weigth is negative for element :'//str(I),ERROR_DEF)
                RETURN
               END IF
              END DO
           END IF

           WTSUM=0.d0
           XMEAN=0.d0
           SOMMWTCARR=0.d0
           !Somme of weight and Mean
           DO I=1,N
             WTSUM = WTSUM + WT(I)
             XMEAN = XMEAN + (WT(I)*X(I))
             SOMMWTCARR = SOMMWTCARR + (WT(I)*WT(I))
           END DO

           IF ( WTSUM == 0 ) THEN
                IFAIL = 2
                CALL log_mess('MATH_QTLMAP_G01AAF: Weigth somme equal 0 ',ERROR_DEF)
                RETURN
           END IF

           XMEAN = XMEAN / WTSUM

           !Standart deviation - Coefficient of skewness - Coefficient of kurtosis
           D = WTSUM - (SOMMWTCARR / WTSUM )

           IF ( D == 0 ) THEN
                IFAIL = 2
                CALL log_mess('MATH_QTLMAP_G01AAF:'//&
                ' Can not compute Standart deviation-Coefficient of skewness- Coefficient of kurtosis',ERROR_DEF)
                RETURN
           END IF

           XMIN = X(1)
           XMAX = X(1)
           BUF= 0.d0
           S2 = 0.d0
           S3 = 0.d0
           S4 = 0.d0
           DO I=1,N
             IF ( XMIN > X(I) ) XMIN = X(I)
             IF ( XMAX < X(I) ) XMAX = X(I)
             BUF = (X(I)-XMEAN)
             BUF2 = WT(I)*BUF*BUF
             S2 = S2 + BUF2
             BUF2 = BUF2*BUF
             S3 = S3 + BUF2
             BUF2 = BUF2*BUF
             S4 = S4 + BUF2
           END DO
           S2 = S2 / D
           S2 = sqrt(S2)

           S3 = S3 / ( D*(S2**3) )
           S4 = ( S4 / ( D*(S2**4) ) ) - 3

#endif
#ifdef DEBUG
           call log_mess('MATH_QTLMAP_G01AAF (OUT) ==== XMEAN:'//str(XMEAN),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01AAF (OUT) ==== S2:'//str(S2),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01AAF (OUT) ==== S3:'//str(S3),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01AAF (OUT) ==== S4:'//str(S4),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01AAF (OUT) ==== XMIN:'//str(XMIN),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01AAF (OUT) ==== XMAX:'//str(XMAX),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01AAF (OUT) ==== WTSUM:'//str(WTSUM),DEBUG_DEF)
           call log_mess('MATH_QTLMAP_G01AAF (OUT) ==== IFAIL:'//str(IFAIL),DEBUG_DEF)
#endif
      end subroutine MATH_QTLMAP_G01AAF


!****************************************************************************************************************

!!****f* m_qtlmap_math/MATH_QTLMAP_G01ECF
!!  NAME
!!    MATH_QTLMAP_G01ECF
!!  DESCRIPTION
!!    Returns the lower or upper tail probability for the 12 distribution with real degrees of freedom, via
!!    the routine name
!!  NOTES
!!   SLATEC :
!!     * DGAMI : incomplete gamma function
!!***
      function MATH_QTLMAP_G01ECF(TAIL, X, DF, IFAIL) result(res)
         integer , intent(out)        :: IFAIL
         real(kind=dp),intent(in)     :: X, DF
         character(len=1) ,intent(in) :: TAIL

#ifdef HAVE_LIBNAG
         real, external :: g01ecf
#else
         double precision,external       :: DGAMI ! SLATEC : incomplete gamma function
         double precision                :: A,B
#endif

         real                            :: res
#ifdef DEBUG
         call log_mess('MATH_QTLMAP_G01ECF (IN) ==== TAIL:'//TAIL,DEBUG_DEF)
         call log_mess('MATH_QTLMAP_G01ECF (IN) ==== X:'//str(X),DEBUG_DEF)
         call log_mess('MATH_QTLMAP_G01ECF (IN) ==== DF:'//str(DF),DEBUG_DEF)
#endif
         IFAIL = 0
         IF ( TAIL /= 'L' .and. TAIL /= 'U' ) THEN
             IFAIL = 1
             call log_mess('MATH_QTLMAP_G01ECF (OUT) ==== IFAIL:'//str(IFAIL),DEBUG_DEF)
             CALL log_mess('MATH_QTLMAP_G01ECF: Bad value of TAIL (U or L):'//TAIL,ERROR_DEF)
             RETURN
         END IF

         IF ( X < 0 ) THEN
             IFAIL = 2
             call log_mess('MATH_QTLMAP_G01ECF (OUT) ==== IFAIL:'//str(IFAIL),DEBUG_DEF)
             CALL log_mess('MATH_QTLMAP_G01ECF: X have to be positive or equal to 0:'//str(X),ERROR_DEF)
             RETURN
         END IF

         IF ( DF <= 0 ) THEN
             IFAIL = 3
             call log_mess('MATH_QTLMAP_G01ECF (OUT) ==== IFAIL:'//str(IFAIL),DEBUG_DEF)
             CALL log_mess('MATH_QTLMAP_G01ECF: DF have to be strictly positive:'//str(DF),ERROR_DEF)
             RETURN
         END IF

#ifdef HAVE_LIBNAG
         RES = G01ECF(TAIL, X, DF, IFAIL)
#else
         A = DGAMI(DF/2,X/2) ! lower incomplete gamma function
         B = DGAMMA(DF/2)
         RES = A / B

       !  PRINT *,RES
       !  STOP
         IF ( TAIL == 'L' ) RETURN
         IF ( TAIL == 'U') RES = 1-RES


#endif
#ifdef DEBUG
         call log_mess('MATH_QTLMAP_G01ECF (OUT) ==== RES:'//str(RES),DEBUG_DEF)
         call log_mess('MATH_QTLMAP_G01ECF (OUT) ==== IFAIL:'//str(IFAIL),DEBUG_DEF)
#endif
         RETURN
      end function MATH_QTLMAP_G01ECF

!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_G01EBF
!!  NAME
!!    MATH_QTLMAP_G01EBF
!!  DESCRIPTION
!!   Returns the lower tail, upper tail or two-tail probability for the Student’s t-distribution with real
!!   degrees of freedom, via the routine name.
!!  NOTES
!!  HISTORY
!!   06/10/2009 : use statlib student function and stop implementation with dbetai from slatec
!!***
      function MATH_QTLMAP_G01EBF(TAIL, T, DF, IFAIL) result(res)
         integer , intent(out)        :: IFAIL
         double precision,intent(in)  :: T, DF
         character(len=1) ,intent(in) :: TAIL
#ifdef HAVE_LIBNAG
         double precision, external :: g01ebf
#endif
         double precision             :: RES

#ifdef DEBUG
         call log_mess('MATH_QTLMAP_G01EBF (IN) ==== TAIL:'//TAIL,DEBUG_DEF)
         call log_mess('MATH_QTLMAP_G01EBF (IN) ==== T:'//str(T),DEBUG_DEF)
         call log_mess('MATH_QTLMAP_G01EBF (IN) ==== DF:'//str(DF),DEBUG_DEF)
#endif
         IF ( TAIL /= 'L' .and. TAIL /= 'U' .and. TAIL /= 'C' .and. TAIL /='S' ) THEN
             IFAIL = 1
             CALL log_mess('MATH_QTLMAP_G01EBF: Bad value of TAIL (U,S,C,L):'//TAIL,ERROR_DEF)
             RETURN
         END IF

         IF ( DF < 1 ) THEN
             IFAIL = 3
             CALL log_mess('MATH_QTLMAP_G01EBF: DF have to be strictly superior to one:'//str(DF),ERROR_DEF)
             RETURN
         END IF


#ifdef HAVE_LIBNAG

         RES = G01EBF(TAIL, T, DF, IFAIL)
#else
         IF ( TAIL == 'L') RES = 1-STUDNT(T,DF,IFAIL)
         IF ( TAIL == 'U') RES = STUDNT(T,DF,IFAIL)
         IF ( TAIL == 'S' ) RES = 2*STUDNT(T,DF,IFAIL)
         IF ( TAIL == 'C' ) RES = 1 - 2*STUDNT(T,DF,IFAIL)

#endif
#ifdef DEBUG
        call log_mess('MATH_QTLMAP_G01EBF (OUT) ==== RES:'//str(RES),DEBUG_DEF)
        call log_mess('MATH_QTLMAP_G01EBF (OUT) ==== IFAIL:'//str(IFAIL),DEBUG_DEF)
#endif
      end function MATH_QTLMAP_G01EBF


!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_G03ACF
!!  NAME
!!    MATH_QTLMAP_G05EHF
!!  DESCRIPTION
!!   Performs a canonical variate (canonical discrimination) analysis.
!!  INPUT
!!
!!     WEIGHT    : 'U' no weights ar used . ’W’ or ’V’, weights are used and must be supplied in WT.
!!                 Constraint: WEIGHT 1⁄4 ’U’; ’W’ or ’V’.
!!                 In the case of WEIGHT 1⁄4 ’W’, the weights are treated as frequencies and the effective number of
!!                 observations is the sum of the weights. If WEIGHT 1⁄4 ’V’, the weights are treated as being
!!                 inversely proportional to the variance of the observations and the effective number of observations is
!!                 the number of observations with non-zero weights.
!!
!!     N         : the number of observations. Constraint: N >= NX + NG.
!!     M         : the total number of variables. Constraint: M >= NX.
!!     X(LDX,M)  : X(i,j) must contain the ith observation for the jth variable, 1<=i<=N, 1<=j<=M
!!     LDX       : the first dimension of the array X LDX >= N
!!     ISX(M)    : indicates whether or not the jth variable is to be included in the analysis.
!!                 If ISX(j) > 0, then the variables contained in the jth column of X is included in the canonical variate analysis,
!!     NX        : the number of variables in the analysis, nx .Constraint : NX >= 1
!!     ING(N)    : ING(i) indicates which group the ith observation is in, for i=1,2,...,n. The effective
!!                 number of groups is the number of groups with non-zero membership.
!!                 Constraint 1<=NG(i)<= NG i=1,..,N
!!     NG        : the number of groups, ng . Constraint: NG >=2
!!     WT        : size(WT)>=N if WEIGHT='W' or 'V' , 1 otherwise.
!!                 if WEIGHT=’W’ or ’V’, then the first n elements of WT must contain the weights to used in the analysis.
!!                 if W(i)=0.0, then the ith observation is not included in the analysis.
!!                 if WEIGHT=’U’, then WT is not referenced.
!!                 Constraints:
!!                             WT(i)>=0.0,for i=1,2,..,N   ,
!!                             SOMME(WT(i))>=NX + effective number of groups , i=1,...,N
!!     LDCVM     : the dimension of the array CVM as declared in the program. Constraint LDCVM >=NG
!!     LDE       : the first dimension of the array E. Constraint: LDE >= min(NX,NG-1).
!!     LDCVX     : the first dimension of the array CVX
!!     TOL       : the value of TOL is used to decide if the variables are of full rank and, if not, what is the
!!                 rank of the variables. The smaller the value of TOL the stricter the criterion for selecting the
!!                 singular value decomposition. If a non-negative value of TOL less than machine precision is
!!                 entered, then the square root of machine precision is used instead.
!!
!!  OUTPUT
!!
!!    NIG(NG)       : NIG(j) gives the number of observations in group j, for j=1,2,...,ng.
!!    CVM(LDCVM,NX) : CVM(i,j) contains the mean of the jth canonical variate for the ith group, for
!!                   i=1,2,...,ng ; j=1,2,...,l the remaining columns, if any, are used as workspace.
!!    E(LDE,6)      : the statistics of the canonical variate analysis.
!!                   E(i,1), the canonical correlations, deltai , for i=1,2,...,l.
!!                   E(i,2), the eigenvalues of the within-group sum of squares matrix ,lambda2, for i=1,...,l.
!!                   E(i,3), the proportion of variation explained by the ith canonical variate, for i=1,...,l.
!!                   E(i,4), the Chi2 statistic for the ith canonical variate, for i=1,...,l.
!!                   E(i,5), the degrees of freedom for Chi2 statistic for the ith canonical variate, for i=1,...,l.
!!                   E(i,6), he significance level for the Chi2 statistic for the ith canonical variate, for i=1,...,l.
!!    NCV           :  he number of canonical variates, l. This will be the minimum of NG-1 and the rank of X.
!!    CVX(LDCVX,NG-1) : the canonical variate loadings. CVX(i,j) contains the loading coefficient for the ith
!!                      variable on the jth canonical variate, for i=1,...,NX ; j=1,...,l; the remaining columns, if
!!                      any, are used as workspace.
!!    IRANKX        :  the rank of the dependent variables.
!!                     If the variables are of full rank then IRANKX 1⁄4 NX.
!!                     If the variables are not of full rank then IRANKX is an estimate of the rank of the dependent
!!                    variables. IRANKX is calculated as the number of singular values greater than TOLÂ (largest singular value).
!!
!!
!!  NOTES
!!***
    subroutine  MATH_QTLMAP_G03ACF(WEIGHT, N, M, X, LDX, ISX, NX, ING, NG, WT, NIG, CVM, &
                LDCVM, E, LDE, NCV, CVX, LDCVX, TOL, IRANKX,IFAIL)
        CHARACTER(LEN=1)                   ,INTENT(in)     :: WEIGHT
        INTEGER                            ,INTENT(in)     :: N, M, LDX, NX, NG
        INTEGER                            ,INTENT(in)     :: LDCVM, LDE, LDCVX
        DOUBLE PRECISION  ,DIMENSION(LDX,M),INTENT(in)     :: X
        INTEGER           ,DIMENSION(M)    ,INTENT(in)     :: ISX
        INTEGER           ,DIMENSION(N)    ,INTENT(in)     :: ING
        DOUBLE PRECISION  ,DIMENSION(:)    ,INTENT(in)     :: WT
        DOUBLE PRECISION                   ,INTENT(in)     :: TOL
   !output
        INTEGER           ,DIMENSION(NG)   ,INTENT(out)    :: NIG
        DOUBLE PRECISION  ,DIMENSION(LDCVM,NX),INTENT(out) :: CVM
        DOUBLE PRECISION  ,DIMENSION(LDE,6),INTENT(out)    :: E
        INTEGER                            ,INTENT(out)    :: NCV
        DOUBLE PRECISION  ,DIMENSION(LDCVX,NG-1),INTENT(out):: CVX

        INTEGER                            ,INTENT(out)    :: IRANKX,  IFAIL



#ifdef HAVE_LIBNAG
      INTEGER                          :: IWK,STAT
      DOUBLE PRECISION  ,DIMENSION(:),ALLOCATABLE  :: WK     ! workspace
      external :: G03ACF

      IF (NX >= NG -1) THEN
         IWK = 1 + N*NX + max ( (5*(NX-1)) + ((NX+1)*NX) ,N)
      ELSE
         IWK = 1 + N*NX + max ( (5*(NX-1)) + ((NG-1)*NX),N)
      ENDIF

      allocate(WK(IWK),STAT=stat)

      if (stat /= 0 ) then
        call log_mess("Can not allocate Workspace for G03ACF",ERROR_DEF)
        stop
      end if

      call G03ACF(WEIGHT, N, M, X, LDX, ISX, NX, ING, NG, WT, NIG, CVM, &
                LDCVM, E, LDE, NCV, CVX, LDCVX, TOL, IRANKX, WK, IWK, &
                IFAIL)
      deallocate(WK);
#else
      INTEGER  :: I,J,IPRINT,IERR,COUNT,M2,K,MATZ,NROT,L,IRANK,IP,JP,MPRIME
      INTEGER, DIMENSION(NG) :: NOG
      DOUBLE PRECISION  :: TOTAL(M,M),BETWEEN(M,M),WITHIN(M,M),MEAN_VAR(M),MEAN_GROUP(NG,M)
      DOUBLE PRECISION  :: FV1(N),IV1(N),tol2,NPRIME
      DOUBLE PRECISION  :: W1(M), W2(M), BETW2(NG,NG), CPROJ(M,NG),MEAN_WEIGHT
      DOUBLE PRECISION  ,DIMENSION(LDX,M)      :: X2
      INTEGER           ,DIMENSION(N)        :: ING2

      REAL :: WR(M),WI(M),A(M,M),Z(M,M)

     ! external :: RSGAB,RG

      IF ( NX < 1 ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : NX < 1",ERROR_DEF);
         IFAIL=1
         RETURN
      END IF

      IF ( NG < 2 ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : NG < 2",ERROR_DEF);
         IFAIL=1
         RETURN
      END IF

      IF ( M < NX ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : M < NX",ERROR_DEF);
         IFAIL=1
         RETURN
      END IF

     IF ( N < (NX+NG) ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : N < NX+NG",ERROR_DEF);
         IFAIL=1
         RETURN
      END IF

     IF ( LDX < N ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : LDX < N",ERROR_DEF);
         IFAIL=1
         RETURN
      END IF

     IF ( LDCVX < NX ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : LDCVX < NX",ERROR_DEF);
         IFAIL=1
         RETURN
      END IF

     IF ( LDCVM < NG ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : LDCVM < NG",ERROR_DEF);
         IFAIL=1
         RETURN
      END IF

     IF ( LDE < min(NX,NG-1) ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : LDE < min(NX,NG-1)",ERROR_DEF);
         IFAIL=1
         RETURN
      END IF

     IF ( WEIGHT/='U' .and. WEIGHT/='W' .and. WEIGHT/='V' ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : WEIGHT /= 'U' or 'V' or 'W' ",ERROR_DEF);
         IFAIL=1
         RETURN
      END IF

      IF ( TOL < 0.d0 ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : TOL < 0.0",ERROR_DEF);
         IFAIL=1
         RETURN
      END IF
    ! IFAIL = 2
       IF ( ( WEIGHT=='V' .or. WEIGHT=='W') .and. minval(WT)<0.d0  ) THEN
         call log_mess("MATH_QTLMAP_G03ACF : WEIGHT:"//WEIGHT//" AND ONE VALUE OF WT < 0.0",ERROR_DEF);
         IFAIL=2
         RETURN
      END IF

    ! IFAIL = 3
    !   IF ( minval(ING) < 1 .or. maxval(ING) > NG ) THEN
    !     call log_mess("MINVAL(ING)="//str(minval(ING)),ERROR_DEF);
    !!     call log_mess("MAXVAL(ING)="//str(maxval(ING)),ERROR_DEF);
     !    call log_mess("NG="//str(NG),ERROR_DEF);
     !    call log_mess("MATH_QTLMAP_G03ACF : A value of ING is < 1 or a value of ING is > NG",ERROR_DEF);
    !     DO i=1,size(ING)
         !  call log_mess("ING("//trim(str(i))//')='//str(ING(i)),ERROR_DEF);
    !     END DO
        ! IFAIL=3
        ! RETURN
!      END IF
    IFAIL=0


    !Weights can be used with the analysis, in which case the weighted means are subtracted from each column
    !and then each row is scaled by an amount square(wi) , where wi is the weight for the ith observation (row).

     NPRIME=0

     DO I=1,N
        X2(I,:) = X(I,:)*wt(i)
        IF (WT(I)>0) NPRIME=NPRIME+WT(I)
     END DO

     IPRINT=3
     IERR=0



     NOG=0
     !*********** Calcul du nombre d observation dans chaque groupe
     NIG=0
     !NPRIME=0
     DO I=1,N
     !  PRINT *,I,":",X(I,:),' g:',ING(I),' w:',WT(I)
       IF ( ING(I) > 0 .AND. (WT(I)>0) ) THEN
         NIG(ING(I))= NIG(ING(I)) +1!WT(I)!+ 1
      !   NPRIME=NPRIME+1
       END IF
     END DO


     !*********** calcul des moyennes de chaque variable
     MEAN_VAR=0d0
     JP=0
     DO J=1,M
      IF (ISX(J)>0) THEN
       JP=JP+1
       DO I=1,N
        IF ( ING(I) > 0 ) THEN
          MEAN_VAR(JP) = X2(I,JP) + MEAN_VAR(JP)
        END IF
       END DO
       MEAN_VAR(JP) = MEAN_VAR(JP) / NPRIME
      END IF
    !  PRINT *,"MEAN:",J,MEAN_VAR(J)
     END DO

     !NB VARIABLES
     MPRIME=JP

     !*********** calcul des moyennes de chaque groupes et variables
     MEAN_GROUP=0d0
     DO I=1,N
      JP=0
      DO J=1,M
       IF (ISX(J)>0) THEN
        JP=JP+1
        IF ( ING(I) > 0 ) THEN
         MEAN_GROUP(ING(I),JP)= X2(I,JP) + MEAN_GROUP(ING(I),JP)
        END IF
       END IF
      END DO
     END DO

     JP=0
     DO J=1,M
      IF (ISX(J)>0) THEN
       JP=JP+1
       DO I=1,NG
        MEAN_GROUP(I,JP)= MEAN_GROUP(I,JP) / NIG(I)
   !    PRINT *,"MEAN_GROUP:",I,J,MEAN_GROUP(I,J)
       END DO
      END IF
     END dO

     !**** Calcul de la matrice covariance TOTAL......V/T
     TOTAL=0.d0
     IP=0
     DO I=1,M
      IF (ISX(I)>0) THEN
       IP=IP+1
       JP=0
       DO J=1,M
        IF (ISX(J)>0) THEN
         JP=JP+1
         DO K=1,N
           IF ( ING(K)>0 .AND. WT(K)>0 ) THEN
            TOTAL(IP,JP)= TOTAL(IP,JP) + (X2(K,IP) - MEAN_VAR(IP)) * (X2(K,JP) - MEAN_VAR(JP))
           END IF
            !PRINT *,IP,JP,TOTAL(IP,JP)
         END DO
         TOTAL(IP,JP)=TOTAL(IP,JP)/NPRIME
      !  PRINT *,"FINAL:",I,J,TOTAL(I,J)
        END IF
       END DO
      END IF
     END DO
     !stop

     !*********** Calcul de Between k (k=1,..,NG) : matrice de variance covariance
     BETWEEN=0.d0
     IP=0
     DO I=1,M
      IF (ISX(I)>0) THEN
       IP=IP+1
       JP=0
       DO J=1,M
        IF (ISX(J)>0) THEN
         JP=JP+1
         DO K=1,NG
          BETWEEN(IP,JP)= BETWEEN(IP,JP) + ( real(NIG(K)) / real(NPRIME) )* &
          (MEAN_GROUP(K,IP) - MEAN_VAR(IP)) * (MEAN_GROUP(K,JP) - MEAN_VAR(JP))
         END DO
        END IF
       END DO
      END IF
     END DO

      !**** Calcul de la matrice covariance WITHIN......
      IP=0
      WITHIN=0.d0
      DO I=1,M
       IF (ISX(I)>0) THEN
        IP=IP+1
        JP=0
        DO J=1,M
         IF (ISX(J)>0) THEN
          JP=JP+1
          DO K=1,NG
           DO L=1,N
            IF ( ING(L)>0) THEN
             WITHIN(IP,JP)= WITHIN(IP,JP) + (X2(L,IP) - MEAN_GROUP(K,IP)) * (X2(L,JP) - MEAN_GROUP(K,JP))
            END IF
            !PRINT *,I,J,TOTAL(I,J)
           END DO
          END DO
          WITHIN(IP,JP)=WITHIN(IP,JP)/NPRIME
        END IF
        END DO
       END IF
     END DO


!     DO I=1,M
!           PRINT *,'TOTAL(',i,')=',TOTAL(i,:)
!     END DO
!
!     DO I=1,M
!           PRINT *,'BETWEEN(',i,')=',BETWEEN(i,:)
!     END DO
!
!     DO I=1,M
!           PRINT *,'WITHIN(',i,')=',WITHIN(i,:)
!     END DO

     call ginv1(WITHIN,M,MPRIME,1.d-15,irank)

   !  call ginv1(TOTAL,M,MPRIME,1.d-15,ifail)
!      DO I=1,M
!          PRINT *,'INV TOTAL(',i,')=',TOTAL(i,:)
!     END DO

   !  A = matmul(TOTAL,BETWEEN)
      A = matmul(WITHIN,BETWEEN)

!      DO I=1,M
!          PRINT *,(A(i,j),j=1,M)
!     END DO
     WR=0
     Wi=0
     MATZ=2 ! eigenvalue and eigenvector
!     call RSGAB (M, M, TOTAL, BETWEEN, W, MATZ, Z, FV1, FV2, IERR)
      call RG (M, MPRIME, A, WR, WI, MATZ, Z, IV1, FV1, IFAIL)


     !call tdiag (M, tol2, a, wr, wi, z, M, ierr)

!     DO I=1,NG
!         PRINT *,'NIG(',I,')=',NIG(I)
!     END DO

!      DO I=1,NG
!        DO J=1,M
!         PRINT *,'CVM(',I,',',J,')=',MEAN_GROUP(I,J)
!        END DO
!     END DO

!     PRINT *,'IERR:',IERR
!     PRINT *,"REAL PART OF EIGEN VALUE"
!     PRINT *,WR
     M2=1
     DO I=2,MPRIME
       IF (WR(M2)<WR(I)) M2=I
     END DO
     CVX=0.d0

     DO J=1,NG-1
      DO I=1,MPRIME
       CVX(I,J)=Z(I,M2)
      END DO
     END DO

!     PRINT *,"IMAGINARY PART OF EIGEN VALUE"
!     PRINT *,WI

!     PRINT *,"EIGEN VECTOR"
!     DO I=1,M
!      print *,(Z(I,J),J=1,M)
!     END DO
!stop
 !    DO I=1,size(WT)
 !!      DO J=1,SIZE(X,2)
   !      D(I,J) = X(I,J)*WT(I)
  !     END DO
   !!  END DO

!     PRINT *,' N:',N,' M:',M,' NX:',NX
!     PRINT *,'N2:',N2
!     PRINT *,'M2:',M2
!
!     PRINT *,'NG:',NG
!     PRINT *,'LDX:',LDX
!
!     PRINT * ,'WEIGHT:',WEIGHT
!
!
!
!      DO I=1,size(NOG)
!       PRINT * ,'NOG(',i,')=',NOG(i)
!     END DO
!
!     PRINT *,'IERR:',IERR
!     DO I=1,M
!       PRINT * ,'MEAN(',i,')=',MEAN_VAR(i)
!     END DO
!
!     DO I=1,SIZE(MEAN_GROUP,1)
!       DO J=1,SIZE(MEAN_GROUP,2)
!           PRINT *,'MEAN_GROUP(',i,',',j,')=',MEAN_GROUP(i,j)
!       END DO
!     END DO
!
!     DO I=1,SIZE(TOTAL,1)
!       DO J=1,SIZE(TOTAL,2)
!           PRINT *,'TOTAL(',i,',',j,')=',TOTAL(i,j)
!       END DO
!     END DO
!
!      DO I=1,SIZE(BETWEEN,1)
!       DO J=1,SIZE(BETWEEN,2)
!           PRINT *,'BETWEEN(',i,',',j,')=',BETWEEN(i,j)
!       END DO
!     END DO
!
!     DO I=1,SIZE(BETW2,1)
!       DO J=1,SIZE(BETW2,2)
!           PRINT *,'BETW2(',i,',',j,')=',BETW2(i,j)
!       END DO
!     END DO
!
!     DO I=1,SIZE(CPROJ,1)
!       DO J=1,SIZE(CPROJ,2)
!           PRINT *,'CPROJ(',i,',',j,')=',CPROJ(i,j)
!       END DO
!     END DO
!
!     DO I=1,SIZE(W1)
!           PRINT *,'W1(',i,')=',W1(i)
!     END DO
!     DO I=1,SIZE(W2)
!           PRINT *,'W2(',i,')=',W2(i)
!     END DO
     ! IFAIL = 4



 !    call stop_application("MATH_QTLMAP_G03ACF pas implemente....")

#endif
    end subroutine MATH_QTLMAP_G03ACF


   subroutine test_G03ACF
     INTEGER :: N=9
     INTEGER :: M=3
     INTEGER :: NX=3
     INTEGER :: NG=3
     CHARACTER(LEN=1) :: WEIGHT='U'

     double precision    TOL
     INTEGER IFAIL, IRX, J, NCV
    double precision  ::CVM(9,9), CVX(9,9), E(9,6),WK(50)
    INTEGER ISX(6), NIG(9)


     DOUBLE PRECISION :: X(9,3),WT(9)
     INTEGER          :: ING(9),I

     data (X(1,i),i=1,3) /13.3,10.6,21.2/
     data (X(2,i),i=1,3) /13.6,10.2,21.0/
     data (X(3,i),i=1,3) /14.2,10.7,21.1/
     data (X(4,i),i=1,3) /13.4,9.4,21.0/
     data (X(5,i),i=1,3) /13.2,9.6,20.1/
     data (X(6,i),i=1,3) /13.9,10.4,19.8/
     data (X(7,i),i=1,3) /12.9,10.0,20.5/
     data (X(8,i),i=1,3) /12.2,9.9,20.7/
     data (X(9,i),i=1,3) /13.9,11.0,19.1/
     data ING /1,2,3,1,2,3,1,2,3/

     ISX=1
     TOL=1.d-5
     IFAIL=0

     PRINT *,X(1,:)
     PRINT *,ING

     CALL MATH_QTLMAP_G03ACF(WEIGHT,N,M,X,9,ISX,NX,ING,NG,WT,&
     NIG,CVM,3,E,3,NCV,CVX,3,TOL,IRX,IFAIL)

    WRITE (*,*)
    WRITE (*,99999) 'Rank of X = ', IRX
    WRITE (*,*)
           WRITE (*,*) 'Canonical     Eigenvalues Percentage    CHISQ       DF SIG'
           WRITE (*,*) 'Correlations              Variation'
           DO 60 I = 1, NCV
              WRITE (*,99998) (E(I,J),J=1,6)
   60      CONTINUE
           WRITE (*,*)
           WRITE (*,*) 'Canonical Coefficients for X'
           DO 80 I = 1, NX
              WRITE (*,99997) (CVX(I,J),J=1,NCV)
   80      CONTINUE
           WRITE (*,*)
           WRITE (*,*) 'Canonical variate means'
           DO 100 I = 1, NG
              WRITE (*,99997) (CVM(I,J),J=1,NCV)
  100      CONTINUE

       STOP

99999 FORMAT (1X,A,I2)
99998 FORMAT (1X,2F12.4,F11.4,F10.4,F8.1,F8.4)
99997 FORMAT (1X,5F9.4)

   end subroutine test_G03ACF


!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_G05EHF
!!  NAME
!!    MATH_QTLMAP_G05EHF
!!  DESCRIPTION
!!   Performs a pseudo-random permutation of a vector of integers.
!!  NOTES
!!   Use order-pack module m_ctrper :
!!   Permute an array randomly, but leaving elements close
!!   to their initial locations (nearbyness is controled by PCLS).
!!***
    subroutine MATH_QTLMAP_G05EHF(INDEX, N, IFAIL)
#ifndef HAVE_LIBNAG
      use m_ctrper
#endif
      INTEGER  , INTENT(IN)                 :: N
      INTEGER  , DIMENSION(N),INTENT(INOUT) :: INDEX
      INTEGER  , INTENT(INOUT)              :: IFAIL

#ifdef HAVE_LIBNAG
      external :: G05EHF
      call G05EHF(INDEX, N, IFAIL)
#else
      integer                               :: i
      IF ( N < 1 ) THEN
        call log_mess("MATH_QTLMAP_G05EHF Dimension ARRAY have to be greater or equal than 1.",ERROR_DEF)
        IFAIL=1
        RETURN
      END IF

      call ctrper(INDEX,1.0)
#endif

    end subroutine MATH_QTLMAP_G05EHF

!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_M01CAF
!!  NAME
!!    MATH_QTLMAP_M01CAF
!!  DESCRIPTION
!!   Rearranges a vector of real numbers into ascending or descending order.
!!  NOTES
!!  ORDERPACK
!!***
      subroutine MATH_QTLMAP_M01CAF(RV, M1, M2, ORDER, IFAIL)
#ifndef HAVE_LIBNAG
                use m_refsor
#endif
                integer , intent(in)         :: M1, M2
                integer , intent(out)        :: IFAIL
                double precision ,intent(inout)  :: RV(M2)
                character(len=1) ,intent(in) :: ORDER
#ifdef HAVE_LIBNAG
#else
                double precision , dimension(M2-M1+1):: DX,DY
                integer i
#endif

		          IF ( M1 < 1 ) THEN
		             IFAIL = 1
		             CALL log_mess('MATH_QTLMAP_M01CAF: M1 < 1 :'//str(M1),ERROR_DEF)
		             RETURN
		          END IF

		          IF ( M2 < 1 ) THEN
		             IFAIL = 1
		             CALL log_mess('MATH_QTLMAP_M01CAF: M2 < 1 :'//str(M2),ERROR_DEF)
		             RETURN
		          END IF

		          IF ( M1 > M2 ) THEN
		             IFAIL = 1
		             CALL log_mess('MATH_QTLMAP_M01CAF: M1 > M2  M1:'//str(M1)//' M2:'//str(M2),ERROR_DEF)
		             RETURN
		          END IF

		          IF ( ORDER /= 'A' .and. ORDER /= 'D' ) THEN
		             IFAIL = 2
		             CALL log_mess('MATH_QTLMAP_M01CAF: Bad value of ORDER (A,D):'//ORDER,ERROR_DEF)
		             RETURN
		         END IF

#ifdef HAVE_LIBNAG
                 call M01CAF(RV, M1, M2, ORDER, IFAIL)
#else
                 call refsor(RV(M1:M2))

                 IF ( ORDER == 'D' .or. ORDER == 'd' ) THEN
                     RV(M1:M2) = RV(M2:M1:-1)
                 END IF

#endif


      end subroutine MATH_QTLMAP_M01CAF

!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_M01DAF
!!  NAME
!!    MATH_QTLMAP_M01DAF
!!  DESCRIPTION
!!    Ranks a vector of real numbers in ascending or descending order.
!!  NOTES
!!   ORDERPACK
!!***
      subroutine MATH_QTLMAP_M01DAF(RV, M1, M2, ORDER, IRANK, IFAIL)
#ifndef HAVE_LIBNAG
          use m_mrgrnk
#endif
          integer , intent(in)             :: M1, M2
          integer , intent(out)            :: IRANK(M2)
          integer , intent(inout)          :: IFAIL
          double precision ,intent(in)     :: RV(M2)
          character(len=1) ,intent(in)     :: ORDER

#ifdef HAVE_LIBNAG
          external :: M01DAF
#ifdef DEBUG_PRINT
          PRINT * ,' --->  M01DAF ** NAG **'
#endif
          call M01DAF(RV, M1, M2, ORDER, IRANK, IFAIL)
#else
          double precision ,dimension(1:(M2-M1+1)) :: XDONT
          integer,dimension(M2) :: RANK
          integer :: i
#ifdef DEBUG_PRINT
          PRINT * ,' --->  IM01DAF'
#endif
          IFAIL = 0
          !On entry: elements M1 to M2 of RV must contain real values to be ranked.

          !Constraintes M1 > 0 and M2 >=M1
          if ( M1 <= 0 .or. M2 < M1 ) then
            IFAIL=1
            return
          end if
          !Constraintes ORDER='A' or 'D'
          if ( ORDER /= 'A' .and. ORDER /='D') then
            IFAIL=2
            return
          end if
         ! On exit: elements M1 to M2 of IRANK contain the ranks of the corresponding elements of RV.
         ! Note that the ranks are in the range M1 to M2: thus, if RV i is the first element in the rank order,
         ! IRANK i is set to M1.

         XDONT = RV(M1:M2)

         call mrgrnk (XDONT, RANK)

         if ( ORDER == 'D') then
            RANK(:) = RANK(M2-M1+1:1:-1)
         end if
         !Rank contient l'indice du rang d'un element du tableau
         do i=1,M2-M1+1
           IRANK(RANK(i)+M1-1)=i
         end do

#endif
      end subroutine MATH_QTLMAP_M01DAF


!****************************************************************************************************************
!!  NAME
!!    IM01DBF
!!  DESCRIPTION
!!    ranks a vector of integer numbers in ascending or descending order.
!!  NOTES
!!***
      subroutine MATH_QTLMAP_M01DBF (IV, M1, M2, ORDER, IRANK, IFAIL)
#ifndef HAVE_LIBNAG
          use m_mrgrnk
#endif
          integer , intent(in)             :: M1, M2
          integer , intent(out)            :: IRANK(M2)
          integer , intent(inout)          :: IFAIL
          integer , intent(in)             :: IV(M2)
          character(len=1) ,intent(in)     :: ORDER
#ifdef HAVE_LIBNAG
          external :: M01DBF
#ifdef DEBUG_PRINT
          PRINT * ,' --->  IM01DBF ** NAG **'
#endif
          call M01DBF(IV, M1, M2, ORDER, IRANK, IFAIL)
#else
          integer ,dimension(1:(M2-M1+1)) :: XDONT
          integer,dimension(M2) :: RANK
          integer :: i
#ifdef DEBUG_PRINT
          PRINT * ,' --->  IM01DBF'
#endif
          IFAIL = 0
          !On entry: elements M1 to M2 of RV must contain real values to be ranked.

          !Constraintes M1 > 0 and M2 >=M1
          if ( M1 <= 0 .or. M2 < M1 ) then
            IFAIL=1
            return
          end if
          !Constraintes ORDER='A' or 'D'
          if ( ORDER /= 'A' .and. ORDER /='D') then
            IFAIL=2
            return
          end if
         ! On exit: elements M1 to M2 of IRANK contain the ranks of the corresponding elements of RV.
         ! Note that the ranks are in the range M1 to M2: thus, if RV i is the first element in the rank order,
         ! IRANK i is set to M1.

         XDONT = IV(M1:M2)

         call mrgrnk (XDONT, RANK)

         if ( ORDER == 'D') then
            RANK(:) = RANK(M2-M1+1:1:-1)
         end if
         IRANK=0

         !Rank contient l'indice du rang d'un element du tableau
         do i=1,M2-M1+1
          IRANK(RANK(i)+M1-1)=i
         end do

#endif
      end subroutine MATH_QTLMAP_M01DBF


Subroutine char_mrgrnk (XDONT, IRNGT)
! __________________________________________________________
!   MRGRNK = Merge-sort ranking of an array
!   For performance reasons, the first 2 passes are taken
!   out of the standard loop, and use dedicated coding.
! __________________________________________________________
! __________________________________________________________
      character(len=*), Dimension (:), Intent (In)  :: XDONT
      Integer, Dimension (:), Intent (Out) :: IRNGT
! __________________________________________________________
      character(len=256) :: XVALA, XVALB
!
      Integer, Dimension (SIZE(IRNGT)) :: JWRKT
      Integer :: LMTNA, LMTNC, IRNG1, IRNG2
      Integer :: NVAL, IIND, IWRKD, IWRK, IWRKF, JINDA, IINDA, IINDB
!
      NVAL = Min (SIZE(XDONT), SIZE(IRNGT))
      Select Case (NVAL)
      Case (:0)
         Return
      Case (1)
         IRNGT (1) = 1
         Return
      Case Default
         Continue
      End Select
!
!  Fill-in the index array, creating ordered couples
!
      Do IIND = 2, NVAL, 2
         If (lle(XDONT(IIND-1),XDONT(IIND))) Then
            IRNGT (IIND-1) = IIND - 1
            IRNGT (IIND) = IIND
         Else
            IRNGT (IIND-1) = IIND
            IRNGT (IIND) = IIND - 1
         End If
      End Do
      If (Modulo(NVAL, 2) /= 0) Then
         IRNGT (NVAL) = NVAL
      End If
!
!  We will now have ordered subsets A - B - A - B - ...
!  and merge A and B couples into     C   -   C   - ...
!
      LMTNA = 2
      LMTNC = 4
!
!  First iteration. The length of the ordered subsets goes from 2 to 4
!
      Do
         If (NVAL <= 2) Exit
!
!   Loop on merges of A and B into C
!
         Do IWRKD = 0, NVAL - 1, 4
            If ((IWRKD+4) > NVAL) Then
               If ((IWRKD+2) >= NVAL) Exit
!
!   1 2 3
!
               If (lle(XDONT(IRNGT(IWRKD+2)),XDONT(IRNGT(IWRKD+3)))) Exit
!
!   1 3 2
!
               If (lle(XDONT(IRNGT(IWRKD+1)),XDONT(IRNGT(IWRKD+3)))) Then
                  IRNG2 = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNG2
!
!   3 1 2
!
               Else
                  IRNG1 = IRNGT (IWRKD+1)
                  IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+2)
                  IRNGT (IWRKD+2) = IRNG1
               End If
               Exit
            End If
!
!   1 2 3 4
!
            If (lle(XDONT(IRNGT(IWRKD+2)),XDONT(IRNGT(IWRKD+3)))) Cycle
!
!   1 3 x x
!
            If (lle(XDONT(IRNGT(IWRKD+1)),XDONT(IRNGT(IWRKD+3)))) Then
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+2) = IRNGT (IWRKD+3)
               If (lle(XDONT(IRNG2),XDONT(IRNGT(IWRKD+4)))) Then
!   1 3 2 4
                  IRNGT (IWRKD+3) = IRNG2
               Else
!   1 3 4 2
                  IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+4) = IRNG2
               End If
!
!   3 x x x
!
            Else
               IRNG1 = IRNGT (IWRKD+1)
               IRNG2 = IRNGT (IWRKD+2)
               IRNGT (IWRKD+1) = IRNGT (IWRKD+3)
               If (lle(XDONT(IRNG1),XDONT(IRNGT(IWRKD+4)))) Then
                  IRNGT (IWRKD+2) = IRNG1
                  If (lle(XDONT(IRNG2),XDONT(IRNGT(IWRKD+4)))) Then
!   3 1 2 4
                     IRNGT (IWRKD+3) = IRNG2
                  Else
!   3 1 4 2
                     IRNGT (IWRKD+3) = IRNGT (IWRKD+4)
                     IRNGT (IWRKD+4) = IRNG2
                  End If
               Else
!   3 4 1 2
                  IRNGT (IWRKD+2) = IRNGT (IWRKD+4)
                  IRNGT (IWRKD+3) = IRNG1
                  IRNGT (IWRKD+4) = IRNG2
               End If
            End If
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 4
         Exit
      End Do
!
!  Iteration loop. Each time, the length of the ordered subsets
!  is doubled.
!
      Do
         If (LMTNA >= NVAL) Exit
         IWRKF = 0
         LMTNC = 2 * LMTNC
!
!   Loop on merges of A and B into C
!
         Do
            IWRK = IWRKF
            IWRKD = IWRKF + 1
            JINDA = IWRKF + LMTNA
            IWRKF = IWRKF + LMTNC
            If (IWRKF >= NVAL) Then
               If (JINDA >= NVAL) Exit
               IWRKF = NVAL
            End If
            IINDA = 1
            IINDB = JINDA + 1
!
!   Shortcut for the case when the max of A is smaller
!   than the min of B. This line may be activated when the
!   initial set is already close to sorted.
!
!          IF (XDONT(IRNGT(JINDA)) <= XDONT(IRNGT(IINDB))) CYCLE
!
!  One steps in the C subset, that we build in the final rank array
!
!  Make a copy of the rank array for the merge iteration
!
            JWRKT (1:LMTNA) = IRNGT (IWRKD:JINDA)
!
            XVALA = XDONT (JWRKT(IINDA))
            XVALB = XDONT (IRNGT(IINDB))
!
            Do
               IWRK = IWRK + 1
!
!  We still have unprocessed values in both A and B
!
               If (XVALA > XVALB) Then
                  IRNGT (IWRK) = IRNGT (IINDB)
                  IINDB = IINDB + 1
                  If (IINDB > IWRKF) Then
!  Only A still with unprocessed values
                     IRNGT (IWRK+1:IWRKF) = JWRKT (IINDA:LMTNA)
                     Exit
                  End If
                  XVALB = XDONT (IRNGT(IINDB))
               Else
                  IRNGT (IWRK) = JWRKT (IINDA)
                  IINDA = IINDA + 1
                  If (IINDA > LMTNA) Exit! Only B still with unprocessed values
                  XVALA = XDONT (JWRKT(IINDA))
               End If
!
            End Do
         End Do
!
!  The Cs become As and Bs
!
         LMTNA = 2 * LMTNA
      End Do
!
      Return
!
End Subroutine char_mrgrnk

!****************************************************************************************************************
!!  NAME
!!    IM01DCF
!!  DESCRIPTION
!!    Ranks a vector of character data in ASCII or reverse ASCII order of a specified substring.
!!
!!******************************************
      subroutine MATH_QTLMAP_M01DCF(CH, M1, M2, L1, L2, ORDER, IRANK, IFAIL)
                   INTEGER        M1, M2, L1, L2, IRANK(M2), IFAIL
                   CHARACTER*(*)  CH(M2)
                   CHARACTER*1    ORDER
#ifdef HAVE_LIBNAG
          external :: M01DCF
#ifdef DEBUG_PRINT
          PRINT * ,' --->  IM01DBF ** NAG **'
#endif
          call M01DCF(CH, M1, M2, L1, L2, ORDER, IRANK, IFAIL)
#else
          CHARACTER(LEN=256) ,dimension(1:(M2-M1+1)) :: XDONT
          integer,dimension(M2) :: RANK
          integer :: i
           IFAIL = 0
          !On entry: elements M1 to M2 of RV must contain real values to be ranked.

          !Constraintes M1 > 0 and M2 >=M1
          if ( M1 <= 0 .or. M2 < M1 ) then
            IFAIL=1
            return
          end if

          IF ( (L2<1) .or. (L1<1) .or. (L1>L2) ) THEN
            IFAIL=1
            RETURN
          END IF

          IF ( SIZE(CH) >= 1) THEN
              IF ( L2 > LEN(CH(1))) THEN
                IFAIL=1
                RETURN
              END IF
          END IF
          !Constraintes ORDER='A' or 'D'
          if ( ORDER /= 'A' .and. ORDER /='R') then
            IFAIL=2
            return
          end if
          DO I=M1,M2
              IF ( len(CH(I)) > 255 ) THEN
                IFAIL=3
                RETURN
              END IF
          END DO

         ! On exit: elements M1 to M2 of IRANK contain the ranks of the corresponding elements of RV.
         ! Note that the ranks are in the range M1 to M2: thus, if RV i is the first element in the rank order,
         ! IRANK i is set to M1.

         DO I=M1,M2
             XDONT(I-M1+1)=CH(I)(L1:L2)
         END DO

         call char_mrgrnk (XDONT, RANK)

         if ( ORDER == 'R') then
            RANK(:) = RANK(M2-M1+1:1:-1)
         end if
         IRANK=0

         !Rank contient l'indice du rang d'un element du tableau
         do i=1,M2-M1+1
          IRANK(RANK(i)+M1-1)=i
         end do

#endif

      end subroutine MATH_QTLMAP_M01DCF


!! Sample
!! Data Ranks
!!  34    9
!!  44    8
!!  89    2
!!  64    7
!!  69    4
!!  69    5
!!  23   10
!!   1   12
!! 999    1
!!  65    6
!!  22   11
!!  76    3

!****************************************************************************************************************
!!  NAME
!!    IM01ZAF
!!  DESCRIPTION
!!    Inverts a permutation, and hence converts a rank vector to an index vector, or vice versa.
!!    There are two common ways of describing a permutation using an integer vector IPERM.
!!    The first uses ranks: IPERMi holds the position to which the ith data element should be moved in order to sort the data; in other words its rank in the sorted order.
!!    The second uses indices: IPERMi holds the current position of the data element which would occur in ith position in sorted order.
!!    For example, given the values
!!     3.5 5.9 2.9 0.5
!!
!!     to be sorted in ascending order, the ranks would be
!!     3 4 2 1
!!
!!     and the indices would be
!!     4 3 1 2
!!
!!  PARAMETERS
!!
!! IPERM(M2) – INTEGER arrayInput/Output
!!    On entry: elements M1 to M2 of IPERM must contain a permutation of the integers M1 to M2.
!!
!!    On exit: these elements contain the inverse permutation of the integers M1 to M2.
!!
!!    M1 and M2 must specify the range of elements used in the array IPERM and the range of values in the permutation, as specified under IPERM.
!!
!!    Constraint:
!!      0<M1≤M2.
!!
!!***
      subroutine MATH_QTLMAP_M01ZAF ( IPERM, M1, M2, IFAIL)
         INTEGER                 , INTENT(IN)    :: M1, M2
         INTEGER  ,DIMENSION(M2) , INTENT(INOUT) :: IPERM
         INTEGER                   , INTENT(OUT) :: IFAIL
#ifdef HAVE_LIBNAG
          external :: M01ZAF
          INTEGER                                 :: I
#ifdef DEBUG_PRINT
          PRINT * ,' --->  M01ZAF ** NAG **'
#endif

          call M01ZAF ( IPERM, M1, M2, IFAIL)
#ifdef DEBUG_PRINT

          PRINT *,"************  ENV NAG *****************"
          DO I=1,M2
            PRINT *,'IPERM(',I,')=',IPERM(I)
          END DO
#endif
#else
         LOGICAL ,DIMENSION(M2)  :: PRES
         INTEGER ,DIMENSION(M2)  :: INDEX_RANK
         INTEGER                 :: I,J
#ifdef DEBUG_PRINT
          PRINT * ,' --->  IM01ZAF'
#endif

         IF ( (M2<1) .or. (M1<1) .or. (M1>M2) ) THEN
           IFAIL=1
           RETURN
         END IF

         PRES=.false.

         DO I=M1,M2
           IF ( (IPERM(I)>M2) .or. IPERM(I)<M1  ) THEN
             IFAIL = 2
             RETURN
           END IF
           IF ( PRES(IPERM(I)) ) THEN
             IFAIL = 3
             RETURN
           END IF
           PRES(IPERM(I)) = .true.
         END DO

         INDEX_RANK=IPERM
         DO I=M1,M2 ! for each rank
          ! find the index of IPERM that contains the minimum
           DO J=M1,M2
            IF ( I == IPERM(J) ) THEN
                INDEX_RANK(I) = J
                EXIT
            END IF
           END DO
         END DO

        IPERM=INDEX_RANK

#ifdef DEBUG_PRINT
        PRINT *,"*******  NO ENV NAG *****************"
        DO I=1,M2
          PRINT *,'IPERM(',I,')=',IPERM(I)
        END DO
#endif

#endif
      end subroutine MATH_QTLMAP_M01ZAF


!****************************************************************************************************************
!!  NAME
!!    IM01DFF
!!  DESCRIPTION
!!     ranks rows M1 to M2 of a matrix, using the data in columns N1 to N2 of those rows. The
!!    ordering is determined by first ranking the data in column N1, then ranking any tied rows according to the
!!    data in column N1 + 1, and so on up to column N2.


      subroutine MATH_QTLMAP_M01DFF(IM, LDM, M1, M2, N1, N2, ORDER, IRANK, IFAIL)
        INTEGER                     , INTENT(IN)    :: LDM,M1,M2,N1,N2
        INTEGER   ,DIMENSION(LDM,N2), INTENT(IN)    :: IM ! column of N1 to N2 of rows M1 to M2 of IM contains INTEGER to be ranked
        CHARACTER(LEN=1)            , INTENT(IN)    :: ORDER  ! 'A' ascending or 'D' descending order
        INTEGER   ,DIMENSION(M2), INTENT(OUT)       :: IRANK  ! elements M1 to M2 of IRANK contain the ranks of the corresponding rows of IM.
!  Note that the ranks are in the range M1 to M2: thus, if the ith row of IM is the first in the rank order,
!  IRANK(i) is set to M1.

        INTEGER                 , INTENT(OUT)       :: IFAIL
#ifdef HAVE_LIBNAG
          external :: M01DFF
#ifdef DEBUG_PRINT
            PRINT *," --> M01DFF ** nag **"
#endif
          call M01DFF(IM, LDM, M1, M2, N1, N2, ORDER, IRANK, IFAIL)
#ifdef DEBUG_PRINT
            PRINT *,"*********** ENV NAG ***********"
            DO I=M1,M2
              PRINT *,'IRANK(',I,')=',IRANK(I)
            END DO
#endif
#else
          INTEGER , DIMENSION(N1:N2,M2) :: INT_IRANK,IIRANK
          INTEGER                       :: I,J,K,IND,IND2,CURRENT_RANK,AFFECTED(M1:M2)
          LOGICAL                       :: FILTERIN(M1:M2),FILTEROUT(M1:M2)
#ifdef DEBUG_PRINT
            PRINT *," --> IM01DFF"
#endif

          IF ( (M2<1).or.(N2<1).or.(M1<1).or.(M1>M2).or.(N1<1).or.(N1>N2).or.(LDM<M2)) THEN
           IFAIL = 1
           RETURN
          END IF

          IF ( ORDER /= 'A' .and. ORDER /= 'a' .and. ORDER /= 'D' .and. ORDER /= 'd') THEN
           IFAIL = 2
           RETURN
          END IF

          ! Build rank vector for each column

          DO I=N1,N2
            CALL MATH_QTLMAP_M01DBF (IM(:,I), M1, M2, ORDER, INT_IRANK(I,:), IFAIL)
            IF ( IFAIL /= 0 ) THEN
               PRINT *,'IM01DFF-->IM01DBF IFAIL=',IFAIL
               RETURN
            END IF
#ifdef DEBUG_PRINT
            PRINT *,"IM01DBF WITH COLUMN:",I
            DO J=M1,M2
              PRINT *,'INT_IRANK(',J,')=',INT_IRANK(I,J)
            END DO
#endif
          END DO ! END I

         !Table of index rank
         DO I=N1,N2
           DO J=M1,M2
             IIRANK(I,INT_IRANK(I,J))=J
           END DO
         END DO



          IRANK=0
          FILTERIN=.true.
          FILTEROUT=.false.
          DO CURRENT_RANK=M1,M2

           DO J=N1,N2
              IF ( .not. is_conflict(LDM,M1,M2,N1,N2,J,FILTERIN,FILTEROUT,IM,INT_IRANK,IIRANK) .or. J==N2)  THEN
               !find the rows
               DO I=M1,M2
                IF ( FILTEROUT(I) ) THEN
                 EXIT
                END IF
               END DO
               !set the rank
               IRANK(I)=CURRENT_RANK
               AFFECTED(CURRENT_RANK)=I
               FILTERIN=.true.
               DO I=M1,CURRENT_RANK
                 FILTERIN(AFFECTED(I))=.false.
               END DO
               EXIT

              ELSE ! THE FILTER OUT contains several TRUE value, we have to search in the next row
                 !initializing the filterin
                 FILTERIN=FILTEROUT
              END IF
            END DO ! J
          END DO  !I

#ifdef DEBUG_PRINT
            PRINT *,"*********** NO ENV NAG ***********"
            DO I=M1,M2
              PRINT *,'IRANK(',I,')=',IRANK(I)
            END DO
#endif

#endif
      end subroutine MATH_QTLMAP_M01DFF

               ! Get the next bolean colum in the jth column from a boolean table (jth -1 column)
               ! 3 2 1  for the rank=1  ==> false  ....  ....
               ! 2 2 1                      true   true  true
               ! 2 2 2                      true   true  ...
               ! 2 3 1                      true   false  ...
      function is_conflict(LDM,M1,M2,N1,N2,J,tabtocomp,nexttocomp,IM,IIRANK,INDEXIRANK) result(RET)
            integer                   , INTENT(IN)      :: M1,M2,J,LDM,N1,N2
            logical,dimension(M1:M2)  , INTENT(IN)      :: tabtocomp
            logical,dimension(M1:M2)  , INTENT(OUT)     :: nexttocomp
            INTEGER   ,DIMENSION(LDM,N2), INTENT(IN)    :: IM
            INTEGER   ,DIMENSION(N1:N2,M2), INTENT(IN)     :: IIRANK,INDEXIRANK

            INTEGER :: I,IND,MINV,TABIND(count(tabtocomp)),SUBIIRANK(count(tabtocomp))
            LOGICAL :: RET
#ifdef DEBUG_PRINT
            PRINT *,'** CREATE SUBTAB ** SIZE:',SIZE(TABIND)
#endif
            IND=1
            DO I=M1,M2
              IF (tabtocomp(I)) THEN
#ifdef DEBUG_PRINT
                PRINT *,'ADD ROW:',I,' RANK INDEX:',IIRANK(J,I)
#endif
                TABIND(IND)=I
                SUBIIRANK(IND)=IIRANK(J,I)
                IND = IND+1
              END IF
            END DO

            !SEARCH THE MINIMUM RANK
            minv=minval(SUBIIRANK)
#ifdef DEBUG_PRINT
            PRINT *,'RANK VALMIN:',minv,' ROWS:',INDEXIRANK(J,minv),' corresponding to the value:',IM(INDEXIRANK(J,minv),J)
#endif
            !CREATE NEW TAB FILTER BOOLEAN
            nexttocomp=.false.
            DO I=1,SIZE(TABIND)
              IND=TABIND(I)
              IF ( IM(IND,J) == IM(INDEXIRANK(J,minv),J) ) THEN
#ifdef DEBUG_PRINT
                 PRINT *,'-->KEEP ROW:',IND
#endif
                 nexttocomp(IND)=.true.
              END IF
            END DO

            !IF ONE VALUE=>NO CONFLICT,WE FIND THE ROW
            IF ( count(nexttocomp) == 1 ) THEN
               RET=.false.
            ELSE
               RET=.true.
            END IF

            RETURN

      end function is_conflict

!****************************************************************************************************************

!!****f* m_qtlmap_math/MATH_QTLMAP_F03ABF
!!  NAME
!!    MATH_QTLMAP_F03ABF
!!  DESCRIPTION
!!    Calculates the determinant of a real symmetric positive-definite matrix using a Cholesky factorization.
!!  NOTES
!!
!!***
      subroutine MATH_QTLMAP_F03ABF(A, IA, N, DET, IFAIL)
          integer      ,intent(in)                   :: IA,N
          double precision,dimension(IA,N), intent(in)  :: A
          double precision,intent(out)                  :: DET
          integer      ,intent(out)                  :: IFAIL

#ifdef HAVE_LIBNAG
          double precision,dimension(N)     :: WKSPCE
#else
          double precision,dimension(IA,N)  :: APRIM
          double precision,dimension(N)     :: eigenv
          double precision                  :: trace, sume
          double precision,parameter        :: TOL = 0.001d0

          IFAIL = 0
#endif
          DET=0.d0
#ifdef HAVE_LIBNAG
          call F03ABF(A, IA, N, DET, WKSPCE, IFAIL)
#else

          APRIM=A
          call kaiser(APRIM, IA, N, eigenv, trace, sume, IFAIL)

          ! Probleme de precision....
          if (trace == 0.d0) then
             IFAIL=3
             DET=0.d0
             RETURN
          end if
          IF ( abs((trace-sume)/trace)>TOL ) THEN
             DET = product(eigenv)
             IFAIL= 3
          END IF

          IF ( IFAIL == 0 ) THEN
            DET = product(eigenv)
          ELSE
            DET = 0.d0
          END IF

#endif
#ifdef DEBUG
          call log_mess('MATH_QTLMAP_F03ABF (OUT) ==== DET:'//str(DET),DEBUG_DEF)
#endif
      end subroutine MATH_QTLMAP_F03ABF

!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_F01ADF
!!  NAME
!!    MATH_QTLMAP_F01ADF
!!  DESCRIPTION
!!   Calculates the approximate inverse of a real symmetric positive-definite matrix, using a Cholesky
!!   factorization.
!!  NOTES
!!   Use DPOFA,DPODI : SLATEC
!!***
       subroutine MATH_QTLMAP_F01ADF (N, A, IA, IFAIL)
          integer         ,intent(in)                     :: IA,N
          double precision,dimension(IA,N) ,intent(inout) :: A
          integer         ,intent(out)                    :: IFAIL
#ifdef HAVE_LIBNAG

#else
          double precision,dimension(IA,N)  :: APRIM
          double precision,dimension(2)     :: DET2
          integer                           :: INFO,I,J
          external                          :: DPOFA,DPODI ! SLATEC
          IFAIL=0
#endif

#ifdef HAVE_LIBNAG
         call F01ADF(N, A, IA, IFAIL)
#else
          APRIM(1:N,1:N)=A(1:N,1:N)
          CALL DPOFA(APRIM, IA, N, INFO)
          IF ( INFO /= 0 ) THEN
              IFAIL=1
              !CALL log_mess('MATH_QTLMAP_F01ADF --> Can not compute inverse of matrix',ERROR_DEF)
              RETURN
          END IF
          CALL DPODI(APRIM,IA,N,DET2,01)

          DO I=2,IA
            DO J=1,(I-1)
                A(I,J)=APRIM(J,(I-1))
            END DO
          END DO
#endif
       end subroutine MATH_QTLMAP_F01ADF

!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_INVDETMATSYM
!!  NAME
!!    MATH_QTLMAP_F01ADF
!!  DESCRIPTION
!!   Calculates the inverse and the derterminant of a real symmetric positive-definite matrix
!!  NOTES
!!   Use DPOFA,DPODI : SLATEC
!!***
       subroutine MATH_QTLMAP_INVDETMATSYM (N, A, IA,DET, IFAIL)
          integer         ,intent(in)                     :: IA,N
          double precision,dimension(IA,N) ,intent(inout) :: A
          double precision                   ,intent(out) :: DET
          integer         ,intent(out)                    :: IFAIL

          double precision,dimension(IA,N)  :: APRIM
          double precision,dimension(2)     :: DET2
          integer                           :: INFO,I,J
          external                          :: DPOFA,DPODI ! SLATEC
          IFAIL=0

          APRIM(1:N,1:N)=A(1:N,1:N)
          CALL DPOFA(APRIM, IA, N, INFO)
          IF ( INFO /= 0 ) THEN
              IFAIL=1
              !CALL log_mess('MATH_QTLMAP_F01ADF --> Can not compute inverse of matrix',ERROR_DEF)
              RETURN
          END IF
          CALL DPODI(APRIM,IA,N,DET2,11)
          DET = DET2(1) * 10.d0**DET2(2)
          DO I=2,IA
            DO J=1,(I-1)
                A(I,J)=APRIM(J,(I-1))
            END DO
          END DO

       end subroutine MATH_QTLMAP_INVDETMATSYM

!!****f* m_qtlmap_math/MATH_QTLMAP_S15AEF
!!  NAME
!!    MATH_QTLMAP_S15AEF
!!  DESCRIPTION
!!   Returns the value of the error function erf x, via the routine name.
!!  NOTES
!!   Use DERF from http://www.kurims.kyoto-u.ac.jp/~ooura/gamerf.html
!!***
    function MATH_QTLMAP_S15AEF(X,IFAIL) result(y)
        double precision,intent(in)  :: X
        integer, intent(inout)       :: IFAIL
        double precision             :: y

#ifdef HAVE_LIBNAG
        double precision,external    :: S15AEF
         y = S15AEF(X, IFAIL)
#else
         y = derf(X)
#endif
    end function MATH_QTLMAP_S15AEF
!****************************************************************************************************************
!!****f* m_qtlmap_math/MATH_QTLMAP_S15ADF
!!  NAME
!!    MATH_QTLMAP_S15ADF
!!  DESCRIPTION
!!   Returns the value of the complementary error function, erfc x, via the routine name.
!!  NOTES
!!   Use DERFC from http://www.kurims.kyoto-u.ac.jp/~ooura/gamerf.html
!!***
    function MATH_QTLMAP_S15ADF(X,IFAIL) result(y)
        double precision,intent(in)  :: X
        integer, intent(inout)       :: IFAIL
        double precision             :: y
#ifdef HAVE_LIBNAG
        double precision,external    :: S15ADF
         y = S15ADF(X, IFAIL)
#else
         y = derfc(X)
#endif
    end function MATH_QTLMAP_S15ADF

!****************************************************************************************************************

       function MATH_QTLMAP_GAMMA_F(N) result(R)
            integer,intent(in)   :: N
            integer :: P,R

            P = N-1
            R = P
            do while ( P > 2 )
              P = P - 1
              R = R*P
            end do

       end function MATH_QTLMAP_GAMMA_F

! ren-raw chen, rutgers business school
! normal inverse
! translate from
!http://home.online.no/~pjacklam/notes/invnorm
! a routine written by john herrero
      function dinvnorm(p) result (X)
      double precision,intent(in) :: p
      double precision :: p_low,p_high
      double precision :: a1,a2,a3,a4,a5,a6
      double precision :: b1,b2,b3,b4,b5
      double precision :: c1,c2,c3,c4,c5,c6
      double precision :: d1,d2,d3,d4
      double precision :: z,q,r,X

      a1=-39.6968302866538
      a2=220.946098424521
      a3=-275.928510446969
      a4=138.357751867269
      a5=-30.6647980661472
      a6=2.50662827745924
      b1=-54.4760987982241
      b2=161.585836858041
      b3=-155.698979859887
      b4=66.8013118877197
      b5=-13.2806815528857
      c1=-0.00778489400243029
      c2=-0.322396458041136
      c3=-2.40075827716184
      c4=-2.54973253934373
      c5=4.37466414146497
      c6=2.93816398269878
      d1=0.00778469570904146
      d2=0.32246712907004
      d3=2.445134137143
      d4=3.75440866190742
      p_low=0.02425
      p_high=1-p_low
      if(p.lt.p_low) goto 201
      if(p.ge.p_low) goto 301
201   q=dsqrt(-2*dlog(p))
      z=(((((c1*q+c2)*q+c3)*q+c4)*q+c5)*q+c6)/ &
     ((((d1*q+d2)*q+d3)*q+d4)*q+1)
      goto 204
301   if((p.ge.p_low).and.(p.le.p_high)) goto 202
      if(p.gt.p_high) goto 302
202   q=p-0.5
      r=q*q
      z=(((((a1*r+a2)*r+a3)*r+a4)*r+a5)*r+a6)*q/ &
     (((((b1*r+b2)*r+b3)*r+b4)*r+b5)*r+1)
      goto 204
302   if((p.gt.p_high).and.(p.lt.1)) goto 203
203   q=dsqrt(-2*dlog(1-p))
      z=-(((((c1*q+c2)*q+c3)*q+c4)*q+c5)*q+c6)/  &
     ((((d1*q+d2)*q+d3)*q+d4)*q+1)
204   X=z
      return
      end function dinvnorm


! error function in double precision
! http://www.kurims.kyoto-u.ac.jp/~ooura/gamerf.html
      function derf(x) result(P)
       double precision  ,intent(in)      :: x
       double precision , dimension(0:64) :: a, b
       double precision                   :: w,t,y,P
       integer                            :: i,k

      data (a(i), i = 0, 12) /                                 &
         0.00000000005958930743d0, -0.00000000113739022964d0,  &
         0.00000001466005199839d0, -0.00000016350354461960d0,  &
         0.00000164610044809620d0, -0.00001492559551950604d0,  &
         0.00012055331122299265d0, -0.00085483269811296660d0,  &
         0.00522397762482322257d0, -0.02686617064507733420d0,  &
         0.11283791670954881569d0, -0.37612638903183748117d0,  &
         1.12837916709551257377d0 /
      data (a(i), i = 13, 25) /                                &
         0.00000000002372510631d0, -0.00000000045493253732d0,  &
         0.00000000590362766598d0, -0.00000006642090827576d0,  &
         0.00000067595634268133d0, -0.00000621188515924000d0,  &
         0.00005103883009709690d0, -0.00037015410692956173d0,  &
         0.00233307631218880978d0, -0.01254988477182192210d0,  &
         0.05657061146827041994d0, -0.21379664776456006580d0,  &
         0.84270079294971486929d0 /
      data (a(i), i = 26, 38) /                                &
         0.00000000000949905026d0, -0.00000000018310229805d0,  &
         0.00000000239463074000d0, -0.00000002721444369609d0,  &
         0.00000028045522331686d0, -0.00000261830022482897d0,  &
         0.00002195455056768781d0, -0.00016358986921372656d0,  &
         0.00107052153564110318d0, -0.00608284718113590151d0,  &
         0.02986978465246258244d0, -0.13055593046562267625d0,  &
         0.67493323603965504676d0 /
      data (a(i), i = 39, 51) /                                &
         0.00000000000382722073d0, -0.00000000007421598602d0,  &
         0.00000000097930574080d0, -0.00000001126008898854d0,  &
         0.00000011775134830784d0, -0.00000111992758382650d0,  &
         0.00000962023443095201d0, -0.00007404402135070773d0,  &
         0.00050689993654144881d0, -0.00307553051439272889d0,  &
         0.01668977892553165586d0, -0.08548534594781312114d0,  &
         0.56909076642393639985d0 /
      data (a(i), i = 52, 64) /                                &
         0.00000000000155296588d0, -0.00000000003032205868d0,  &
         0.00000000040424830707d0, -0.00000000471135111493d0,  &
         0.00000005011915876293d0, -0.00000048722516178974d0,  &
         0.00000430683284629395d0, -0.00003445026145385764d0,  &
         0.00024879276133931664d0, -0.00162940941748079288d0,  &
         0.00988786373932350462d0, -0.05962426839442303805d0,  &
         0.49766113250947636708d0 /
      data (b(i), i = 0, 12) /                                 &
         -0.00000000029734388465d0, 0.00000000269776334046d0,  &
         -0.00000000640788827665d0, -0.00000001667820132100d0, &
         -0.00000021854388148686d0, 0.00000266246030457984d0,  &
         0.00001612722157047886d0, -0.00025616361025506629d0,  &
         0.00015380842432375365d0, 0.00815533022524927908d0,  &
         -0.01402283663896319337d0, -0.19746892495383021487d0,  &
         0.71511720328842845913d0 /
      data (b(i), i = 13, 25) /                                  &
         -0.00000000001951073787d0, -0.00000000032302692214d0,   &
         0.00000000522461866919d0, 0.00000000342940918551d0,     &
         -0.00000035772874310272d0, 0.00000019999935792654d0,    &
         0.00002687044575042908d0, -0.00011843240273775776d0,    &
         -0.00080991728956032271d0, 0.00661062970502241174d0,    &
         0.00909530922354827295d0, -0.20160072778491013140d0,    &
         0.51169696718727644908d0 /
      data (b(i), i = 26, 38) /                                  &
         0.00000000003147682272d0, -0.00000000048465972408d0,    &
         0.00000000063675740242d0, 0.00000003377623323271d0,     &
         -0.00000015451139637086d0, -0.00000203340624738438d0,   &
         0.00001947204525295057d0, 0.00002854147231653228d0,     &
         -0.00101565063152200272d0, 0.00271187003520095655d0,    &
         0.02328095035422810727d0, -0.16725021123116877197d0,    &
         0.32490054966649436974d0 /
      data (b(i), i = 39, 51) /                                  &
         0.00000000002319363370d0, -0.00000000006303206648d0,    &
         -0.00000000264888267434d0, 0.00000002050708040581d0,    &
         0.00000011371857327578d0, -0.00000211211337219663d0,    &
         0.00000368797328322935d0, 0.00009823686253424796d0,     &
         -0.00065860243990455368d0, -0.00075285814895230877d0,   &
         0.02585434424202960464d0, -0.11637092784486193258d0,    &
         0.18267336775296612024d0 /
      data (b(i), i = 52, 64) /                                  &
         -0.00000000000367789363d0, 0.00000000020876046746d0,    &
         -0.00000000193319027226d0, -0.00000000435953392472d0,   &
         0.00000018006992266137d0, -0.00000078441223763969d0,    &
         -0.00000675407647949153d0, 0.00008428418334440096d0,    &
         -0.00017604388937031815d0, -0.00239729611435071610d0,   &
         0.02064129023876022970d0, -0.06905562880005864105d0,    &
         0.09084526782065478489d0 /
      w = abs(x)
      if (w .lt. 2.2d0) then
          t = w * w
          k = int(t)
          t = t - k
          k = k * 13
          y = ((((((((((((a(k) * t + a(k + 1)) * t +              &
             a(k + 2)) * t + a(k + 3)) * t + a(k + 4)) * t +      &
             a(k + 5)) * t + a(k + 6)) * t + a(k + 7)) * t +       &
             a(k + 8)) * t + a(k + 9)) * t + a(k + 10)) * t +     &
             a(k + 11)) * t + a(k + 12)) * w
      else if (w .lt. 6.9d0) then
          k = int(w)
          t = w - k
          k = 13 * (k - 2)
          y = (((((((((((b(k) * t + b(k + 1)) * t + &
             b(k + 2)) * t + b(k + 3)) * t + b(k + 4)) * t + &
             b(k + 5)) * t + b(k + 6)) * t + b(k + 7)) * t + &
             b(k + 8)) * t + b(k + 9)) * t + b(k + 10)) * t + &
             b(k + 11)) * t + b(k + 12)
          y = y * y
          y = y * y
          y = y * y
          y = 1 - y * y
      else
          y = 1
      end if
      if (x .lt. 0) y = -y
      P = y
      end function derf
!

      ! complementary error function in double precision
!! http://www.kurims.kyoto-u.ac.jp/~ooura/gamerf.html
      function derfc(x) result(y)
      double precision , intent(in) :: x
      double precision              :: y,t,u

      double precision,parameter :: pa = 3.97886080735226000d+00
         double precision,parameter ::  p0 = 2.75374741597376782d-01
         double precision,parameter ::  p1 = 4.90165080585318424d-01
         double precision,parameter ::  p2 = 7.74368199119538609d-01
         double precision,parameter ::  p3 = 1.07925515155856677d+00
         double precision,parameter ::  p4 = 1.31314653831023098d+00
         double precision,parameter ::  p5 = 1.37040217682338167d+00
         double precision,parameter ::  p6 = 1.18902982909273333d+00
         double precision,parameter ::  p7 = 8.05276408752910567d-01
         double precision,parameter ::  p8 = 3.57524274449531043d-01
         double precision,parameter ::  p9 = 1.66207924969367356d-02
         double precision,parameter ::  p10 = -1.19463959964325415d-01
         double precision,parameter ::  p11 = -8.38864557023001992d-02

         double precision,parameter ::  p12 = 2.49367200053503304d-03
         double precision,parameter ::  p13 = 3.90976845588484035d-02
         double precision,parameter ::  p14 = 1.61315329733252248d-02
         double precision,parameter ::  p15 = -1.33823644533460069d-02
         double precision,parameter ::  p16 = -1.27223813782122755d-02
         double precision,parameter ::  p17 = 3.83335126264887303d-03
         double precision,parameter ::  p18 = 7.73672528313526668d-03
         double precision,parameter ::  p19 = -8.70779635317295828d-04
         double precision,parameter ::  p20 = -3.96385097360513500d-03
         double precision,parameter ::  p21 = 1.19314022838340944d-04
         double precision,parameter ::  p22 = 1.27109764952614092d-03

      t = pa / (pa + abs(x))
      u = t - 0.5d0
      y = (((((((((p22 * u + p21) * u + p20) * u +         &
         p19) * u + p18) * u + p17) * u + p16) * u +       &
         p15) * u + p14) * u + p13) * u + p12
      y = ((((((((((((y * u + p11) * u + p10) * u +        &
         p9) * u + p8) * u + p7) * u + p6) * u + p5) * u + &
         p4) * u + p3) * u + p2) * u + p1) * u + p0) * t * &
         exp(-x * x)
      if (x .lt. 0) y = 2 - y

      end function derfc
!

! Gamma function in double precision
! ! http://www.kurims.kyoto-u.ac.jp/~ooura/gamerf.html
      function dgamma(x) result(RES)
      double precision , intent(in) :: x
      double precision              :: w,y,RES
      integer                       :: k,n

         double precision,parameter :: p0 = 0.999999999999999990d+00
         double precision,parameter :: p1 = -0.422784335098466784d+00
         double precision,parameter :: p2 = -0.233093736421782878d+00
         double precision,parameter :: p3 = 0.191091101387638410d+00
         double precision,parameter :: p4 = -0.024552490005641278d+00
         double precision,parameter :: p5 = -0.017645244547851414d+00
         double precision,parameter :: p6 = 0.008023273027855346d+00

         double precision,parameter :: p7 = -0.000804329819255744d+00
         double precision,parameter :: p8 = -0.000360837876648255d+00
         double precision,parameter :: p9 = 0.000145596568617526d+00
         double precision,parameter :: p10 = -0.000017545539395205d+00
         double precision,parameter :: p11 = -0.000002591225267689d+00
         double precision,parameter :: p12 = 0.000001337767384067d+00
         double precision,parameter :: p13 = -0.000000199542863674d+00
      n = nint(x - 2)
      w = x - (n + 2)
      y = ((((((((((((p13 * w + p12) * w + p11) * w + p10) *  &
         w + p9) * w + p8) * w + p7) * w + p6) * w + p5) *    &
         w + p4) * w + p3) * w + p2) * w + p1) * w + p0
      if (n .gt. 0) then
          w = x - 1
          do k = 2, n
              w = w * (x - k)
          end do
      else
          w = 1
          do k = 0, -n - 1
              y = y * (x + k)
          end do
      end if
      RES = w / y
      end function dgamma


      SUBROUTINE kaiser(a, nrows, n, eigenv, trace, sume, ier)

!  EIGENVALUES AND VECTORS OF A SYMMETRIC +VE DEFINITE MATRIX,
!  USING KAISER'S METHOD.
!  REFERENCE: KAISER,H.F. 'THE JK METHOD: A PROCEDURE FOR FINDING THE
!  EIGENVALUES OF A REAL SYMMETRIC MATRIX', COMPUT.J., VOL.15, 271-273, 1972.

!  ARGUMENTS:-
!  A       = INPUT, AN ARRAY CONTAINING THE MATRIX
!            OUTPUT, THE COLUMNS OF A CONTAIN THE NORMALIZED EIGENVECTORS
!            OF A.   N.B. A IS OVERWRITTEN !
!  NROWS   = INPUT, THE FIRST DIMENSION OF A IN THE CALLING PROGRAM.
!  N       = INPUT, THE ORDER OF A, I.E. NO. OF COLUMNS.
!            N MUST BE <= NROWS.
!  EIGENV()= OUTPUT, A VECTOR CONTAINING THE ORDERED EIGENVALUES.
!  TRACE   = OUTPUT, THE TRACE OF THE INPUT MATRIX.
!  SUME    = OUTPUT, THE SUM OF THE EIGENVALUES COMPUTED.
!            N.B. ANY SYMMETRIC MATRIX MAY BE INPUT, BUT IF IT IS NOT +VE
!            DEFINITE, THE ABSOLUTE VALUES OF THE EIGENVALUES WILL BE FOUND.
!            IF TRACE = SUME, THEN ALL OF THE EIGENVALUES ARE POSITIVE
!            OR ZERO.   IF SUME > TRACE, THE DIFFERENCE IS TWICE THE SUM OF
!            THE EIGENVALUES WHICH HAVE BEEN GIVEN THE WRONG SIGNS !
!  IER     = OUTPUT, ERROR INDICATOR
!             = 0 NO ERROR
!             = 1 N > NROWS OR N < 1
!             = 2 FAILED TO CONVERGE IN 10 ITERATIONS

!  LATEST REVISION - 6 September 1990
!  Fortran 90 version - 20 November 1998

!*************************************************************************


REAL (kind=dp), INTENT(IN OUT) :: a(:,:)
INTEGER, INTENT(IN)       :: nrows
INTEGER, INTENT(IN)       :: n
REAL (kind=dp), INTENT(OUT)    :: eigenv(:)
REAL (kind=dp), INTENT(OUT)    :: trace
REAL (kind=dp), INTENT(OUT)    :: sume
INTEGER, INTENT(OUT)      :: ier

! Local variables

REAL (kind=dp), PARAMETER :: small = 1.0e-12_dp, zero = 0.0_dp, half = 0.5_dp,  &
                        one = 1.0_dp
INTEGER   :: i, iter, j, k, ncount, nn
REAL (kind=dp) :: absp, absq, COS, ctn, eps, halfp, p, q, SIN, ss, TAN, temp, xj, xk

!   CALCULATE CONVERGENCE TOLERANCE, EPS.
!   CALCULATE TRACE.   INITIAL SETTINGS.

ier = 1
IF(n < 1 .OR. n > nrows) RETURN
ier = 0
iter = 0
trace = zero
ss = zero
DO j = 1,n
  trace = trace + a(j,j)
  DO i = 1,n
    ss = ss + a(i,j)**2
  END DO
END DO
sume = zero
eps = small*ss/n
nn = n*(n-1)/2
ncount = nn

!   ORTHOGONALIZE PAIRS OF COLUMNS J & K, K > J.

20 DO j = 1,n-1
  DO k = j+1,n

!   CALCULATE PLANAR ROTATION REQUIRED

    halfp = zero
    q = zero
    DO i = 1,n
      xj = a(i,j)
      xk = a(i,k)
      halfp = halfp + xj*xk
      q = q + (xj+xk) * (xj-xk)
    END DO
    p = halfp + halfp
    absp = ABS(p)

!   If P is very small, the vectors are almost orthogonal.
!   Skip the rotation if Q >= 0 (correct ordering).

    IF (absp < eps .AND. q >= zero) THEN
      ncount = ncount - 1
      IF (ncount <= 0) GO TO 160
      CYCLE
    END IF

!   Rotation needed.

    absq = ABS(q)
    IF(absp <= absq) THEN
      TAN = absp/absq
      COS = one/SQRT(one + TAN*TAN)
      SIN = TAN*COS
    ELSE
      ctn = absq/absp
      SIN = one/SQRT(one + ctn*ctn)
      COS = ctn*SIN
    END IF
    COS = SQRT((one + COS)*half)
    SIN = SIN/(COS + COS)
    IF(q < zero) THEN
      temp = COS
      COS = SIN
      SIN = temp
    END IF
    IF(p < zero) SIN = -SIN

!   PERFORM ROTATION

    DO i = 1,n
      temp = a(i,j)
      a(i,j) = temp*COS + a(i,k)*SIN
      a(i,k) = -temp*SIN + a(i,k)*COS
    END DO
  END DO
END DO
ncount = nn
iter = iter + 1
IF(iter < 10) GO TO 20
ier = 2

!   CONVERGED, OR GAVE UP AFTER 10 ITERATIONS

160 DO j = 1,n
  temp = SUM( a(1:n,j)**2 )
  eigenv(j) = SQRT(temp)
  sume = sume + eigenv(j)
END DO

!   SCALE COLUMNS TO HAVE UNIT LENGTH

DO j = 1,n
  IF (eigenv(j) > zero) THEN
    temp = one/eigenv(j)
  ELSE
    temp = zero
  END IF
  a(1:n,j) = a(1:n,j)*temp
END DO

RETURN
END SUBROUTINE kaiser

!Student's t-distribution. (17/1968)
!START OF AS 3
      FUNCTION PROBST(T, IDF, IFAULT) result(prob)
!
!        ALGORITHM AS 3  APPL. STATIST. (1968) VOL.17, P.189
!
!        STUDENT T PROBABILITY (LOWER TAIL)
!
      DOUBLE PRECISION , INTENT(IN) :: T
      INTEGER          , INTENT(IN) :: IDF
      INTEGER          ,INTENT(OUT) :: IFAULT
      DOUBLE PRECISION :: A, B, C, F, G1, S, FK, ZERO, ONE, TWO, HALF, ZSQRT, ZATAN
      DOUBLE PRECISION :: prob
      INTEGER :: IM2,IOE,K,KS
!
!        G1 IS RECIPROCAL OF PI
!
      DATA ZERO, ONE, TWO, HALF, G1/0.0, 1.0, 2.0,  0.5, 0.3183098862/
!
      ZSQRT(A) = SQRT(A)
      ZATAN(A) = ATAN(A)
!
      IFAULT = 1
      PROB = ZERO
      IF (IDF .LT. 1) RETURN
      IFAULT = 0
      F = IDF
      A = T / ZSQRT(F)
      B = F / (F + T ** 2)
      IM2 = IDF - 2
      IOE = MOD(IDF, 2)
      S = ONE
      C = ONE
      F = ONE
      KS = 2 + IOE
      FK = KS
      IF (IM2 .LT. 2) GOTO 20
      DO 10 K = KS, IM2, 2
      C = C * B * (FK - ONE) / FK
      S = S + C
      IF (S .EQ. F) GOTO 20
      F = S
      FK = FK + TWO
   10 CONTINUE
   20 IF (IOE .EQ. 1) GOTO 30
      PROB = HALF + HALF * A * ZSQRT(B) * S
      RETURN
   30 IF (IDF .EQ. 1) S = ZERO
      PROB = HALF + (A * B * S + ZATAN(A)) * G1
      RETURN
      END FUNCTION PROBST
!END OF AS 3

      FUNCTION STUDNT (T, DOFF, IFAULT) result(ut)
!
!     ALGORITHM AS 27  APPL. STATIST. VOL.19, NO.1
!
!     Calculate the upper tail area under Student's t-distribution
!
!     Translated from Algol by Alan Miller
!
      INTEGER IFAULT
      DOUBLE PRECISION T, DOFF,UT
!
!     Local variables
!
      DOUBLE PRECISION V, X, TT, TWO, FOUR, ONE, ZERO, HALF
      DOUBLE PRECISION A1, A2, A3, A4, A5, B1, B2, &
          C1, C2, C3, C4, C5, D1, D2,  &
          E1, E2, E3, E4, E5, F1, F2,  &
          G1, G2, G3, G4, G5, H1, H2,  &
          I1, I2, I3, I4, I5, J1, J2
      LOGICAL POS
      DATA TWO /2.0/, FOUR /4.0/, ONE /1.0/, ZERO /0.0/, HALF /0.5/
      DATA A1, A2, A3, A4, A5 /0.09979441, -0.581821, 1.390993,      &
          -1.222452, 2.151185/, B1, B2 /5.537409, 11.42343/
      DATA C1, C2, C3, C4, C5 /0.04431742, -0.2206018, -0.03317253,  &
          5.679969, -12.96519/, D1, D2 /5.166733, 13.49862/
      DATA E1, E2, E3, E4, E5 /0.009694901, -0.1408854, 1.88993,     &
          -12.75532, 25.77532/, F1, F2 /4.233736, 14.3963/
      DATA G1, G2, G3, G4, G5 /-9.187228E-5, 0.03789901, -1.280346,  &
          9.249528, -19.08115/, H1, H2 /2.777816, 16.46132/
      DATA I1, I2, I3, I4, I5 /5.79602E-4, -0.02763334, 0.4517029,   &
          -2.657697, 5.127212/, J1, J2 /0.5657187, 21.83269/
!
!     Check that number of degrees of freedom > 4.
!
      IF (DOFF .LT. TWO) THEN
	IFAULT = 1
	UT = - ONE
	RETURN
      END IF
!
      IF (DOFF .LE. FOUR) THEN
	IFAULT = DOFF
      ELSE
	IFAULT = 0
      END IF
!
!     Evaluate series.
!
      V = ONE / DOFF
      POS = (T .GE. ZERO)
      TT = ABS(T)
      X = HALF * (ONE +                                          &
         TT * (((A1 + V * (A2 + V * (A3 + V * (A4 + V * A5)))) / &
             (ONE - V * (B1 - V * B2))) +                        &
         TT * (((C1 + V * (C2 + V * (C3 + V * (C4 + V * C5)))) / &
             (ONE - V * (D1 - V * D2))) +                        &
         TT * (((E1 + V * (E2 + V * (E3 + V * (E4 + V * E5)))) / &
             (ONE - V * (F1 - V * F2))) +                        &
         TT * (((G1 + V * (G2 + V * (G3 + V * (G4 + V * G5)))) / &
             (ONE - V * (H1 - V * H2))) +                        &
         TT * ((I1 + V * (I2 + V * (I3 + V * (I4 + V * I5)))) /  &
             (ONE - V * (J1 - V * J2))) ))))) ** (-8)
      IF (POS) THEN
	UT = X
      ELSE
	UT = ONE - X
      END IF
!
      RETURN
      END FUNCTION STUDNT

!Complete beta function
  function betaC(x,y) result(RES)
    double precision , intent(in)  :: x,y
    double precision   :: RES

    RES = dgamma(x)*dgamma(y)/dgamma(x+Y)

  end function betaC


!Incomplete gamma function
      function gammds (y,p,ifault) result(RES)
!
!        Algorithm AS 147  Appl. Statist. (1980) Vol. 29, No. 1
!
!        Computes the incomplete gamma integral for positive
!        parameters y,p using an infinite series
!
!        Auxiliary function required: ALNGAM = CACM algorithm 291
!
!	     AS239 should be considered as an alternative to AS147
!
      double precision ,intent(in):: y,p
      integer, intent(out)  :: ifault
      double precision :: RES,e,zero,one,uflo,arg,c,a,f
      data e/1.0d-9/, zero/0.0d0/, one/1.0d0/, uflo/1.0d-37/
!
!        Checks admissibility of arguments and value of f
!
      ifault = 1
      RES = zero
      if(y.le.zero .or. p.le.zero) return
      ifault = 2
!
!        alngam is natural log of gamma function
!
      arg = p*log(y)-log(dgamma(p+one))-y
      if(arg.lt.log(uflo)) return
      f = exp(arg)
      if(f.eq.zero) return
      ifault = 0
!
!          Series begins
!
      c = one
      RES = one
      a = p
    1 a = a+one
      c = c*y/a
      RES = RES+c
      if (c/RES.gt.e) goto 1
      RES = RES*f
      return
      end function gammds

!!SLATEC
!!------

     FUNCTION D1MACH (I) RESULT(RES)
         INTEGER , INTENT(IN) :: I
         DOUBLE PRECISION :: RES
         IF ( I == 1 ) RES = tiny ( RES )
         IF ( I == 2 ) RES = huge ( RES )
         IF ( I == 3 ) RES = epsilon ( RES ) / radix ( RES )
         IF ( I == 4 ) RES = epsilon ( RES )
         IF ( I == 5 ) RES = log ( dble(radix ( RES )) )
     END FUNCTION D1MACH


!*DECK DBETAI
      FUNCTION DBETAI (X, PIN, QIN) RESULT (RES)
!***BEGIN PROLOGUE  DBETAI
!***PURPOSE  Calculate the incomplete Beta function.
!***LIBRARY   SLATEC (FNLIB)
!***CATEGORY  C7F
!***TYPE      DOUBLE PRECISION (BETAI-S, DBETAI-D)
!***KEYWORDS  FNLIB, INCOMPLETE BETA FUNCTION, SPECIAL FUNCTIONS
!***AUTHOR  Fullerton, W., (LANL)
!***DESCRIPTION
!
!   DBETAI calculates the DOUBLE PRECISION incomplete beta function.
!
!   The incomplete beta function ratio is the probability that a
!   random variable from a beta distribution having parameters PIN and
!   QIN will be less than or equal to X.
!
!     -- Input Arguments -- All arguments are DOUBLE PRECISION.
!   X      upper limit of integration.  X must be in (0,1) inclusive.
!   PIN    first beta distribution parameter.  PIN must be .GT. 0.0.
!   QIN    second beta distribution parameter.  QIN must be .GT. 0.0.
!
!***REFERENCES  Nancy E. Bosten and E. L. Battiste, Remark on Algorithm
!                 179, Communications of the ACM 17, 3 (March 1974),
!                 pp. 156.
!***ROUTINES CALLED  D1MACH, DLBETA, XERMSG
!***REVISION HISTORY  (YYMMDD)
!   770701  DATE WRITTEN
!   890531  Changed all specific intrinsics to generic.  (WRB)
!   890911  Removed unnecessary intrinsics.  (WRB)
!   890911  REVISION DATE from Version 3.2
!   891214  Prologue converted to Version 4.0 format.  (BAB)
!   900315  CALLs to XERROR changed to CALLs to XERMSG.  (THJ)
!   920528  DESCRIPTION and REFERENCES sections revised.  (WRB)
!***END PROLOGUE  DBETAI
      DOUBLE PRECISION X, PIN, QIN, ALNEPS, ALNSML, C, EPS, FINSUM, P, &
                       PS, Q, SML, TERM, XB, XI, Y, D1MACH, DLBETA, P1
      LOGICAL FIRST
      INTEGER I,IB,N
      DOUBLE PRECISION RES
      SAVE EPS, ALNEPS, SML, ALNSML, FIRST
      DATA FIRST /.TRUE./
!***FIRST EXECUTABLE STATEMENT  DBETAI
      IF (FIRST) THEN
         EPS = D1MACH(3)
         ALNEPS = LOG (EPS)
         SML = D1MACH(1)
         ALNSML = LOG (SML)
      ENDIF
      FIRST = .FALSE.
!
      IF (X .LT. 0.D0 .OR. X .GT. 1.D0) THEN
         CALL STOP_APPLICATION('DEBATI : X IS NOT IN THE RANGE (0,1)')
      END IF
      IF (PIN .LE. 0.D0 .OR. QIN .LE. 0.D0) THEN
        CALL STOP_APPLICATION ('DBETAI : P AND/OR Q IS LE ZERO')
      END IF
!
      Y = X
      P = PIN
      Q = QIN
      IF (Q.LE.P .AND. X.LT.0.8D0) GO TO 20
      IF (X.LT.0.2D0) GO TO 20
      Y = 1.0D0 - Y
      P = QIN
      Q = PIN
!
 20   IF ((P+Q)*Y/(P+1.D0).LT.EPS) GO TO 80
!
! EVALUATE THE INFINITE SUM FIRST.  TERM WILL EQUAL
! Y**P/BETA(PS,P) * (1.-PS)-SUB-I * Y**I / FAC(I) .
!
      PS = Q - AINT(Q)
      IF (PS.EQ.0.D0) PS = 1.0D0
      XB = P*LOG(Y) - DLBETA(PS,P) - LOG(P)
      RES = 0.0D0
      IF (XB.LT.ALNSML) GO TO 40
!
      RES = EXP (XB)
      TERM = RES*P
      IF (PS.EQ.1.0D0) GO TO 40
      N = MAX (ALNEPS/LOG(Y), 4.0D0)
      DO 30 I=1,N
        XI = I
        TERM = TERM * (XI-PS)*Y/XI
        RES = RES + TERM/(P+XI)
 30   CONTINUE
!
! NOW EVALUATE THE FINITE SUM, MAYBE.
!
 40   IF (Q.LE.1.0D0) GO TO 70
!
      XB = P*LOG(Y) + Q*LOG(1.0D0-Y) - DLBETA(P,Q) - LOG(Q)
      IB = MAX (XB/ALNSML, 0.0D0)
      TERM = EXP(XB - IB*ALNSML)
      C = 1.0D0/(1.D0-Y)
      P1 = Q*C/(P+Q-1.D0)
!
      FINSUM = 0.0D0
      N = Q
      IF (Q.EQ.DBLE(N)) N = N - 1
      DO 50 I=1,N
        IF (P1.LE.1.0D0 .AND. TERM/EPS.LE.FINSUM) GO TO 60
        XI = I
        TERM = (Q-XI+1.0D0)*C*TERM/(P+Q-XI)
!
        IF (TERM.GT.1.0D0) IB = IB - 1
        IF (TERM.GT.1.0D0) TERM = TERM*SML
!
        IF (IB.EQ.0) FINSUM = FINSUM + TERM
 50   CONTINUE
!
 60   RES = RES + FINSUM
 70   IF (Y.NE.X .OR. P.NE.PIN) RES = 1.0D0 - RES
      RES = MAX (MIN (RES, 1.0D0), 0.0D0)
      RETURN
!
 80   RES = 0.0D0
      XB = P*LOG(MAX(Y,SML)) - LOG(P) - DLBETA(P,Q)
      IF (XB.GT.ALNSML .AND. Y.NE.0.0D0) RES = EXP(XB)
      IF (Y.NE.X .OR. P.NE.PIN) RES = 1.0D0 - RES
!
      RETURN
      END FUNCTION DBETAI

      subroutine ginv1(a,n,m,tol,irank)
! returns generalized inverse of matrix x of size n x n declared
! as m x m. tol is working zero and irank returns the rank of
! the matrix. w is a work vector of size m,
! by rohan fernando, slightly structured by i. misztal 05/05/87

      integer :: n,m
      double precision a(m,m),w(m),re,sum,tol
      integer :: i,j,iim1,ii,iii,irank
      irank=n
      do 10 i=1,n
         do 20 j=1,i-1
              re=a(i,j)
              do 20 ii=i,n
20                 a(ii,i)=a(ii,i)-re*a(ii,j)
         if (a(i,i).lt.tol) then
              a(i,i)=0.0
              do 45 ii=i+1,n
45                 a(ii,i)=0.
           irank=irank-1
           else
              a(i,i)=sqrt(a(i,i))
              do 40 ii=i+1,n
40                a(ii,i)=a(ii,i)/a(i,i)
         endif
10    continue

      do 100 i=1,n
         if (a(i,i).eq.0.) then
              do 150 ii=i+1,n
150                a(ii,i)=0.
           else
              a(i,i)=1.0/ a(i,i)
              do 200 ii=i+1,n
200               w(ii)=0.0
              do 300 ii=i+1,n
                  iim1=ii-1
                  re=a(iim1,i)
                  do 400 iii=ii,n
400                   w(iii)=w(iii)-a(iii,iim1)*re
                  if (a(ii,ii).eq.0.) then
                      a(ii,i)=0.
                    else
                      a(ii,i)=w(ii)/a(ii,ii)
                  endif
300           continue
          endif
100     continue

      do 110 j=1,n
         do 110 i=j,n
              sum=0
              do 130 ii=i,n
130                sum=sum+a(ii,j)*a(ii,i)
110           a(i,j)=sum
      do 600 i=1,n
          do 600 j=i,n
600           a(i,j)=a(j,i)
      return
      end subroutine ginv1


      SUBROUTINE tdiag (n, tol, a, d, e, z, maxdim, ifault)

! Code converted using TO_F90 by Alan Miller
! Date: 2003-12-03  Time: 12:16:20

INTEGER, INTENT(IN)      :: n
REAL (dp), INTENT(OUT)   :: tol
INTEGER, INTENT(IN)  :: maxdim
REAL (dp), INTENT(IN)    :: a(maxdim,maxdim)
REAL (dp), INTENT(OUT)   :: d(maxdim)
REAL (dp), INTENT(OUT)   :: e(maxdim)
REAL (dp), INTENT(OUT)   :: z(maxdim,maxdim)
INTEGER, INTENT(OUT)     :: ifault

REAL (dp), PARAMETER  :: eta = 1.0D-37, eps = 1.0D-14

!    Algorithm as 60.1 appl.statist. (1973) vol.22 no.2

!    Reduces real symmetric matrix to tridiagonal form

!    tol is a machine dependent constant , tol = eta/eps , where
!    eta is the smallest positive number representable in the
!    computer and eps is the smallest positive number for which
!    1+eps.ne.1.

!      eta=eps*tol
!      eps=0.7105427358e-14
!      tol=0.3131513063e-293
!      precis=1.0e-14

!      NB
!        Real constants must be <= 15 decimal digits
!        the range of a real constant is from 1.0e-293 to 1.0e+322

INTEGER    :: i, i1, j, j1, k, l
REAL (dp)  :: f, g, h, hh,zero,one

zero=0.d0
one=1.d0
tol=eta/eps
ifault=1
IF (n <= 1) RETURN
ifault=0
DO  i=1,n
  z(i,1:i)=a(i,1:i)
END DO
i=n
DO  i1=2,n
  l=i-2
  f=z(i,i-1)
  g=zero
  IF (l < 1) GO TO 30
  g = SUM( z(i,1:l)**2 )
  30 h=g + f*f

!       If g is too small for orthogonality to be guaranteed, the
!       transformation is skipped

  IF (g > tol) GO TO 40
  e(i)=f
  d(i)=zero
  GO TO 100

  40 l=l+1
  g=SQRT(h)
  IF (f >= zero) g=-g
  e(i)=g
  h=h - f*g
  z(i,i-1)=f-g
  f=zero
  DO  j=1,l
    z(j,i)=z(i,j)/h
    g=zero

!       Form element of a * u

    DO  k=1,j
      g=g + z(j,k)*z(i,k)
    END DO
    IF (j >= l) GO TO 70
    j1=j+1
    DO  k=j1,l
      g=g + z(k,j)*z(i,k)
    END DO

!       Form element of p

    70 e(j)=g/h
    f=f + g*z(j,i)
  END DO

!       Form k

  hh=f/(h+h)

!       Form reduced a

  DO  j=1,l
    f=z(i,j)
    g=e(j)-hh*f
    e(j)=g
    DO  k=1,j
      z(j,k)=z(j,k) - f*e(k) - g*z(i,k)
    END DO
  END DO
  d(i)=h
  100 i=i-1
END DO
d(1)=zero
e(1)=zero

!       Accumulation of transformation matrices

DO  i=1,n
  l=i-1
  IF (d(i) == zero .OR. l == 0) GO TO 140
  DO  j=1,l
    g=zero
    DO  k=1,l
      g=g + z(i,k)*z(k,j)
    END DO
    DO  k=1,l
      z(k,j)=z(k,j) - g*z(k,i)
    END DO
  END DO
  140 d(i)=z(i,i)
  z(i,i)=one
  IF (l == 0) CYCLE
  DO  j=1,l
    z(i,j)=zero
    z(j,i)=zero
  END DO
END DO
RETURN
END SUBROUTINE tdiag


SUBROUTINE lrvt (n, precis, d, e, z, ifault, maxdim)

INTEGER, INTENT(IN)        :: n
REAL (dp), INTENT(OUT)     :: precis
INTEGER, INTENT(IN)        :: maxdim
REAL (dp), INTENT(IN OUT)  :: d(maxdim)
REAL (dp), INTENT(OUT)     :: e(maxdim)
REAL (dp), INTENT(IN OUT)  :: z(maxdim,maxdim)
INTEGER, INTENT(OUT)       :: ifault

!    Algorithm AS 60.2 appl.statist. (1973) vol.22, no.2

!    Finds latent roots and vectors of tridiagonal matrix

INTEGER, PARAMETER    :: mits = 30
REAL (dp), PARAMETER  :: two = 2.0D0

INTEGER    :: i, i1, j, jj, k, l, m, m1, n1
REAL (dp)  :: b, c, f, g, h, p, pr, r, s,zero,one
zero=0.d0
one=1.d0
precis=1.0D-14
ifault=2
IF (n <= 1) RETURN
ifault=1
n1=n-1
DO  i=2,n
  e(i-1)=e(i)
END DO
e(n)=zero
b=zero
f=zero
DO  l=1,n
  jj=0
  h=precis*(ABS(d(l))+ABS(e(l)))
  IF (b < h) b=h

!       Look for small sub-diagonal element

  DO  m1=l,n
    m=m1
    IF (ABS(e(m)) <= b) EXIT
  END DO
  IF (m == l) GO TO 90
  40 IF (jj == mits) RETURN
  jj=jj+1

!       Form shift

  p=(d(l+1)-d(l))/(two*e(l))
  r=SQRT(p*p + one)
  pr=p+r
  IF (p < zero) pr=p-r
  h=d(l)-e(l)/pr
  DO  i=l,n
    d(i)=d(i)-h
  END DO
  f=f+h

!       QL transformation

  p=d(m)
  c=one
  s=zero
  m1=m-1
  i=m
  DO  i1=l,m1
    j=i
    i=i-1
    g=c*e(i)
    h=c*p
    IF (ABS(p) >= ABS(e(i))) GO TO 60
    c=p/e(i)
    r=SQRT(c*c + one)
    e(j)=s*e(i)*r
    s=one/r
    c=c/r
    GO TO 70
    60 c=e(i)/p
    r=SQRT(c*c + one)
    e(j)=s*p*r
    s=c/r
    c=one/r
    70 p=c*d(i) - s*g
    d(j)=h + s*(c*g + s*d(i))

!       Form vector

    DO  k=1,n
      h=z(k,j)
      z(k,j)=s*z(k,i) + c*h
      z(k,i)=c*z(k,i) - s*h
    END DO
  END DO
  e(l)=s*p
  d(l)=c*p
  IF (ABS(e(l)) > b) GO TO 40
  90 d(l)=d(l) + f
END DO

!       Order latent roots and vectors

DO  i=1,n1
  k=i
  p=d(i)
  i1=i+1
  DO  j=i1,n
    IF (d(j) <= p) CYCLE
    k=j
    p=d(j)
  END DO
  IF (k == i) CYCLE
  d(k)=d(i)
  d(i)=p
  DO  j=1,n
    p=z(j,i)
    z(j,i)=z(j,k)
    z(j,k)=p
  END DO
END DO
ifault=0
RETURN
END SUBROUTINE lrvt




end module m_qtlmap_math
