!!****p* QTLMap/TOOLS
!!  NAME
!!    TOOLS
!!  DESCRIPTION
!!  Package tools :
!!     contains tools modules : mathematical and optimlization subroutines, lexer and parser subroutines
!!
!!  25/09/2009
!!  COPYRIGHT
!!***

!!****m* TOOLS/m_qtlmap_tools
!!  NAME
!!    m_qtlmap_tools
!!  DESCRIPTION
!!   misc. routines.
!!  NOTES
!!
!!  BUGS
!!
!!  SEE ALSO
!!
!!  COPYRIGHT
!!***
!! You can use this space for remarks that should not be included
!! in the documentation.
!!/
module m_qtlmap_tools
      use m_qtlmap_constant
      use m_qtlmap_data
      use m_qtlmap_haplotype_data
      use m_qtlmap_log
      use QTLMAP_TYPES

      implicit none
      private

      public :: file_exist
      public :: get_real
      public :: get_int
      public :: check_allocate
      public :: next_word
      public :: parse_real_array
      public :: xaldane
      public :: xosambi
      public :: stop_on_error
      public :: stop_application
      public :: str

#ifndef VARYING_STRING_MOD
      public :: REMOVE
      public :: EXTRACT
      public :: GET
#endif

       !------------ SERVICES ASSOCIATED WITH DATA ----------------------------
      public :: set_base_and_step
      public :: get_long_step_morgan
      public :: set_absi
      public :: get_ilong
      public :: get_dx
      public :: get_pos
      public :: get_npo
      public :: get_maxnpo
      public :: init_pheno
      public :: set_pheno
      public :: get_pheno
      public :: get_ind_pheno
      public :: release_data

      public :: new
      public :: release
      public :: copy

      public :: get_maxnbgenotypedam

      public :: clean_lrt_solution


!!****f* m_qtlmap_tools/new
!! NAME
!!   new
!! DESCRIPTION
!!   generic interface to allocate internal array of a qtlmap variable with a complex type
!! SOURCE

      interface new
        module procedure new_lrt_solution
      end interface new
!!***

!!****f* m_qtlmap_tools/release
!! NAME
!!   release
!! DESCRIPTION
!!   generic interface to release internal array of a qtlmap variable with a complex type
!! SOURCE
      interface release
        module procedure release_lrt_solution, &
                         release_incidence_solution, &
                         release_description_type
      end interface release
!!***

!!****f* m_qtlmap_tools/copy
!! NAME
!!   copy
!! DESCRIPTION
!!   generic interface to copy a variable of qtlmap type
!! SOURCE
      interface copy
        module procedure copy_description_type
      end interface copy

!!***

!!****f* m_qtlmap_tools/str
!! NAME
!!   str
!! DESCRIPTION
!!   generic interface to convert a float/double or an integer to a string
!! SOURCE
      interface str
        module procedure iget_str, &
                         fget_str, &
                         dget_str
      end interface str
!!***

!!****f* m_qtlmap_tools/get_real
!! NAME
!!   get_real
!! DESCRIPTION
!!   generic interface to convert a word into a float
!! SOURCE
      interface get_real
        module procedure get_real_char
      end interface get_real
!!***

!!****f* m_qtlmap_tools/get_int
!! NAME
!!   get_int
!! DESCRIPTION
!!   generic interface to convert a word into an integer
!! SOURCE
      interface get_int
        module procedure get_int_char
      end interface get_int
!!***

!!****f* m_qtlmap_tools/next_word
!! NAME
!!   next_word
!! DESCRIPTION
!!   generic interface to read a word from line (parsing context)
!! SOURCE
      interface next_word
        module procedure next_word_char
      end interface next_word
!!***

      contains

!!****f* m_qtlmap_tools/file_exist
!! NAME
!!   file_exist
!! DESCRIPTION
!!   Test if a local file exist (the programe stop if not)
!! INPUTS
!!  file_in              : the pathname file
!!
!! SOURCE
!!
              subroutine file_exist(file_in)
      			character(len=*),intent(inout) ::file_in
      			logical::existe
      			inquire( file=file_in, exist=existe)

      			if(.not. existe) then
                   call stop_application('The file ['//trim(file_in)//'] does not exist.')
      			endif
              end subroutine file_exist
!!***

!!****f* m_qtlmap_tools/stop_on_error
!! NAME
!!   stop_on_error
!! DESCRIPTION
!!   stop the application with an error message on a read file context
!! INPUTS
!!  iostat              : iostat error code
!!  file                : the pathname file
!!  lineNumber          : the number of the current line
!!  message             : a message error
!!
!! SOURCE
!!
            subroutine stop_on_error(iostat,file,lineNumber,message)
              integer , intent(in)        ::iostat
              character(len=*),intent(in) ::file
              integer , intent(in)        ::lineNumber
              character(len=*),intent(in) ::message

              if (iostat /= 0 .and. iostat/=-2) then
                call log_mess('*** Read error ['//trim(file)//'] ' // &
                'line '//trim(iget_str(lineNumber))//':bad specification [message:'//adjustr(message)//'] ***',ERROR_DEF)
                stop 'read error'
              endif

            end subroutine stop_on_error
!!***

!!****f* m_qtlmap_tools/stop_application
!! NAME
!!   stop_application
!! DESCRIPTION
!!   stop the application with an error message
!! INPUTS
!!  message              : an error message
!!
!! SOURCE
!!
            subroutine stop_application(message)
              character(len=*),intent(in) ::message
              call log_mess('*** message error :'//message//' ***',ERROR_DEF)
              stop 'runtime error'

            end subroutine stop_application
!!***

!!****f* m_qtlmap_tools/parse_real_array
!! NAME
!!   parse_real_array
!! DESCRIPTION
!!   get a set of real value from a line
!! INPUTS
!!  nvalue               : number of value
!! INPUTS/OUTPUT
!!  line_to_parse        : the line
!! OUTPUTS
!!  tabV                 : a vector of double precision
!!  is_ok                : a flag error
!! SOURCE
!!
           subroutine parse_real_array(line_to_parse,nvalue,tabV,is_ok)
              character(len=*),intent(inout)                 :: line_to_parse
              integer, intent(in)                            :: nvalue
              real(kind=dp) ,dimension(nvalue),intent(out)   :: tabV
              logical, intent(inout),optional                :: is_ok
              character(len=LEN_BUFFER_WORD)                 :: w_token
              integer                                        :: i,ns,ne
              character(len=1000)                            :: endline

              if ( trim(line_to_parse)=='' ) then
                 if ( present(is_ok) ) is_ok = .false.
                 return
              end if

              ns=index(line_to_parse,"[")
              if ( ns == 0 ) then
                 if ( present(is_ok) ) is_ok = .false.
                 return
              end if
              ne=index(line_to_parse,"]")

              if ( ne == 0 ) then
                 if ( present(is_ok) ) is_ok = .false.
                 return
              end if

              endline = line_to_parse(ne+1:)
              line_to_parse = line_to_parse(ns+1:ne-1)

              do i=1,nvalue
                 w_token=trim(next_word_char(line_to_parse,is_ok))
                 if (is_ok) tabV(i)=get_real(w_token,is_ok)
                 if ( .not. is_ok ) then
                   if ( present(is_ok) ) is_ok = .false.
                   return
                 end if
              end do

              if ( trim(line_to_parse) /= '' ) then
                 if ( present(is_ok) ) is_ok = .false.
                 return
              end if

              line_to_parse = endline

           end subroutine parse_real_array
!!***

!!****f* m_qtlmap_tools/next_word_char
!! NAME
!!    next_word_char
!! DESCRIPTION
!!    get the last word at the right side of a line and remove it.
!!
!! INPUTS/OUTPUT
!!  line_to_parse        : the line.
!! OUTPUTS
!!  is_ok                : a flag error
!! RETURN
!!  the word
!! SOURCE
!!
           FUNCTION next_word_char(line_to_parse,is_ok) result(w_token)

            character(len=*) ,intent(inout)        :: line_to_parse
            logical, intent(inout),optional        :: is_ok
            character(len=LEN_BUFFER_WORD)         :: w_token
            integer                                :: nd

            nd = 1
            w_token = ''
            line_to_parse=adjustl(line_to_parse)

            if ( trim(line_to_parse)=='' ) then
                 if ( present(is_ok) ) is_ok = .false.
                 return
            end if


            do while (nd == 1 .and. w_token == '')
              !! char(9) pour les tabulation.....
              nd=SCAN(line_to_parse," ;"//char(9))

              if ( nd == 0) then
                  w_token = trim(line_to_parse)
              else
                  w_token = trim(EXTRACT(line_to_parse,1,nd-1))
              end if


              if (w_token == '') then
               	line_to_parse = REMOVE(line_to_parse,1,nd)
              end if
            end do

            if  (nd == 0) then ! END OF LINE
               w_token = trim(EXTRACT(line_to_parse,1))
               line_to_parse = ''
               if ( present(is_ok) ) is_ok = .true.
               return
            end if

            if (nd > 0) then  ! at least one non-terminator character in the word
                w_token = trim(EXTRACT(line_to_parse,1,nd-1))
                if ( present(is_ok) ) is_ok = .true.
            else
                w_token = ''
                if ( present(is_ok) ) is_ok = .false.
            end if


            line_to_parse = adjustl(REMOVE(line_to_parse,1,nd))
           return
           end function next_word_char
!!***

!!****f* m_qtlmap_tools/check_allocate
!! NAME
!!    check_allocate
!! DESCRIPTION
!!    check a iostat error code. print an error message and stop the programe if the iostat error /= 0
!! INPUTS
!!  iostat        : the iostat error code
!!  error_message : an error message
!!
!! OUTPUTS
!!  is_ok             : boolean
!! RETURN
!!   a double precision
!! SOURCE
!!
          subroutine check_allocate(iostat,error_message)
          !ios_id where read the parameter
          integer,intent(in)::iostat
          !message to print if iostat /= 0
          character(len=*),optional,intent(in)::error_message

            if (iostat /= 0) then
               print *, "Allocation error:[iostat:",iostat,"]"
               if ( present (error_message) ) then
                  print *,'**'//error_message//'**'
               endif
               stop 'allocation error'
             end if
          end subroutine check_allocate
!!***

!!****f* m_qtlmap_tools/get_real_char
!! NAME
!!    get_real_char
!! DESCRIPTION
!!    conversion of a character word in a double precision using the fortran read routine
!! INPUTS
!!  string_to_convert : the word
!!
!! OUTPUTS
!!  is_ok             : boolean
!! RETURN
!!   a double precision
!! SOURCE
!!
          function get_real_char(string_to_convert,is_ok) result(real_value)
             character(len=*),intent(in)      :: string_to_convert
             logical,optional,intent(inout)       :: is_ok
             real (kind=dp)                       :: real_value
             character(len=LEN_BUFFER_WORD)       :: word_token
             integer                              :: ios
             character(len=LEN_DEF)               :: c,string
             logical                              :: is_int

             string = trim(string_to_convert)
             is_int=.true.

             do while ( string /='' )
               c = EXTRACT (string , 1, 1)
               if (c == '.') then
                 is_int=.false.
                 exit
               end if
               string = REMOVE(string,1,1)
             end do

             if ( .not. is_int ) then
                word_token = string_to_convert!CHAR(string_to_convert)
                read (word_token,FMT_REAL,IOSTAT=ios) real_value
                if ( present(is_ok) ) then
                   is_ok = ( ios == 0 )
                end if
                if (  ios /= 0 ) real_value = REAL_NOT_DEFINED
             else
                if (present(is_ok)) then
                  real_value = get_int(string_to_convert,is_ok)
                else
                  real_value = get_int(string_to_convert)
                end if
             end if
             string=''
             return

          end function get_real_char
!!***

!!****f* m_qtlmap_tools/get_int_char
!! NAME
!!    get_int_char
!! DESCRIPTION
!!    conversion of a character word in an integer using the fortran read routine
!! INPUTS
!!  string_to_convert : the word
!!
!! OUTPUTS
!!  is_ok             : boolean
!! RETURN
!!   an integer
!! SOURCE
!!
          function get_int_char(string_to_convert,is_ok) result(int_value)
             character(len=*),intent(in)      :: string_to_convert
             logical,optional,intent(inout)       :: is_ok
             integer                              :: int_value
             character(len=LEN_BUFFER_WORD)       :: word_token
             integer                              :: ios

             word_token = trim(string_to_convert)!CHAR(string_to_convert)
             read (word_token,FMT_INT,IOSTAT=ios) int_value

             if ( present(is_ok) ) then
                is_ok = ( ios == 0 )
             endif

             if (  ios /= 0 ) int_value = INT_NOT_DEFINED
             word_token =''
             return

          end function get_int_char
!!***

!!****f* m_qtlmap_tools/iget_str
!! NAME
!!    iget_str
!! DESCRIPTION
!!    conversion of an integer into a character word
!! INPUTS
!!  value_int : the word
!!
!! RETURN
!!   an integer
!! SOURCE
!!
          function iget_str(value_int) result(res)
             integer,intent(in)                   :: value_int
             character(len=20)                    :: res
             character(len=LEN_DEF)                 :: str_value,sign_value,tempvs,tempvs2
             integer                              :: v,reste,ascii_zero

             if (value_int<0) then
                sign_value='-'
                v = -value_int
             else
                sign_value=''
                v = value_int
             end if

             ascii_zero = iachar('0')
             str_value=''
             do while (v /=0 )
                reste = mod(v,10)
                tempvs=achar(ascii_zero+reste)
                tempvs2=trim(str_value)
                str_value = trim(tempvs)//trim(tempvs2)
                v = v / 10
             end do

             str_value = trim(sign_value)//trim(str_value)
             if (str_value=='' .or. str_value=='-') str_value='0'
             res = trim(str_value)

             return

          end function iget_str
!!***

!!****f* m_qtlmap_tools/fget_str
!! NAME
!!    fget_str
!! DESCRIPTION
!!    conversion of a a float into a character word
!! INPUTS
!!  value : the word
!!
!! RETURN
!!   a float
!! SOURCE
!!
          function fget_str(value) result(res)
             real           ,intent(in)           :: value
             character(len=20)                    :: res
             real (kind=dp)                       :: v

             v = value
             res = dget_str(v)

          end function fget_str
!!***

!!****f* m_qtlmap_tools/dget_str
!! NAME
!!    dget_str
!! DESCRIPTION
!!    conversion of a double precision into a character word
!! INPUTS
!!  value : the word
!!
!! RETURN
!!   a double precision
!! SOURCE
!!
          function dget_str(value) result(res)
             real (kind=dp) ,intent(in)           :: value
             character(len=20)                    :: res
             character(len=LEN_DEF)               :: str_value,c,sign_value,tempvs,tempvs2,temppoint
             integer                              :: ascii_zero,reste,decimal
             integer                              :: val_int,io
             real(kind=dp)                        :: v
             !real                                 :: v
             logical                              :: have_dec

             res=''
             write (res,fmt='(f13.2)',iostat=io) value
             if ( io /= 0 ) then
               res='<NO-DEF>'
             else
               res=adjustl(res)
             end if
             return

          end function dget_str
!!***

!!****f* m_qtlmap_tools/xaldane
!! NAME
!!    xaldane
!! DESCRIPTION
!!
!! INPUTS
!!  d :
!!
!! RETURN
!!
!! NOTES
!!     Calcul du taux de recombinaison en fonction de la distance de xaldane
!! SOURCE
!!
          function xaldane(d) result (val)
            real (kind=dp)   , intent(in)  :: d
            real  (kind=dp)                :: val
            val=0.5d0*(1.d0-dexp(-2.d0*d))
            return
          end function xaldane
!!***

!!****f* m_qtlmap_tools/xosambi
!! NAME
!!    xosambi
!! DESCRIPTION
!!
!! INPUTS
!!  d :
!!
!! RETURN
!!
!! NOTES
!!     Calcul du taux de recombinaison en fonction de la distance de Kosambi
!! SOURCE
!!
          function xosambi(d) result (val)
             real (kind=dp)   , intent(in)  :: d
             real  (kind=dp)                :: val,z

             z=dexp(4.d0*d)
             val=0.5d0*((z-1.d0)/(z+1.d0))
          return
          end function xosambi
!!***

!!****f* m_qtlmap_tools/REMOVE
!! NAME
!!    REMOVE
!! DESCRIPTION
!!    Remove a subset of a line
!! INPUTS
!!  string : the line
!!  start  : start index of the subset
!!  finish : end index of the subset
!!
!! RETURN
!!  the new line without the subset
!!
!! SOURCE
!!
        function REMOVE (string,start,finish) result(out)
           character(len=*) , intent(in)         :: string
           integer , intent(in) ,optional        :: start
           integer , intent(in) ,optional        :: finish

           character(len=len(string))                :: out
           integer                               :: s,e

           s = 1
           e = len(string)
           if (present(start)) then
             if (start > 1 .and. start <= len(string)) then
                s = start
             end if
           end if

           if (present(finish)) then
             if (finish >= start  .and. finish <= len(string)) then
                e = finish
             end if
           end if

           out=''

           if ( s > 1) then
              out=trim(string(1:s))
           end if

           if ( e < len(string) ) then
             out = trim(out) // string(e+1:len(string))
           end if

        end function REMOVE
!!***

!!****f* m_qtlmap_tools/EXTRACT
!! NAME
!!    EXTRACT
!! DESCRIPTION
!!    Get a subset of a character line
!! INPUTS
!!  string : the line
!!  start  : start index of the subset
!!  finish : end index of the subset
!!
!! RETURN
!!  the subset of the line
!!
!! SOURCE
!!
        function EXTRACT (string,start,finish) result(out)
           character(len=*) , intent(in)         :: string
           integer , intent(in) ,optional        :: start
           integer , intent(in) ,optional        :: finish

           character(len=len(string))                :: out
           integer                               :: s,e

           s=1
           e=len(string)

           if (present(start)) then
             if (start > 1 .and. start <= len(string)) then
                s = start
             end if
           end if

           if (present(finish)) then
             if (finish >= start  .and. finish <= len(string)) then
                e = finish
             end if
           end if

           out=trim(string(s:e))

        end function
!!***

!!****f* m_qtlmap_tools/GET
!! NAME
!!    GET
!! DESCRIPTION
!!    Read a line from an unit file
!! INPUTS/OUTPUTS
!!   string : the line read
!! INPUTS
!!   unit   : the unit record
!!  maxlen  : maximum length of the line
!!
!! OUTPUTS
!!   iostat : iostat error
!! SOURCE
!!
        subroutine GET (unit,string,maxlen,iostat)
           integer                , intent(in)      :: unit
           character(len=LEN_LINE) , intent(inout)  :: string
           integer , intent(in),optional            :: maxlen
           integer , intent(out),optional           :: iostat

           character(len=LEN_LINE)                   :: word
           integer :: eof

           eof = 0
           string=''
           read (unit,iostat=eof,fmt="(a)/") string
           if ( present(iostat) ) iostat = eof

        end subroutine GET
!!***

!!****f* m_qtlmap_tools/set_base_and_step
!! NAME
!!    set_base_and_step
!! DESCRIPTION
!!    Set the step (variable PAS) and the base dimension (variable BASE_STEP) according a Morgan step.
!!
!! INPUTS
!!     stepMorgan : the step in Morgan
!! SOURCE
!!
        subroutine set_base_and_step(stepMorgan)
         character(len=*)  ,intent(in) :: stepMorgan
         integer :: i,s
         character(len=LEN_W) :: buf

         buf=""
         buf=adjustl(stepMorgan)
         s=len(trim(buf))
         !on cherche le point.....
         do i=1,s
           if (buf(i:i) == '.' ) then
              exit
           end if
         end do
         ! pas de point...
         if ( i > s ) then
            BASE_STEP = 1
            PAS = get_int(buf)
         else
            if ( (s-i) >= 10 ) then
              call stop_application("** Bad definition of step **")
            end if
            BASE_STEP = 10**(s-i)
            buf=trim(buf(:i-1))//trim(buf(i+1:))
            do while ( buf(1:1)=='0' )
               buf=trim(buf(2:))
            end do
            PAS = get_int(buf)
         end if

         call log_mess("BASE STEP : "//trim(str(BASE_STEP)),DEBUG_DEF)
         call log_mess("     STEP : "//trim(str(PAS)),DEBUG_DEF)

       end subroutine set_base_and_step
!!***

!!****f* m_qtlmap_tools/get_long_step_morgan
!!  NAME
!!    get_long_step_morgan
!!  DESCRIPTION
!!    Get the the step in Morgan
!!
!! SOURCE
       function get_long_step_morgan() result(l)
         real(kind=dp) :: l

         l = real(pas) / BASE_STEP

       end function get_long_step_morgan
!!***

!!****f* m_qtlmap_tools/set_absi
!! NAME
!!    set_absi
!! DESCRIPTION
!!    build the global array absi. absi is a corresponding vector with a sampled positon.
!!    absi(ch,iposi)=<position in morgan of iposi>, iposi is an integer  1<=iposi<=get_ilong(ch)
!! HISTORY
!!  07/09/2010 - add correspond vector with marker position
!! SOURCE
       subroutine set_absi()
         integer :: i,chr,l,n,ix,ilk,nlong
         real :: v,lastv,lastvm,lastvf
         logical :: vOk,vmOk,vfOk

         integer ,allocatable , dimension(:,:) :: corr_average_absi,corr_mal_absi,corr_femal_absi

         allocate (corr_average_absi(nchr,maxval(nmk)))
         allocate (corr_mal_absi(nchr,maxval(nmk)))
         allocate (corr_femal_absi(nchr,maxval(nmk)))
         corr_average_absi=0
         corr_mal_absi=0
         corr_femal_absi=0

         if ( nchr == 0 ) then
           call stop_application("Devel error: Can not set absci before the map structure!")
         end if

         allocate (absi(nchr,get_maxnpo()))

         do chr=1,nchr
           l = get_ilong(chr)
           n=0
           do ix=0,l,pas
              n=n+1
              absi(chr,n)=get_dx(chr,ix)
           end do

          nlong=n
          do ilk=1,nmk(chr)
            lastv = 10000.d0;lastvm = 10000.d0;lastvf = 10000.d0
            vOk = .false.
            vmOk = .false.
            vfOk = .false.
            do n=1,nlong
             !! average map

               v = ( posi(chr,ilk) - absi(chr,n) ) / posi(chr,ilk)

               if (  v <= 0.0 .or. n == nlong ) then

                 if ( v < 0.d0 ) then ! maybe the last value was nearest....
                   if ( abs(v) < abs(lastv) )  then
                      corr_average_absi(chr,ilk)=n
                   else
                      corr_average_absi(chr,ilk)=n-1
                   end if
                 else
                     corr_average_absi(chr,ilk)=n
                 end if
               vOk = .true.
              end if
              lastv = v
              !! mal map

               v = ( posim(chr,ilk) - absi(chr,n) ) / posim(chr,ilk)

               if (  v <= 0.0 .or. n == nlong ) then

                 if ( v < 0.d0 ) then ! maybe the last value was nearest....
                   if ( abs(v) < abs(lastvm) )  then
                      corr_mal_absi(chr,ilk)=n
                   else
                      corr_mal_absi(chr,ilk)=n-1
                   end if
                 else
                     corr_mal_absi(chr,ilk)=n
                 end if
               vmOk = .true.
              end if
              lastvm = v

              !! femal map

               v = ( posif(chr,ilk) - absi(chr,n) ) / posif(chr,ilk)

               if (  v <= 0.0 .or. n == nlong ) then

                 if ( v < 0.d0 ) then ! maybe the last value was nearest....
                   if ( abs(v) < abs(lastvf) )  then
                      corr_femal_absi(chr,ilk)=n
                   else
                      corr_femal_absi(chr,ilk)=n-1
                   end if
                 else
                     corr_femal_absi(chr,ilk)=n
                 end if
               vfOk = .true.
              end if
              lastvf = v


            if ( vOk .and. vmOk .and. vfOk ) exit

            end do
            call log_mess("CH "//str(chr)//" m"//str(ilk)//":"//&
             str(corr_average_absi(chr,ilk))//" "//&
             str(corr_mal_absi(chr,ilk))//" "//&
             str(corr_femal_absi(chr,ilk)),DEBUG_DEF);
           end do
         end do

         deallocate (corr_average_absi)
         deallocate (corr_mal_absi)
         deallocate (corr_femal_absi)
   !      stop
       end subroutine set_absi
!!***

!!****f* m_qtlmap_tools/get_ilong
!!  NAME
!!    get_ilong
!!  DESCRIPTION
!!    Get the number of sampled position among the selected chromosome
!!  INPUTS
!!    index_chr  : index of chromosome
!!  RESULT
!!    The chromosome size in number of position sampled
!!  NOTES
!!  SEE ALSO
!!  m_qtlmap_data/nmk
!!  m_qtlmap_data/posi
!!
!! SOURCE
       function get_ilong(index_chr) result(ilong)
           integer, intent(in)         :: index_chr
           integer                     :: ilong

           if ( index_chr > nchr ) then
             print *,'ind:',index_chr
             print *,'dev error: get_ilong can not access to index chr'
             stop
           end if
           ilong=nint(BASE_STEP*(posi(index_chr,nmk(index_chr))-posi(index_chr,1)))
           return

       end function get_ilong
!!***

!!****f* m_qtlmap_tools/get_dx
!!  NAME
!!    get_dx
!!  DESCRIPTION
!!    Get the position in Morgan
!!
!!  INPUTS
!!    index_chr   : index of chromosome
!!    ix          : the position index
!!
!!  RETURN
!!   the position in Morgan
!!  NOTES
!!
!!  BUGS
!!
!! SOURCE

        function get_dx(index_chr,ix) result(dx)
           integer, intent(in)         :: index_chr,ix
           real(kind=dp)               :: dx

           if ( index_chr > nchr ) then
             print *,'ind:',index_chr
             print *,'dev error: get_ilong can not access to index chr'
             stop
           end if

           dx=posi(index_chr,1)+(dble(ix)/BASE_STEP)
          ! print *,"ix:",ix,"dx:",dx
           return

        end function get_dx
!!***

!!****f* m_qtlmap_tools/get_pos
!!  NAME
!!    get_pos
!!  DESCRIPTION
!!    Get the index of the sampled position
!!
!!  INPUTS
!!    posInMorgan : position in Morgan
!!
!!  RETURN
!!    the index of sampled position
!!  NOTES
!!
!!  BUGS
!!
!! SOURCE
        function get_pos(posInMorgan) result(npo)
           real(kind=dp), intent(in)         :: posInMorgan
           integer                     :: npo

           npo=int(BASE_STEP*posInMorgan)
           npo=nint(real(npo)/real(PAS))

           return

        end function get_pos
!!***

!!****f* m_qtlmap_tools/get_npo
!!  NAME
!!    get_npo
!!  DESCRIPTION
!!    Getting the number of postion depending the chromosome , step analysis and ilong (size of the chromosome)
!!  INPUTS
!!    * index_chr     -- Chromosome number
!!  RESULT
!!    The number of position to computing Probabilities of transmission and likelihood
!!  EXAMPLE
!!
!!  NOTES
!!
!!  BUGS
!!
!!  SEE ALSO
!!  DATA/m_qtlmap_data/get_ilong
!! SOURCE
       function get_npo(index_chr) result(npo)
           integer, intent(in)         :: index_chr
           integer                     :: npo
           npo =  (get_ilong(index_chr) / pas)+1
           return

       end function get_npo
!!***

!!****f* m_qtlmap_tools/get_maxnpo
!!  NAME
!!    get_npo
!!  DESCRIPTION
!!    Getting the maximum npo
!!  RESULT
!!    maximum number of position on a chromosome
!!  EXAMPLE
!!
!!  NOTES
!!
!!  BUGS
!!
!!  SEE ALSO
!!  m_qtlmap_data/get_ilong
!! SOURCE
       function get_maxnpo() result(maxnpo)
           integer                     :: chr,npo,maxnpo

           maxnpo=0

           do chr=1,nchr
             npo = get_npo(chr)
             maxnpo = max(maxnpo,npo)
           end do

           return

       end function get_maxnpo
!!***

!!****f* m_qtlmap_tools/init_pheno
!!  NAME
!!    init_pheno
!!  DESCRIPTION
!!     allocate arrays to manage the internal encoded format of allele : value_pheno,nb_value_pheno
!!  INPUTS
!!    nch : number of chromosome to manage
!!  NOTES
!!
!!  BUGS
!!
!!  SEE ALSO
!! SOURCE
       subroutine init_pheno(nch)
         integer ,intent(in) :: nch
         allocate (value_pheno(nch,256))
         allocate (nb_value_pheno(nch))
         value_pheno=""
         nb_value_pheno=0

       end subroutine init_pheno
!!***

!!****f* m_qtlmap_tools/set_pheno
!!  NAME
!!    set_pheno
!!  DESCRIPTION
!!    Setting a new identifiant number allele with allele string value or get the value of the identifiant
!!    if this value exist.
!!    * Reallocation or Allocation (first call) of the value_pheno for each call
!!  INPUTS
!!    * value     -- Value of the allele string
!!  RESULT
!!    The idenfiant allele number
!!  EXAMPLE
!!
!!  NOTES
!!
!!  BUGS
!!
!!  SEE ALSO
!!  m_qtlmap_data/value_pheno
!!  m_qtlmap_data/VAL_MIN_INDEX_PHENO
!!  m_qtlmap_data/VAL_MAX_INDEX_PHENO
!! SOURCE
       function set_pheno(c,value) result(res)
           integer                 ,intent(in)  :: c
           character(len=LEN_DEF)  ,intent(in)  :: value
           integer(kind=KIND_PHENO)           :: res
           integer                            :: nbvalue
           integer                            :: i,ch
           character(len=LEN_DEF) , dimension (:,:)    ,allocatable  :: buf_value_pheno
           res = VAL_MIN_INDEX_PHENO

           do i=1,nb_value_pheno(c)
              if ( value_pheno(c,i) == trim(value) ) then
                  return
              end if
              res = res + 1
           end do

           !manage error
           if ( i > VAL_MAX_INDEX_PHENO ) then
             nbvalue = VAL_MAX_INDEX_PHENO
             nbvalue = nbvalue - VAL_MIN_INDEX_PHENO

             write (0,*) "* Too many allele detected for internal data structure "
             write (0,*) "number max [",nbvalue,']'
             do ch=1,nchr
             print *,"------chromo : "//trim(chromo(ch))//"---------- size: ",nb_value_pheno(ch)
             print *,(trim(value_pheno(ch,i))//",",i=1,nb_value_pheno(ch))
             end do
             stop
           end if

           if ( size(value_pheno,2) <= nb_value_pheno(c) ) then
            allocate (buf_value_pheno(nchr,size(value_pheno,2)))
            buf_value_pheno = value_pheno
            deallocate(value_pheno)
            allocate( value_pheno(nchr,size(buf_value_pheno,2)+256) )
            value_pheno=""
            do ch=1,nchr
             do i=1,nb_value_pheno(ch)
              value_pheno(ch,i)=trim(buf_value_pheno(ch,i))
             end do
            end do
            deallocate(buf_value_pheno)
           end if

           nb_value_pheno(c)=nb_value_pheno(c)+1
           value_pheno(c,nb_value_pheno(c))=trim(value)
           return

       end function set_pheno
!!***

!!****f* m_qtlmap_tools/get_pheno
!!  NAME
!!    get_pheno
!!  DESCRIPTION
!!    Get the corresponding identifiant number of a coded initial allele string
!!  INPUTS
!!    * value    -- Value of the allele string
!!  RESULT
!!    The string value of the allele
!!  EXAMPLE
!!
!!  NOTES
!!
!!  BUGS
!!
!!  SEE ALSO
!!  m_qtlmap_data/value_pheno
!!  m_qtlmap_data/VAL_MIN_INDEX_PHENO
!! SOURCE
       function get_pheno(c,value) result(res)
           integer                   ,intent(in)  :: c
           integer(kind=KIND_PHENO)  ,intent(in)  :: value
           character(len=LEN_DEF)                 :: res
           integer                                :: ind,i

           ind = value - VAL_MIN_INDEX_PHENO + 1
           if ( ind < 1 .or. ind > nb_value_pheno(c)) then
              write (0,*) "Devel error : get_pheno * none index correspond [",value,"]"
              print *, (trim(value_pheno(c,i))//" ",i=1,nb_value_pheno(c))
              stop
           end if
           res = value_pheno(c,ind)
       end function get_pheno
!!***

!!****f* m_qtlmap_tools/get_ind_pheno
!!  NAME
!!    get_ind_pheno
!!  DESCRIPTION
!!    Get the corresponding internal identifiant number of a coded initial allele string
!!  INPUTS
!!    * c      : the chromosome
!!    * value  : Value of the string allele
!!  RESULT
!!    The identifiant value of the value allele
!!  EXAMPLE
!!
!!  NOTES
!!
!!  BUGS
!!
!!  SEE ALSO
!!  m_qtlmap_data/value_pheno
!!  m_qtlmap_data/VAL_MIN_INDEX_PHENO
!! SOURCE
       function get_ind_pheno(c,value) result(res)
           integer                 ,intent(in)  :: c
           character(len=LEN_DEF)  ,intent(in)  :: value
           integer(kind=KIND_PHENO)             :: res
           integer :: i

            do i=1,nb_value_pheno(c)
              if ( value_pheno(c,i) == value ) then
                  res = VAL_MIN_INDEX_PHENO + i + 1
                  return
              end if
           end do
           write (0,*) "Devel error : get_ind_pheno * none value correspond [",trim(value),"]"
           stop
       end function get_ind_pheno
!!***

!!****f* m_qtlmap_tools/release_data
!!  NAME
!!    release_data
!!  DESCRIPTION
!!   release all arrays from the module m_qtlmap_data
!! SOURCE
   subroutine release_data
     integer :: ic

     deallocate (repro)
     deallocate (gpere)
     deallocate (gmere)
     deallocate (pere)
     deallocate (mere)
     deallocate (ngmgp)
     deallocate (nrgm)
     deallocate (reppere)
     deallocate (repmere)
     deallocate (animal)
     deallocate (nmp)
     deallocate (ndm)
     deallocate (femelle)
     deallocate (repfem)

     deallocate (nmk)
     deallocate (mark)
     if (allocated(ch)) deallocate (ch)
     deallocate (posi)
     deallocate (posim)
     deallocate (posif)
     deallocate (rm)
     deallocate (rf)
     deallocate (absi)

     deallocate (numero)
     deallocate (pheno)
     deallocate (value_pheno)
     deallocate (nb_value_pheno)
     deallocate (corregp)
     deallocate (corregm)
     deallocate (correr)
     deallocate (correp)
     deallocate (correm)
     deallocate (corred)
     deallocate (nall)
     deallocate (presentg)
     deallocate (alleles)
     deallocate (pc_all)

     if (allocated(bete)) deallocate (bete)
     deallocate (modele)
     deallocate (carac)
     deallocate (natureY)
     deallocate (y)
     deallocate (cd)
     deallocate (xmut)
     deallocate (sigt)
     deallocate (ydiscretord)
     if (allocated (ycategorial)) deallocate (ycategorial)
     deallocate (indicemod)
     deallocate (nmod)
     deallocate (seuil)
     deallocate (prop)
     deallocate (presentc)
     deallocate (ndelta)
     deallocate (corperf)
     deallocate (namefix)
     deallocate (nlev)
     deallocate (listelev)
     deallocate (nivx)
     deallocate (namecov)
     deallocate (covar)
     deallocate (chromo)

     deallocate (estime)
     deallocate (nmumest)
     deallocate (namest)
     if (allocated(iam)) deallocate (iam)
     deallocate (estfem)

     if (allocated(h2)) deallocate (h2)
     if (allocated(RhoP)) deallocate (RhoP)
     if (allocated(RhoG)) deallocate (RhoG)
     if (allocated(ue)) deallocate (ue)
     if (allocated(soglia)) deallocate (soglia)
     if (allocated(freqmodsimultrait)) deallocate (freqmodsimultrait)
     if (allocated(nbmodsimultrait)) deallocate (nbmodsimultrait)
     if (allocated(posiqtl)) deallocate (posiqtl)
     if (allocated(chrqtl)) deallocate (chrqtl)
     if (allocated(nbmodsimultrait)) deallocate (xlim)

     if (allocated (listModelTrait)) then
      do ic=1,ncar
       call release_model_trait(listModelTrait(ic))
      end do
      deallocate (listModelTrait)
     end if

   end subroutine release_data
!!***

!!****f* m_qtlmap_tools/new_lrt_solution
!!  NAME
!!    new_lrt_solution
!!  DESCRIPTION
!!   initialize a variable of type TYPE_LRT_SOLUTION
!!  INPUTS/OUTPUTS
!!    lrtsol   : the variable to initialize
!! SOURCE
      subroutine new_lrt_solution(nqtl,lrtsol)
        integer   , intent(in) :: nqtl
        type(TYPE_LRT_SOLUTION) , intent(inout) :: lrtsol
        integer  :: npo,i,iq,ip,jm
        ! solution under hypothesis nqtl
        lrtsol%nqtl=nqtl

        ! begin index <--0
        allocate (lrtsol%lrtmax(0:nqtl-1))
        allocate (lrtsol%nxmax(0:nqtl-1))
        allocate (lrtsol%chrmax(0:nqtl-1))

        lrtsol%lrtmax = 0.d0
        lrtsol%nxmax = 0
        lrtsol%chrmax = 0

        if ( nqtl == 1 .or. nqtl == 2 ) then
          npo = get_maxnpo()
          if ( nqtl == 1) then
           allocate (lrtsol%lrt1(nchr,npo))
           allocate (lrtsol%pater_eff(nchr,np,npo))
           allocate (lrtsol%mater_eff(nchr,nm,npo))
           allocate (lrtsol%xlrp(nchr,np,npo))
           allocate (lrtsol%xlrm(nchr,nm,npo))
           lrtsol%pater_eff=0.d0
           lrtsol%mater_eff=0.d0
           lrtsol%xlrp=0.d0
           lrtsol%xlrm=0.d0
          end if

          if ( nqtl == 2 ) then
            allocate (lrtsol%lrt0_2(nchr,nchr,npo,npo))
            allocate (lrtsol%lrt1_2(nchr,nchr,npo,npo))
            allocate (lrtsol%pater_eff2(nchr,nchr,np,npo,npo,2))
            allocate (lrtsol%mater_eff2(nchr,nchr,nm,npo,npo,2))
            allocate (lrtsol%xlrp2(nchr,nchr,np,npo,npo))
            allocate (lrtsol%xlrm2(nchr,nchr,nm,npo,npo))
            lrtsol%lrt0_2=0.d0
            lrtsol%lrt1_2=0.d0
            lrtsol%pater_eff2=0.d0
            lrtsol%mater_eff2=0.d0
            lrtsol%xlrp2=0.d0
            lrtsol%xlrm2=0.d0
          end if
       end if
      end subroutine new_lrt_solution
!!***

!!****f* m_qtlmap_tools/release_lrt_solution
!!  NAME
!!    release_lrt_solution
!!  DESCRIPTION
!!   release a variable of type TYPE_LRT_SOLUTION
!!  INPUTS/OUTPUTS
!!    lrtsol   : the variable to release
!! SOURCE
      subroutine release_lrt_solution(lrtsol)
       type(TYPE_LRT_SOLUTION) , intent(inout) :: lrtsol
       integer :: ip,jm

       if ( lrtsol%nqtl < 0 ) return

       if ( associated (lrtsol%lrtmax) ) deallocate (lrtsol%lrtmax)
       if ( associated (lrtsol%chrmax) ) deallocate (lrtsol%chrmax)
       if ( associated (lrtsol%nxmax) ) deallocate (lrtsol%nxmax)

       if ( lrtsol%nqtl == 1 ) then
        if ( associated (lrtsol%lrt1) ) deallocate (lrtsol%lrt1)
        if ( associated (lrtsol%pater_eff) ) deallocate (lrtsol%pater_eff)
        if ( associated (lrtsol%mater_eff) ) deallocate (lrtsol%mater_eff)
        if ( associated (lrtsol%xlrp) ) deallocate (lrtsol%xlrp)
        if ( associated (lrtsol%xlrm) ) deallocate (lrtsol%xlrm)
       end if

       if ( lrtsol%nqtl == 2 ) then
         if ( associated (lrtsol%lrt0_2) ) deallocate (lrtsol%lrt0_2)
         if ( associated (lrtsol%lrt1_2) ) deallocate (lrtsol%lrt1_2)
         if ( associated (lrtsol%pater_eff2) ) deallocate (lrtsol%pater_eff2)
         if ( associated (lrtsol%mater_eff2) ) deallocate (lrtsol%mater_eff2)
         if ( associated (lrtsol%xlrp2) ) deallocate (lrtsol%xlrp2)
         if ( associated (lrtsol%xlrm2) ) deallocate (lrtsol%xlrm2)
       end if

      end subroutine release_lrt_solution
!!***

!!****f* m_qtlmap_tools/clean_lrt_solution
!!  NAME
!!    clean_lrt_solution
!!  DESCRIPTION
!!   release all sub data of TYPE_LRT_SOLUTION
!!  INPUTS/OUTPUTS
!!    lrtsol   : the variable to release
!! SOURCE
      subroutine clean_lrt_solution(lrtsol)
       type(TYPE_LRT_SOLUTION) , intent(inout) :: lrtsol
       integer :: ip,jm

       if ( lrtsol%nqtl < 0 ) return

       if ( lrtsol%nqtl == 1 ) then
        if ( associated (lrtsol%lrt1) ) deallocate (lrtsol%lrt1)
        if ( associated (lrtsol%pater_eff) ) deallocate (lrtsol%pater_eff)
        if ( associated (lrtsol%mater_eff) ) deallocate (lrtsol%mater_eff)
        if ( associated (lrtsol%xlrp) ) deallocate (lrtsol%xlrp)
        if ( associated (lrtsol%xlrm) ) deallocate (lrtsol%xlrm)
       end if

       if ( lrtsol%nqtl == 2 ) then
         if ( associated (lrtsol%lrt0_2) ) deallocate (lrtsol%lrt0_2)
         if ( associated (lrtsol%lrt1_2) ) deallocate (lrtsol%lrt1_2)
         if ( associated (lrtsol%pater_eff2) ) deallocate (lrtsol%pater_eff2)
         if ( associated (lrtsol%mater_eff2) ) deallocate (lrtsol%mater_eff2)
         if ( associated (lrtsol%xlrp2) ) deallocate (lrtsol%xlrp2)
         if ( associated (lrtsol%xlrm2) ) deallocate (lrtsol%xlrm2)
       end if

      end subroutine clean_lrt_solution
!!***

!!****f* m_qtlmap_tools/release_incidence_solution
!!  NAME
!!    release_incidence_solution
!!  DESCRIPTION
!!   release a variable of type TYPE_INCIDENCE_SOLUTION
!!  INPUTS/OUTPUTS
!!    obj   : the variable to release
!! SOURCE
      subroutine release_incidence_solution(obj)
        type(TYPE_INCIDENCE_SOLUTION) , intent(inout) :: obj

        if ( associated (obj%sig) ) deallocate (obj%sig)
        if ( associated (obj%unknown_dam_sig) ) deallocate (obj%unknown_dam_sig)
        if ( associated (obj%paramaterValue) ) deallocate (obj%paramaterValue)
        if ( associated (obj%parameterVecsol) ) deallocate (obj%parameterVecsol)
        if ( associated (obj%eqtl_print) ) deallocate (obj%eqtl_print)
        if ( associated (obj%parameterPrecis) ) deallocate (obj%parameterPrecis)
        if ( associated(obj%rhoi) ) deallocate (obj%rhoi)

        if ( associated (obj%groupeName) ) deallocate (obj%groupeName)
        if ( associated (obj%nbParameterGroup) ) deallocate (obj%nbParameterGroup)
        if ( associated (obj%parameterName) ) deallocate (obj%parameterName)
        if (associated(obj%qtl_groupeName)) deallocate (obj%qtl_groupeName)

      end subroutine release_incidence_solution
!!***

!!****f* m_qtlmap_tools/copy_description_type
!!  NAME
!!    copy_description_type
!!  DESCRIPTION
!!   copy the variable in_desc of type TYPE_INCIDENCE_SOLUTION to out_desc
!!  INPUTS
!!    in_desc  : the variable to copy
!!    out_desc : the copy
!! SOURCE
      recursive subroutine copy_description_type(in_desc,out_desc)
       type(DESC_EFFECT_TYPE) , intent(in) :: in_desc
       type(DESC_EFFECT_TYPE) , intent(out) :: out_desc
       integer :: i

        out_desc%name =  in_desc%name
        out_desc%start =  in_desc%start
        out_desc%end =  in_desc%end
        out_desc%haveSubDesc =  in_desc%haveSubDesc

        if ( in_desc%haveSubDesc ) then
          allocate ( out_desc%listSubDesc(size(in_desc%listSubDesc)) )
          do i=1,size(in_desc%listSubDesc)
            call copy_description_type(in_desc%listSubDesc(i),out_desc%listSubDesc(i))
          end do
        end if

    end subroutine copy_description_type
!!***

!!****f* m_qtlmap_tools/release_description_type
!!  NAME
!!    release_description_type
!!  DESCRIPTION
!!   release a variable of type DESC_EFFECT_TYPE
!!  INPUTS/OUTPUTS
!!    inout_desc  : the variable to release
!! SOURCE
    recursive subroutine release_description_type(inout_desc)
      type(DESC_EFFECT_TYPE) , intent(inout) :: inout_desc
      integer :: i

        if ( inout_desc%haveSubDesc ) then
          do i=1,size(inout_desc%listSubDesc)
            call release_description_type(inout_desc%listSubDesc(i))
          end do
          deallocate (inout_desc%listSubDesc)
        end if

    end subroutine release_description_type
!!***

!!****f* m_qtlmap_tools/release_model_trait
!!  NAME
!!    release_model_trait
!!  DESCRIPTION
!!   release a variable of type model_trait
!!  INPUTS/OUTPUTS
!!    inout_desc  : the variable to release
!! SOURCE
    recursive subroutine release_model_trait(inout_desc)
      type(model_trait) , intent(inout) :: inout_desc

      if ( associated(inout_desc%nbint) ) then
         deallocate (inout_desc%nbint)
      end if

      if ( associated(inout_desc%indexFixedEffect) ) then
         deallocate (inout_desc%indexFixedEffect)
      end if

      if ( associated(inout_desc%indexCovariate) ) then
         deallocate (inout_desc%indexCovariate)
      end if

      if ( associated(inout_desc%indexFixedEffectWithInteraction) ) then
         deallocate (inout_desc%indexFixedEffectWithInteraction)
      end if

    end subroutine release_model_trait
!!***


!!****f* m_qtlmap_tools/get_maxnbgenotypedam
!!  NAME
!!    get_maxnbgenotypedam
!!  DESCRIPTION
!!   get the maximum number of genotype (all dams confused) found by the haplotype on the genome wide
!!  RESULTS
!!    maxng  : the variable to release
!! SOURCE
      function get_maxnbgenotypedam() result(maxng)
         integer             :: maxng

         maxng = maxval(ngenom(:,nmp(np+1)+1))

      end function get_maxnbgenotypedam
!!***

end module m_qtlmap_tools
