/*
 *
 * O.Filangi : linearisation de la vraissemblance -  calcul simultané des position sur le génome
 * -deviceemu => emulation
 * nvcc -g-G -deviceemu --cuda analyse_whole-genome.cu
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <cuda.h>

#define _HOST_DEBUG_ 0

/* number of thread available by block */
#define MAX_BLOCKDIM 256

#define BLOCK_WORK  16

#define MAXNTNIV_LOCALTHREAD_MEMORY 60

#define DT float

void print_info_memory() {
	unsigned int free,total;
	cuMemGetInfo(&free, &total) ;
	printf("===============> free: %f Mb (%lu bytes)   total:%f Mb (%lu bytes) \n",double(free)/(1024*1024),free,double(total)/(1024*1024),total);
}

class  QTLMapStructDeviceDataLinear {
public:
	DT * contingence ; /* matrice de contingence les effets statiques sont dans les premieres colonne. */
	DT * Y          ; /* Vecteur de Performance Y                           [NSIM][NDMAX]        */
	int nqtl        ; /* hypothese courante */
	int np          ; /* nombre de famille de pere */
	int *sizeFam    ; /* *nombre de descendant par famille de pere          [NP] */
	int nsim        ; /* nombre de simulation */
	int npos        ; /* nombre de position teste */
	int ndmax       ; /* taille du tableau Y */
	int nd          ; /* nombre de descendants a prendre en compte */
	int nLevelFix   ; /* nombre d effet non dependant de la position */
	int nLevelVar   ; /* nombre d'effet dependant de la position */
	int ntniv       ; /* nLevelFix + nLevelVar */
	int * corrIndexCol ; /* ordre des effets a prendre en compte */
	DT seuil_cho    ; /* seuil pour l estimation des effets */

public:
	void init(   DT * xinc_d,
			DT * Y_d,
			int *corrLevelColPtr,
			double *seuil_choPtr,
			int * ndPtr,
			int * nkdPtr,
			int * nsimPtr,
			int * npositionPtr,
			int * nLevelFixPtr,
			int * nLevelVarPtr,
			int * npPtr,
			int * sizeFamilyNp);

	virtual ~QTLMapStructDeviceDataLinear();
	void releaseDevice();
} ;

class QTLMapStructDeviceWorkLinear {
public :
	DT * XX         ; /* matrice d'incidence                                [NPOS][NTNIV][NTNIV] */
	DT * triang     ; /* matrice temporaire de la decomposition Cholesky    [NPOS][NTNIV][NTNIV] */
	DT * rhs        ; /* Vecteur RHS                                        [NPOS][NSIM][NTNIV]  */
	DT * isigsq     ; /* variance residuelle des Hypothese < NQTL           [NSIM][NP] */
public :
	/* init de XX, triang, rhs */
	void initResolution(const QTLMapStructDeviceDataLinear & data);

	/* init de isigsq */
	void initCalculLRT(const QTLMapStructDeviceDataLinear & data,DT * sigsquare_host);

	virtual ~QTLMapStructDeviceWorkLinear();

	void releaseDeviceResolution();

	void releaseDeviceCalculLRT();

	/* recuperation du tableau vecsol stoque sur le device */
	void getAllVecsolDevice(const QTLMapStructDeviceDataLinear & data,int nbBloc,int nbThread,int *vecsol) ;

} ;

class QTLMapStructDeviceSolutionLinear {
public:
	DT * osigsq     ; /* Vecteur des variances residuelles                  [NPOS][NSIM][NP]     */
	DT * bestim     ; /* Vecteur des solution des effets                    [NPOS][NSIM][NTNIV ] */
	DT * lrt        ; /* rapport de vraisemblance                           [NPOS][NSIM] */
public :
	void init(const QTLMapStructDeviceDataLinear & data);

	virtual ~QTLMapStructDeviceSolutionLinear();

	void getAllMaxLrt(const QTLMapStructDeviceDataLinear & data,int nbBloc,int nbThread,DT * lrtMax,int * posMax);

	void releaseDevice();

} ;

template <class T>
class Utils {
public:
	void printFloatDeviceArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D);
	void printFloatDeviceArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D);
	void printFloatHostArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D);
	void printFloatHostArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D);
	static void getArrayDeviceToHost(int nbBloc,int nbThread,int size,T * inDeviceArray, T *outHostArray);
};

/*******************************************************************************************************************************************/

void QTLMapStructDeviceDataLinear::init(DT * xinc_d,
		DT * Y_d,
		int *corrLevelColPtr,
		double *seuil_choPtr,
		int * ndPtr,
		int * nkdPtr,
		int * nsimPtr,
		int * npositionPtr,
		int * nLevelFixPtr,
		int * nLevelVarPtr,
		int * npPtr,
		int * sizeFamilyNp) {

	printf("*********************** INIT DATA *********************************\n");
	seuil_cho   = (DT) *seuil_choPtr;
	seuil_cho   = sqrt(seuil_cho);
	ndmax       = *ndPtr       ; /* taille reelle de la matrice de contingence : lignes */
	nd          = *nkdPtr      ; /* taille de la population pris en compte */
	nsim        = *nsimPtr     ;
	npos        = *npositionPtr;
	nLevelFix   = *nLevelFixPtr ;
	nLevelVar   = *nLevelVarPtr;
	np          = *npPtr;
	ntniv       = nLevelFix + nLevelVar ;

	printf("seuil_cho  SQ : %f\n",seuil_cho);
	printf("ndmax         : %d\n",ndmax);
	printf("nd            : %d\n",nd);
	printf("nsim          : %d\n",nsim);
	printf("np            : %d\n",np);
	printf("npos          : %d\n",npos);
	printf("nLevelFix     : %d\n",nLevelFix);
	printf("nLevelVar     : %d\n",nLevelVar);
	for (int i=0;i<np;i++) printf("sizeFamilyNp[%d] : %d\n",i,sizeFamilyNp[i]);
	for (int i=0;i<ntniv;i++) printf("corrLevelCol[%d] : %d\n",i,corrLevelColPtr[i]);
	printf("Y[1]         : %f\n",Y_d[0]);
	printf("Y[%d]        : %f\n",Y_d[nd-1]);


	/* Y */
	int size = nsim * ndmax * sizeof(DT);
	int err = cudaMalloc(&Y,size);
	if ( err ) {
		printf("Error allocation Y ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	err = (int)cudaMemcpy(Y, Y_d, size, cudaMemcpyHostToDevice);
	if ( err ) {
		printf("Error copy host to device Y ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	/* contingence */
	size = ndmax*(nLevelFix+(nLevelVar)*npos)*sizeof(DT) ;
	err = cudaMalloc(&contingence,size);
	if ( err ) {
		printf("Error allocation contingence ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	err = (int)cudaMemcpy(contingence, xinc_d, size, cudaMemcpyHostToDevice);
	if ( err ) {
		printf("Error copy host to device contingence ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	/* sizeFamilyNp */
	size = np*sizeof(int) ;
	err = cudaMalloc(&sizeFam,size);
	if ( err ) {
		printf("Error allocation sizeFam ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	err = (int)cudaMemcpy(sizeFam, sizeFamilyNp, size, cudaMemcpyHostToDevice);
	if ( err ) {
		printf("Error copy host to device sizeFam ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	/* corrIndex */
	size=ntniv * sizeof(int);
	err = cudaMalloc(&corrIndexCol,size);
	if ( err ) {
		printf("Error allocation corrIndexCol ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	err = (int)cudaMemcpy(corrIndexCol, corrLevelColPtr, size, cudaMemcpyHostToDevice);
	if ( err ) {
		printf("Error copy host to device corrIndexCol ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	print_info_memory();
}

QTLMapStructDeviceDataLinear::~QTLMapStructDeviceDataLinear() {
}

void QTLMapStructDeviceDataLinear::releaseDevice(){
	cudaFree(contingence);
	cudaFree(Y);
	cudaFree(sizeFam);
	cudaFree(corrIndexCol);
}


void QTLMapStructDeviceWorkLinear::initCalculLRT(const QTLMapStructDeviceDataLinear & data,DT * sigsquare_host) {
	printf("*********************** INIT ISIGSQ *********************************\n");
	/* isigsq */
	int size = data.np*data.nsim*sizeof(DT) ;
	int err = cudaMalloc(&isigsq,size);
	if ( err ) {
		printf("Error allocation  sigsquare_device ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	err = (int)cudaMemcpy(isigsq, sigsquare_host, size, cudaMemcpyHostToDevice);
	if ( err ) {
		printf("Error copy host to device sigsquare_device ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	print_info_memory();
}

void QTLMapStructDeviceWorkLinear::releaseDeviceCalculLRT() {
	printf("*********************** END ISIGSQ *********************************\n");
	cudaFree(isigsq);
}



void QTLMapStructDeviceWorkLinear::initResolution(const QTLMapStructDeviceDataLinear & data) {
	printf("*********************** INIT WORK *********************************\n");
	/* XX */
	int size = data.ntniv*data.ntniv*data.npos*sizeof(DT) ;
	int err = cudaMalloc(&XX,size);
	if ( err ) {
		printf("Error allocation XX ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}

	/* triang */
	err = cudaMalloc(&triang,size);
	if ( err ) {
		printf("Error allocation triang ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	/* RHS */
	size = data.nsim*data.npos*data.ntniv*sizeof(DT);
	err = cudaMalloc(&rhs,size);
	if ( err ) {
		printf("Error allocation rhs ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	print_info_memory();
}

QTLMapStructDeviceWorkLinear::~QTLMapStructDeviceWorkLinear() {

}

void QTLMapStructDeviceWorkLinear::releaseDeviceResolution(){
	cudaFree(rhs);
	cudaFree(XX);
	cudaFree(triang);
}

__global__ void calcul_vecsol(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,int *vec_device) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */

	if (ipos < data.npos ) {
		for(int i=0;i<data.ntniv;i++) {
			vec_device[ipos+i*data.npos] = (work.triang[ipos+i*data.npos+i*data.npos*data.ntniv] > data.seuil_cho );
		}
	}

}

void QTLMapStructDeviceWorkLinear::getAllVecsolDevice(const QTLMapStructDeviceDataLinear & data,int nbBloc,int nbThread,int * vecsol) {

	int * vec_device;
	int size = data.npos*data.ntniv*sizeof(int);
	int errInt = cudaMalloc(&vec_device,size);
	if ( errInt ) {
		printf("Error allocation memory GPU QTLMapStructDeviceWorkLinear::getAllVecsolDevice = %d \n",errInt);
		exit(1);
	}

	calcul_vecsol<<<nbBloc,nbThread>>>(data,*this,vec_device);
	cudaThreadSynchronize();
	cudaError_t err = cudaMemcpy(vecsol, vec_device, size, cudaMemcpyDeviceToHost);
	if ( err ) {
		printf("Error QTLMapStructDeviceWorkLinear::getAllVecsolDevice = %d \n",err);
		exit(1);
	}
	cudaFree(vec_device);
}

void QTLMapStructDeviceSolutionLinear::init(const QTLMapStructDeviceDataLinear & data) {
	printf("*********************** INIT SOLUTION *********************************\n");
	/* bestim */
	int size = data.nsim*data.npos*data.ntniv*sizeof(DT);
	int err = cudaMalloc(&bestim,size);
	if ( err ) {
		printf("Error allocation bestim ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	/* osigsq */
	size = data.nsim*data.npos*data.np*sizeof(DT);
	err = cudaMalloc(&osigsq,size);
	if ( err ) {
		printf("Error allocation osigsq ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}
	/* lrt */
	size = data.nsim*data.npos*sizeof(DT);
	err = cudaMalloc(&lrt,size);
	if ( err ) {
		printf("Error allocation lrt ERR=%d , SIZE=%d \n",err,size);
		exit(1);
	}

	print_info_memory();

}

QTLMapStructDeviceSolutionLinear::~QTLMapStructDeviceSolutionLinear() {

}

/**
 *
 * set_maxLRT
 *
 *
 */
__global__ void set_maxLRT(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceSolutionLinear solution,DT* lrtMax,int * posMax) {
	int isim = blockIdx.x * blockDim.x + threadIdx.x ; /* index de la simul  */

	/******* DESCENT LU *****/
	if ( isim < data.nsim ) {
		DT  mL = -999.9;
		int pMax;
		for (int i=0;i<data.npos;i++) {
			if (mL < solution.lrt[i]) {
				mL   = solution.lrt[i];
				pMax = i+1;
			}
		}
		lrtMax[isim] = mL;
		posMax[isim] = pMax;
	}
}


void QTLMapStructDeviceSolutionLinear::getAllMaxLrt(const QTLMapStructDeviceDataLinear & data,int nbBloc,int nbThread,DT * lrtMax,int * posMax) {

	int * posMax_device;
	DT  * lrtMax_device;
	int size = data.nsim ;

	int errInt = cudaMalloc(&posMax_device,size*sizeof(int));
	if ( errInt ) {
		printf("Error allocation memory GPU QTLMapStructDeviceSolutionLinear::getAllMaxLrt = %d \n",errInt);
		exit(1);
	}
	errInt = cudaMalloc(&lrtMax_device,size*sizeof(DT));
	if ( errInt ) {
		printf("Error allocation memory GPU QTLMapStructDeviceSolutionLinear::getAllMaxLrt = %d \n",errInt);
		exit(1);
	}

	set_maxLRT<<<nbBloc,nbThread>>>(data,*this,lrtMax_device,posMax_device);
	cudaThreadSynchronize();
	cudaError_t err = cudaMemcpy(lrtMax, lrtMax_device, size*sizeof(DT), cudaMemcpyDeviceToHost);
	if ( err ) {
		printf("Error QTLMapStructDeviceSolutionLinear::getAllMaxLrt MEMCOPY DEVICE TO HOST = %d \n",err);
		exit(1);
	}
	err = cudaMemcpy(posMax, posMax_device, size*sizeof(int), cudaMemcpyDeviceToHost);
	if ( err ) {
		printf("Error QTLMapStructDeviceSolutionLinear::getAllMaxLrt MEMCOPY DEVICE TO HOST = %d \n",err);
		exit(1);
	}
	cudaFree(lrtMax_device);
	cudaFree(posMax_device);

}



void QTLMapStructDeviceSolutionLinear::releaseDevice(){
	cudaFree(bestim);
	cudaFree(osigsq);
	cudaFree(lrt);
}

template <class T>
void Utils<T>::printFloatDeviceArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D) {
	int size = dim1*dim2;

	T array2DHost[size] ;
	cudaError_t err = cudaMemcpy(array2DHost, array2D, size*sizeof(T), cudaMemcpyDeviceToHost);
	if ( err == 0 ) {
		printf("****************************\n");
		for (int i=0;i<nb1;i++ ) {
			for (int j=0;j<nb2;j++ ) {
				printf(" %f ",(float)array2DHost[dim1*j+i]);
			}
		}
		printf("\n -- \n");
	} else {
		printf("Error Memcpy %d\n",err);
		exit(1);
	}
}

template <class T>
void Utils<T>::printFloatDeviceArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D) {
	int size = dim1*dim2*dim3;

	T array3DHost[size] ;
	cudaError_t err = cudaMemcpy(array3DHost, array3D, size*sizeof(T), cudaMemcpyDeviceToHost);
	if ( err == 0 ) {
		printf("****************************\n");
		for (int i=0;i<nb1;i++ ) {
			for (int j=0;j<nb2;j++ ) {
				for (int k=0;k<nb3 ;k++ ) {
					printf(" %f ",(float)array3DHost[dim1*dim2*k+dim1*j+i]);
				}
				printf("\n");
			}
			printf("\n -- \n");
		}
	} else {
		printf("Error Memcpy %d\n",err);
		exit(1);
	}
}

template <class T>
void Utils<T>::printFloatHostArray2D(int dim1,int dim2,int nb1,int nb2,T* array2D) {

}

template <class T>
void Utils<T>::printFloatHostArray3D(int dim1,int dim2,int dim3,int nb1,int nb2,int nb3,T* array3D) {

}

template <class T>
void Utils<T>::getArrayDeviceToHost(int nbBloc,int nbThread,int size,T * inDeviceArray, T *outHostArray) {
	cudaError_t err = cudaMemcpy(outHostArray,inDeviceArray, size*sizeof(T), cudaMemcpyDeviceToHost);
	if ( err ) {
		printf("Error QTLMapStructDeviceWorkLinear::getAllVecsolDevice = %d \n",err);
		exit(1);
	}
}

/*******************************************************************************************************************************************/

/**
 *
 * Modele XT_X
 *
 *
 *
 */
__global__ void XT_X(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	__shared__ int corrIndex[MAXNTNIV_LOCALTHREAD_MEMORY] ;
	if ( threadIdx.x == 0 ) {
		for (int i=0;i<data.ntniv;i++)
			corrIndex[i] = data.corrIndexCol[i];
	}

	__syncthreads();

	int scale1,scale2;
	int ntniv = data.ntniv;

	/* ndmax*npos*iniv+npos*kd+ipos */
	/* access contigence kd + (iniv*NKD) + (ipos*(NKD*NTNIVMAX))  */

	/* Optimisation : Des operations peuvent etre communes a tous les threads si les niveaux dependant de la position sont connus !!!!*/
	/* XINC ACCESS => ndmax*npos*iniv+npos*kd+ipos */
	//ipos+ iniv*maxpos +jniv*ntniv*maxpos ;
	if ( ipos < data.npos ) {
		int k1=0;
		for (int iniv = 0;iniv<ntniv;iniv++) {
			if ( iniv < data.nLevelFix ) {
				scale1 = iniv*data.ndmax;
			} else {
				scale1 = (data.nLevelFix+(ipos * data.nLevelVar)+k1 )*data.ndmax;
				k1++;
			}
			int k2=0;
			for (int jniv=0;jniv<data.ntniv;jniv++) {
				if ( jniv < data.nLevelFix ) {
					scale2 = jniv*data.ndmax;
				} else {
					scale2 = ( data.nLevelFix+(ipos * data.nLevelVar)+k2 )*data.ndmax;
					k2++;
				}
				DT v = 0.0;
				for (int kd = 0;kd<data.nd;kd++ ) {
					v += (data.contingence[scale1 + kd] * data.contingence[scale2 + kd]);
				}
				work.XX[ipos+ corrIndex[iniv]*data.npos +corrIndex[jniv]*ntniv*data.npos] = v;
			}
		}
	}
}

/**
 *
 * Cholesky_Decomposition
 *
 *
 *
 */
__global__ void Cholesky_Decomposition(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {

	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */

	DT * triang = work.triang ;
	int ntniv = data.ntniv;

	if ( ipos < data.npos ) {

		for (int i=0;i<ntniv;i++) {
			for (int j=0;j<ntniv;j++) {
				triang[ipos+i*data.npos+j*ntniv*data.npos] = 0.0;
			}
		}

		for (int j=0;j<data.ntniv;j++) {
			DT v = work.XX[ipos+j*data.npos +j*ntniv*data.npos]; //XX[j][j] ;

			for (int k=0;k<=j-1;k++) {
				// [j][j] = [j][j] - [k][j]*[k][j]
				v -= triang[ipos+k*data.npos+j*ntniv*data.npos]*triang[ipos+k*data.npos+j*ntniv*data.npos];
			}
			//si estimable .. on ne met pas de structure vecseuil==> on utilise la diagonale de triangle pour savoir l effet est estimable
			triang[ipos+j*data.npos +j*ntniv*data.npos] =  sqrt(v);
			if ( triang[ipos+j*data.npos +j*ntniv*data.npos] > data.seuil_cho ) {
				for (int i=j+1;i<ntniv;i++ ) {
					triang[ipos+i*data.npos +j*ntniv*data.npos] = work.XX[ipos+i*data.npos +j*ntniv*data.npos];//XX[i][j];
					for ( int k=0;k<=j-1;k++) {
						triang[ipos+i*data.npos +j*ntniv*data.npos] = triang[ipos+i*data.npos +j*ntniv*data.npos] -  (triang[ipos+i*data.npos +k*ntniv*data.npos] * triang[ipos+j*data.npos +k*ntniv*data.npos]);
					}
					triang[ipos+i*data.npos +j*ntniv*data.npos] = triang[ipos+i*data.npos +j*ntniv*data.npos] / triang[ipos+j*data.npos +j*ntniv*data.npos];
					triang[ipos+j*data.npos +i*ntniv*data.npos] = triang[ipos+i*data.npos +j*ntniv*data.npos] ;
				}
			}
		}
	}

}

/**
 *
 * Set_RHS
 *
 *  RHS [NPOS][NSIM][NTNIV]
 *  Y   [NSIM][NKD]
 */
__global__ void Set_RHS(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	__shared__ int corrIndex[MAXNTNIV_LOCALTHREAD_MEMORY] ;
	if ( threadIdx.x == 0 ) {
		for (int i=0;i<data.nLevelFix+data.nLevelVar;i++)
			corrIndex[i] = data.corrIndexCol[i];
	}

	__syncthreads();
	int ntniv = data.ntniv ;

	if ( ipos < data.npos ) {
		int k=0;
		int scale;
		for (int iniv=0;iniv<ntniv;iniv++) {
			if ( iniv < data.nLevelFix ) {
				scale = iniv*data.ndmax;
			} else {
				scale = (data.nLevelFix+(ipos * data.nLevelVar)+k )*data.ndmax;
				k++;
			}
			for (int isim=0;isim<data.nsim;isim++) {
				int index = ipos + isim*data.npos + corrIndex[iniv]*data.nsim*data.npos ;
				work.rhs[index]=0;
				if (work.triang[ipos+iniv*data.npos+iniv*data.npos*ntniv]>data.seuil_cho) {
					for (int kd = 0;kd<data.nd;kd++ ) {
						work.rhs[index] += data.contingence[scale + kd]*data.Y[isim+kd*data.nsim];
					}
				}
			}
		}
	}
}

/**
 *
 * Resolution LU (Partie descendante descendante )
 *
 *
 */
__global__ void Resolve_LU(QTLMapStructDeviceDataLinear data,
		QTLMapStructDeviceWorkLinear work,
		QTLMapStructDeviceSolutionLinear solution) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */

	int ntniv = data.ntniv ;
	DT temp[MAXNTNIV_LOCALTHREAD_MEMORY];

	/******* DESCENT LU *****/
	if ( ipos < data.npos ) {
		for (int isim=0;isim<data.nsim;isim++) {
			int scale = ipos + isim*data.npos ;
			for (int i=0;i<ntniv;i++) {
				if (work.triang[ipos+i*data.npos+i*ntniv*data.npos]>data.seuil_cho) {
					temp[i] = work.rhs[scale + i*data.nsim*data.npos];
					for (int j=i-1;j>=0;j--) {
						if (work.triang[ipos+j*data.npos+j*ntniv*data.npos]>data.seuil_cho) {
							temp[i] -= temp[j]*work.triang[ipos+i*data.npos+j*ntniv*data.npos];
						}
					}
					temp[i] = temp[i] / work.triang[ipos+i*data.npos+i*ntniv*data.npos];
				}
			}
			for (int i=ntniv-1;i>=0;i--) {
				int indexI = scale + i*data.nsim*data.npos;
				solution.bestim[indexI] = 0 ;
				if (work.triang[ipos+i*data.npos+i*ntniv*data.npos]>data.seuil_cho) {
					solution.bestim[indexI] = temp[i];
					for (int j=i+1;j<ntniv;j++) {
						if (work.triang[ipos+j*data.npos+j*ntniv*data.npos]>data.seuil_cho) {
							int indexJ = scale + j*data.nsim*data.npos;
							solution.bestim[indexI] -= solution.bestim[indexJ]*work.triang[ipos+i*data.npos+j*ntniv*data.npos];
						}
					}
					solution.bestim[indexI] = solution.bestim[indexI] / work.triang[ipos+i*data.npos+i*ntniv*data.npos];
				}
			}
		}//fin for isim
	}
}

/**
 *
 * Set_SIGSQ
 * XB = Y - ( X' . Bestim )
 * SIGSQ = SUM(XB^2)
 */
__global__ void Set_SIGSQ(QTLMapStructDeviceDataLinear data,
		QTLMapStructDeviceWorkLinear work,
		QTLMapStructDeviceSolutionLinear solution) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	__shared__ int corrIndex[MAXNTNIV_LOCALTHREAD_MEMORY] ;
	if ( threadIdx.x == 0 ) {
		for (int i=0;i<data.nLevelFix+data.nLevelVar;i++)
			corrIndex[i] = data.corrIndexCol[i];
	}
	__syncthreads();

	int ntniv = data.ntniv ;

	if ( ipos < data.npos ) {
		for (int isim=0;isim<data.nsim;isim++) {
			int scale = ipos + isim*data.npos ;
			DT xb = 0 ;
			for (int kd = 0;kd<data.nd;kd++ ) {
				int k=0;
				int scale_cont;
				DT v=0;
				for (int iniv=0;iniv<ntniv;iniv++) {
					if ( iniv < data.nLevelFix ) {
						scale_cont = iniv*data.ndmax;
					} else {
						scale_cont = (data.nLevelFix+(ipos * data.nLevelVar)+k )*data.ndmax;
						k++;
					}

					if (work.triang[ipos+iniv*data.npos+iniv*ntniv*data.npos]>data.seuil_cho) {
						v += data.contingence[scale_cont + kd]*solution.bestim[scale + corrIndex[iniv]*data.nsim*data.npos];
					}
				}

				v = data.Y[isim+kd*data.nsim] - v;
				xb += v*v ;
			}
			DT v = xb / data.nd ;

			for (int ip=0;ip<data.np;ip++) {
				solution.osigsq[ipos+isim*data.npos+ip*data.npos*data.nsim] = v ;
			}
		}
	}
}


/**
 *
 * Set_LRT
 *
 *
 */
__global__ void Set_LRT(QTLMapStructDeviceDataLinear data,QTLMapStructDeviceWorkLinear work,QTLMapStructDeviceSolutionLinear solution) {
	int ipos = blockIdx.x * blockDim.x + threadIdx.x ; /* position */
	__shared__ int sizeFam[MAXNTNIV_LOCALTHREAD_MEMORY] ;
	if ( threadIdx.x == 0 ) {
		for (int ip=0;ip<data.np;ip++)
			sizeFam[ip] = data.sizeFam[ip];
	}
	__syncthreads();


	/******* DESCENT LU *****/
	if ( ipos < data.npos ) {
		for (int isim=0;isim<data.nsim;isim++) {
			solution.lrt[ipos+isim*data.npos] = 0 ;
			int k = ipos + isim * data.npos ;
			DT v = 0;
			for (int ip=0;ip<data.np;ip++) {
				v += sizeFam[ip] * (log(work.isigsq[isim+ip*data.nsim]) - log(solution.osigsq[k+ip*data.npos*data.nsim]));
			}
			solution.lrt[ipos+isim*data.npos] = v;
		}//fin for isim
	}
}



/*
  xinc  = matrice d'incidence
  nkd   = nombre de ligne
  ntniv = nombre de colonne
  seuil_choPtr = seuil d'estimabilite pour la solution

  column major order ===> offset = row + column*NUMROWS
  XINC ACCESS => ndmax*npos*iniv+npos*kd+ipos

  XINC(construction : level fixe puis levels variables)
  Mu MuPere1 MuPere2 MuMere1 MuMere2 QTLPere1Pos1 QTLPere2Pos1 QTLMere1Pos1 QTLMere2Pos1 ... QTLPere1PosN QTLPere2PosN QTLMere1PosN QTLMere2PosN

  nLevelFixPtr : 5 (Moyenne gen + 4 effet polygenique)
  nLevelVarPtr : 4 (4 effet qtl a estimer)
  corrLevelColPtr(1) = 1    moyenne generale
  corrLevelColPtr(2) = 6    effet qtl pere 1
  corrLevelColPtr(3) = 7    effet qtl pere 2
  corrLevelColPtr(4) = 8    effet qtl mere 1
  corrLevelColPtr(5) = 9    effet qtl mere 2
  corrLevelColPtr(6) = 2    effet polygenique pere 1
  corrLevelColPtr(7) = 3    effet polygenique pere 2
  corrLevelColPtr(8) = 4    effet polygenique mere 1
  corrLevelColPtr(9) = 5    effet polygenique mere 2
 */
extern "C" void cuda_homoscedatic_resolv_genome_(
		int *nqtlPtr,                         /* IN  : Number of QTL */
		DT  *sigsquare,                       /* IN  : variance residuelle sous les hypothese 0 à NQTL-1 tableau [NSIM][NP] */
		int *nLevelFixPtr,                    /* IN  : nombre de niveaux fixe a la position : moyenne genrale, polygenic, effet fixe, covariables */
		int *nLevelVarPtr,                    /* IN  : nombre de niveau variable a la position */
		int *corrLevelColPtr,                 /* IN  : Correspondance du niveau associe a une colonne de la matrice de contingence */
		DT *xinc_d,                           /* IN  : matrice de contingence [nd,ntnivparam+nbposition]*/
		DT *Y_d,                              /* IN  : performance [nsim,nd]*/
		int *nsimPtr,                         /* IN  : nombre d'individu total  : nombre de ligne reelle de XINC */
		int *ndPtr,                           /* IN  : nombre d'individu total  : nombre de ligne reelle de XINC */
		int *nkdPtr,                          /* IN  : nombre d'individu pris en compte dans les matrices de contingence */
		int *npPtr,                           /* IN  : nombre de famille de pere */
		int *sizeFamilyNp,                    /* IN  : nombre de descendant par pere tableau [NP] */
		int *ntnivmaxPtr,                     /* IN  : nombre maximum de niveau : nombre de colonne reelle de XINC */
		int *npositionPtr,                    /* IN  : number of position estimation */
		double * seuil_choPtr,                /* IN  : seuil pour considere grace a la decomposition de cholesky si un niveau est estimable */
		int    * vecsol,                      /* OUT : tableau de booleen de taille [NPOSITION,NTNIV]  pour l'estimabilite de chaque niveau */
		DT * bestim,                          /* OUT : tableau solution de l'estimation de chaque niveau (taille [NPOSITION,NSIM,NTNIV]) */
		DT * osigsquare,                      /* OUT : tableau des variances residuelles pour chaque familles de pere (taille [NPOSITION,NSIM,NP])  */
		DT * lrt,                             /* OUT : LRT (taille [NPOSITION,NSIM])  */
		DT * maxLRT,                          /* OUT : Value of the Maximul LRT  [NSIM] */
		int *maxPosition                      /* OUT : Index of the position maximum tableau [NSIM] */
) {

	QTLMapStructDeviceDataLinear      data ;
	QTLMapStructDeviceWorkLinear      work ;
	QTLMapStructDeviceSolutionLinear  solution ;

	int err ;
	printf("****************** INFO DEVICE ******************************\n");
	int nbDevice ;
	cudaGetDeviceCount(&nbDevice);
	printf("NB DEVICE     : %d\n",nbDevice);

	cudaDeviceProp  prop;

	for (int i=0;i<nbDevice;i++) {
		cudaGetDeviceProperties(&prop,i);
		printf(" ******* ===  %s  ===  *******\n",prop.name);
		printf("  totalGlobalMem : %d \n",prop.totalGlobalMem);
		printf("  sharedMemPerBlock : %d \n",prop.sharedMemPerBlock);
		printf("  regsPerBlock : %d \n",prop.regsPerBlock);
		printf("  warpSize : %d \n",prop.warpSize);
		printf("  maxThreadsPerBlock : %d \n",prop.maxThreadsPerBlock);
		printf("  maxThreadsDim : %d,%d,%d \n",prop.maxThreadsDim[0],prop.maxThreadsDim[1],prop.maxThreadsDim[2]);
		printf("  maxGridSize : %d,%d,%d \n",prop.maxGridSize[0],prop.maxGridSize[1],prop.maxGridSize[2]);
		printf("  clockRate : %d \n",prop.clockRate);
		printf("  totalConstMem : %d \n",prop.totalConstMem);
		printf("  multiProcessorCount : %d \n",prop.multiProcessorCount);
	}

	data.init(xinc_d,Y_d,corrLevelColPtr,seuil_choPtr,ndPtr,nkdPtr,nsimPtr,npositionPtr,nLevelFixPtr,nLevelVarPtr,npPtr,sizeFamilyNp);
	work.initResolution(data) ;
	solution.init(data);

	if ( data.ntniv > MAXNTNIV_LOCALTHREAD_MEMORY ) {
		printf("ntniv too big......maximum allowed :%d",MAXNTNIV_LOCALTHREAD_MEMORY);
		return;
	}

	err = (int)cuInit(0);
	if ( err ) {
		printf(" cuInit error : %d \n",err);
		exit(1);
	}
	//cublasInit();
	print_info_memory();
	int nbblock = ceil(data.npos / MAX_BLOCKDIM) + 1 ;

	// ***********************************
	// Allocate vectors in device memory
	// ***********************************

	printf("\nNB BLOCKS  : %d \n",nbblock);
	printf("\nNB THREADS : %d \n",MAX_BLOCKDIM);

	/* X'.X */
	XT_X<<<nbblock,MAX_BLOCKDIM>>>(data,work);
	cudaThreadSynchronize();
	//printf("X'.X\n----------------------------------------------------------\n");
	//outilDT.printFloatDeviceArray3D(data.npos,data.ntniv,data.ntniv,2,data.ntniv,data.ntniv,work.XX);

	/* Cholesky decomp */
	Cholesky_Decomposition<<<nbblock,MAX_BLOCKDIM>>>(data,work);
	cudaThreadSynchronize();
	//printf("Cholesky decomp\n----------------------------------------------------------\n");
	//outilDT.printFloatDeviceArray3D(data.npos,data.ntniv,data.ntniv,2,data.ntniv,data.ntniv,work.triang);

	/* RHS */
	Set_RHS<<<nbblock,MAX_BLOCKDIM>>>(data,work);
	cudaThreadSynchronize();
	//printf("RHS\n----------------------------------------------------------\n");
	//outilDT.printFloatDeviceArray3D(data.npos,data.nsim,data.ntniv,2,1,data.ntniv,work.rhs);

	/* Solution */
	Resolve_LU<<<nbblock,MAX_BLOCKDIM>>>(data,work,solution);
	cudaThreadSynchronize();
	//	printf("Resolution LU\n----------------------------------------------------------\n");
	//	outilDT.printFloatDeviceArray3D(data.npos,data.nsim,data.ntniv,2,1,data.ntniv,solution.bestim);

	/* Variances residuelles */
	Set_SIGSQ<<<nbblock,MAX_BLOCKDIM>>>(data,work,solution);
	cudaThreadSynchronize();
	//printf("SIGSQ\n----------------------------------------------------------\n");
	//outilDT.printFloatDeviceArray3D(data.npos,data.nsim,data.np,2,1,data.np,solution.osigsq);

	/* Recuperation des resultats */
	Utils<DT>::getArrayDeviceToHost(nbblock,MAX_BLOCKDIM,data.npos*data.nsim*data.np,solution.osigsq,osigsquare);
	Utils<DT>::getArrayDeviceToHost(nbblock,MAX_BLOCKDIM,data.npos*data.nsim*data.np,solution.osigsq,osigsquare);

	//work.getAllSigDevice(data,nbblock,MAX_BLOCKDIM,osigsquare) ;
	work.getAllVecsolDevice(data,nbblock,MAX_BLOCKDIM,vecsol) ;

	/* On libere la memoire du GPU partiellement */
	work.releaseDeviceResolution();

	/* Calcul LRT */

	work.initCalculLRT(data,sigsquare);
	Set_LRT<<<nbblock,MAX_BLOCKDIM>>>(data,work,solution);
	cudaThreadSynchronize();

	/* Recuperation des LRTs */
	Utils<DT>::getArrayDeviceToHost(nbblock,MAX_BLOCKDIM,data.npos*data.nsim,solution.lrt,lrt);

	if (data.nsim < 1 ) { /* Si analyse seulement on cherche le maximum sur le Host */
		maxLRT[0]      = -999.9;
		maxPosition[0] = -1;
		for (int i=0;i<data.npos;i++) {
			if ( maxLRT[0] < lrt[i]) {
				maxLRT[0]      = lrt[i];
				maxPosition[0] = i+1 ; /* index pour fortran */
			}
		}
	} else { /* Sinon on recherche les maximum pour chaque simul en meme temps via cuda */
		solution.getAllMaxLrt(data,nbblock,MAX_BLOCKDIM,maxLRT,maxPosition);
	}

	work.releaseDeviceCalculLRT();
	data.releaseDevice();
	solution.releaseDevice();
}
